#!/usr/bin/perl
# *** P856 version 0.43          -- date: Jan 31 2017 ***
# Philips P856 emulator with 32k 16 bit words
#
# After boot, enter date (YY MM DD) and time (HH MM)
#
# ADAPTED TO DOM DOM4A OR DOM8/DOS5 predefined PRIORITY INTERRUPT LEVELS
#
my $version = 0.43;
use strict;
use bytes;
use Term::ReadKey;
use Time::HiRes qw(gettimeofday tv_interval);
use IO::Handle;

# CONFIGURATION default
#=============================================================================================
# specify default files to be attached
my $filelpt = "lp.txt";
my $logfile = "log.txt";
my $filemd0 = "dos5.img";
my $filemd1 = "emptyX1215pack.img";
my $lptwin  = "";
my $vt = "";

use constant LINUX => 0;
#use constant LINUX => 1;  # can be set in case operated under LINUX with XWINDOWS started

if(LINUX)
{
    # uses a vt for the lineprinter,
    $vt = "/dev/pts/";
    system("xterm -title \"LPT\" -e \"perl vt.pl\" &"); # open an xterm for the lineprinter
}

my $DEBUG    = 0;     # 1 to show debug information

##############################################################################################

# limitations:
#=============
# 1) there is no "user mode", the CPU operates always in "system mode".
#    (LKM, traps, interrupts switch to system mode by setting the UM flipflop to 0. 
#     The SMD operation sets the UM flopflop to 1. As the cpu operates always in 
#     system mode however: there is no checking on illegal use of registers and 
#     instructions in user mode)
# 2) high speed arithmetic and the multiplex channel (IOP) are implemented (no dma).
#
# version 0.1 is based on P860 version 0.9 -- date: June 30 2008
#
# modifications:
# ==============
# version 0.2
#  - support 4 moving head disk spindels (iso 2) on two disk units on a single CU
#  - time measurement changed because it went wrong with Continue
#  - rtc 10Hz => 50Hz (because the DOM monitor likes that)
#  - setting of the GC flipflops (set mode instruction) is changed
#  - trap mechanism (for location 7E) changed
#
# version 0.3
#  - changes the moving head disk drive ready logic
#  - bug removed from TM and TNM instructions
#
# version 0.31
#  - comments added
#
# version 0.32 (04012009)
#  - trap /7E handling changed (invalid instruction:
#    also operating when interrupt is disabled and the PLR is not changed.
#  - RTN A15 copies ENB from stack (and does not set ENB like on P855/60/52)
#  - UM flipflop handled as it should be (interrupts, traps, SMD operation)
#    checking on invalid instruction/register usage in User Mode is not implemented.
#
# version 0.33 (25012009)
#  - load command addapted to proper process file extensions
#
# version 0.34 (08092009)
#  - bug removed from NGR instruction
#
# version 0.35 (20042010)
#  - bug load command (files can be loaded with .LM or .lm extension)
#  - comments in C1S instruction changed
#  - X1215 seek code simplified 
#  - ldhex command added
#  - To start with ENB after reset. change line 5975 in reset_cpu()
#
# version 0.36 (31102010)
#  - bug DV instruction (negative 2-complement result was wrong sometimes) 
#
# version 0.37 (07012011)
#  - add logging 
#  - make the dump command able to log its output (no other parts of the program use the prlog function yet)
#
# version 0.38 
#  - div: code changed (18072011)
#  - div: 1st test divisor, then dividend on zero (04082011)
#
# version 0.39 
#  - trace now always displays on console, also when using a trace file (30092011)
#  - TTY: Enter will always respond with a 0D (Windows and Linux)       (24102011)
#
# version 0.40 (31122011)
#  - pc as name added to A0 + some editing
#
# version 0.41 (18062012)
#  - CR set by RER instruction
#  - implicit ENB in User Mode <<<<< removed again
#  - LKM in inhibit mode skips DATA after LKM to exe the next instruction
#
# version 0.42 (17022013)
#  - start with default files attached and rtc on
#  - under LINUX a virtual terminal can be used to show lineprinter output
#  - logging 
#  - attach/detach md0,md1,md2,md3 changed
#
# version 0.43 (29012017)
#  - seek mechanism X1215 changed
#  - trace r option: only traces program instructions in a particular region, including
#    symbolic tracing
##############################################################################################

#
# Author: Theo Engel
# Info:   Info@theoengel.nl
#
# Usage:  p856-xx.pl [<do-script-name>]  (where xx is some version number)
#
# The emulator is built as a prototype, because at least some hardware details of the original 
# P856 are still lacking. The emulator is partly based on concepts used in SIMH (e.g. the event 
# mechanism). Also the control commands mirror some of the SIMH scp commands.
# (for SIMH, look at simh.trailing-edge.com)
#
# Currently implemented control commands:
# - run [<address>]										(includes reset-all)
# - go  [<address>]                                     (like run, but no reset)
# - continue|c
# - step|s [<nbr of instructions to execute>]           (no argument = execute 1 instruction)
# - do <filename>                                       (filename is name of the do-script)
# - breakpoint|bp set|s   <breakpoint address>
# - breakpoint|bp clear|c <breakpoint address>|all|a
# - breakpoint|bp print|p                               (show active breakpoints)
# - examine|ex <register>                               (show register content)
# - examine|ex m <address>                              (interactively shows content of <address>, <address>+2, .. until s is replied)
# - deposit|de <register> [<16 bit hex value>]          (in case no content specified, then it is requested from the keyboard)
# - deposit|de m <address>                              (interactively changes content of <address>, <address>+2, .. until s is replied)
# - deposit|de m <address> <16 bit hex value>
# - reset [cpu|tty|rtc|ptr|ptp|lpt|crd|md0|md1|md2|md3|mem|power] (no argument = reset-all; reset power = reset-all + reset-mem + P=0)
# - trace|tr [on|off] [<filename>]                      (no argument = on; traceoutput => file in case file is specified)
# - rtrace|rt on|off <la> [<ba>] [<ea>]                 (trace instructions executed between <ba> and <ea>; <la>=load addr; <ba>,<ea> rel to <la>)
#                                                       (no <ba> => da=0; no <ea> => region is upto 0xFFFE)
# - load <filename.abs>                                 (load absolute (8*8 format) executable with an .abs extension)
# - load <filename.rel>|<filename.LM|lm> <address>|mon  (load relocatable with a .rel or .LM extension at <address>) (mon loads a monitor at address 0)
# - ldhex|ldh [s] <filename.hex>                        (load absolute (hex format); with s specified, file content is printed)
# - boot ptr [h|r]                                      (h=halt,r= run after load; no argument = run, starting at 1st address)
# - boot md0|md1|md2|md3 [h|r]                          (ipl from mdx; h=halt,r= run after load; no argument = run)
# - attach|att ptr|ptp|lpt|crd|md0|md1|md2|md3 <filename> (attach a simulated peripheral to a file)
# - detach|det ptr|ptp|lpt|crd|md0|md1|md2|md3
# - state [cpu|tty|rtc|ptr|ptp|lpt|crd|md0|md1|md2|md3]	(no argument = state cpu+tty+rtc+ptr+ptp+lpt+crd+md0+md1+md2+md3)
# - queue|q                                             (show event queue)
# - echo <string>                                       (character string is displayed on console)
# - dump|du <ba> <ea> [<fn.abs>]                        (dump in hex memory area on screen or on papertape in abs 8+8 format)
# - console 64|128                                      (select either 64 or 128 character set; 64 is default)
# - rtc [on|off]                                        (start/stop realtime clock; ask state in case of no argument)
# - log [on|off] | <file>                               (log command progress in a logfile; off switches logging off; on resumes logging)
# - exit
#
# |    means logical OR
# [..] means OPTIONAL argument
# <address> is hex value of 0..FFFF
# <register> is P,A0..A15
# <string> is double quoted character string and may include codes like \n
# The key combination CntrE stops the emulation
# The key combination CntrP simulates a Panel interrupt
# do-scripts do not support arguments
#
# The following devices are currently are supported:
# 1) TTY - keyboard/printer
# 2) High speed papertape reader (PTR)  - Digitronics 25540EP, 333/600 cps
# 3) High speed papertape punch (PTP)   - Facit 4070, 75 cps
# 4) Real time clock (RTC)              - Variable frequency (set to 50Hz)
# 5) Line printer (LPT)                 - Data products, 132col, 245/670 lpm
# 6) Card reader (CRD)                  - Documation M200, 285 cpm
# 7) Moving head disk (MD0,MD1,MD2,MD3) - 2 * Philips X1215 (two units, 4 spindels)
# 8) Single IOP (address 0), connecting: LTP, CRD and MD0/MD1/MD2/MD3 on 3 subchannels
#
# Adaptions required to run another preconfigured monitor (without adding peripherals/iop's):
# 1) allocated priority interrupts (bit constants) 
# 2) allocated device addresses (%DEV)
#
# Questions/Remarks
# 1) which part of the PSW is stored/restored on/from the stack?
#    assumptions:
#    - CF           : whole PSW stored on the stack
#    - Interrupt    : whole PSW stored on the stack
#    - Depending on the register being used, part of the PSW is restored with RTN:
#      RTN always restores saved CR; 
#      RTN A15 also restores the saved PLR/UM-bit from stack and enables interrupt (P855,P860,P852)
#      (different on P856,P857: copies ENB bit from stack) 
#      However this implementation mimics the P855/P860/P852 behaviour at this moment: changed in version 0.32 to P856/57 behaviour
#
# 2) reset state? ENB=1? PL=00? 
#    The P856/57 Service Manual provides the answer (page 1-6 chapter 1.20) "At power-on time
#    and at every Master Clear from the control panel:
#    - the CPU is set to Enable interrupt mode,
#    - program level is set to 63
#    - all internal and external interrupts are reset
#
#    This implementation has at this moment for debugging: ENB=0 an PL=63 as reset state
#
# 3) adapt LDKL, ADKL SUKL so they load proper A0?
#     CWK instruction changed to make it possible to compare the long constant
#     with the right value of register P (instruction address + 4)
#     still adapt LDKL, ADKL SUKL ?? (other long constant may not have A0 as r1)
#
# 4) The double shift instructions DLA,DRA,DLN,DRN bypass bit-0 of A2
#    On the P855/P860?? and certainly on P852/P856/P857 bit-0 of A2 is
#    not changed during shift actions. On later models P854 .. bit-0
#    of A2 is always reset to 0 (also in case of shift count = 0). In this
#    implementation bit-0 of A2 is always reset to 0 (at this moment)
#
# 5) For some instructions on the real machine the l/s bit is not significant 
#    and may contain a 0 or a 1. The assembler puts in those cases always a 0
#    in the l/s bit In this implementation (at this moment) a 0 is always assumed 
#    in the l/s bit in those cases. A 1 will result in the detection of an illegal
#    instruction.
#
# Todo:
# 1) The disk ready mechanism needs a change to make it possible to
#    change a disk (cartridge) after to system is booted
# 2) The trap mechanism needs a change: a trap is not a (priority) interrupt
#    and is also possible when the CPU is in inhibit mode. A trap does
#    not change the priority level (operates like a CFI A15,/7E)
#    (done in version 0.32)
#
# Differences with the P855/860 (on hsi level)
# - interrupt system: 
#   -interrupt locations at other place in memory			V
#   -no common line: all interrupts handled by priority		V
#   -RTN A15 restores PLR, CR, ENB, SU from the stack
#    (on P855,860 and 852, ENB is always set to 1)          V
#   -remove P855/P860 set mask/read mask instructions       V
#   -on P855/860 a  stackoverflow interrupt is generated
#    when the A15 stackpointer is /100. On the P856/57
#    the stack overflow interrupt is generated when the
#    A15 stack becomes < /100                               V
#   -on the P855/860 an invalid instruction code generates
#    an LKM interrupt (when the interrupt is enabled). On the
#    P852/56/57 an invalid instruction generates a trap to
#    location /7E also when the interrupt is not enabled.   V
#    the trap acts like an interrupt: 
#    - invalid instruction not attemted to be executed
#    - instruction address and processpor state saved in stack
#    - interrupts are inhibited (if notyet inhibited)
#    - an indirect branche is made to location /7E: to the trap routine
#    - the program level (PLR) is NOT changed.
#    (a trap to /7E is also generated when a privileged instruction is
#     executed in User Mode or when the A15 stack is use in User Mode.
#     This behaviour is however not(yet) implemented)
# - multiplex channel replaced by IO processors
#   -IO processor channels controlled by WER instructions and external registers V
#   -BOOT from device using IOP iso MPX is different        V
# - device drivers
#   - other interrupt mechanism (priority iso common line)  V
#   - mpx => iop registers                                  V
# - a few additional instructions:
#   - 2-COMPLEMENT: NGR                                     V
#   - CLEAR MEMORY: CMR, CM, CM*                            V
#   - EXECUTE: EXR, EXR*, EXK, EX, EX*                      V
#   - WRITE EXTERNAL REGISTER: WER							V
#   - READ  EXTERNAL REGISTER: RER                          V
# - RTC: no start/stop with CIO instruction but via
#   control panel                                           V
# - Moving head disk differs in several ways
#   -X1210 had 203 cylinders, the X1215 (connected to P856) had 204 cylinders
#   -CIO SEEK operation (absolute cylinder iso delta)       V
#   -2 disks/4 heads with single actuator position per unit V
#   -drive ready mechanism
#   -status word bits                                       V
#   -CU supports 2 units/4 drives                           V

# ascii conversion table
my %ch;
$ch{0x20}=' ';
$ch{0x21}='!';
$ch{0x22}="\"";
$ch{0x23}="\#";
$ch{0x24}='$';
$ch{0x25}='%';
$ch{0x26}='&';
$ch{0x27}="\'";
$ch{0x28}='(';
$ch{0x29}=')';
$ch{0x2a}='*';
$ch{0x2b}='+';
$ch{0x2c}=',';
$ch{0x2d}='-';
$ch{0x2e}='.';
$ch{0x2f}="\/";
$ch{0x30}='0';
$ch{0x31}='1';
$ch{0x32}='2';
$ch{0x33}='3';
$ch{0x34}='4';
$ch{0x35}='5';
$ch{0x36}='6';
$ch{0x37}='7';
$ch{0x38}='8';
$ch{0x39}='9';
$ch{0x3a}="\:";
$ch{0x3b}="\;";
$ch{0x3c}='<';
$ch{0x3d}='=';
$ch{0x3e}='>';
$ch{0x3f}='?';
$ch{0x40}='@';
$ch{0x41}='A';
$ch{0x42}='B';
$ch{0x43}='C';
$ch{0x44}='D';
$ch{0x45}='E';
$ch{0x46}='F';
$ch{0x47}='G';
$ch{0x48}='H';
$ch{0x49}='I';
$ch{0x4a}='J';
$ch{0x4b}='K';
$ch{0x4c}='L';
$ch{0x4d}='M';
$ch{0x4e}='N';
$ch{0x4f}='O';
$ch{0x50}='P';
$ch{0x51}='Q';
$ch{0x52}='R';
$ch{0x53}='S';
$ch{0x54}='T';
$ch{0x55}='U';
$ch{0x56}='V';
$ch{0x57}='W';
$ch{0x58}='X';
$ch{0x59}='Y';
$ch{0x5a}='Z';
$ch{0x5b}='[';
$ch{0x5c}="\\";
$ch{0x5d}=']';
$ch{0x5e}='^';
$ch{0x5f}='_';

 # digit conversion
my %cvdig;             
$cvdig{'0'}=0x0;
$cvdig{'1'}=0x1;
$cvdig{'2'}=0x2;
$cvdig{'3'}=0x3;
$cvdig{'4'}=0x4;
$cvdig{'5'}=0x5;
$cvdig{'6'}=0x6;
$cvdig{'7'}=0x7;
$cvdig{'8'}=0x8;
$cvdig{'9'}=0x9;
$cvdig{'A'}=0xa;
$cvdig{'B'}=0xb;
$cvdig{'C'}=0xc;
$cvdig{'D'}=0xd;
$cvdig{'E'}=0xe;
$cvdig{'F'}=0xf;

# register conversion table
my %R16;
$R16{"P"}   = 0;	# 0000
$R16{"PC"}  = 0;	# 0000
$R16{"A0"}  = 0;	# 0000
$R16{"A1"}  = 0x2;	# 0010
$R16{'A2'}  = 0x4;	# 0100
$R16{'A3'}  = 0x6;	# 0110
$R16{'A4'}  = 0x8;	# 1000
$R16{'A5'}  = 0xA;	# 1010
$R16{'A6'}  = 0xC;	# 1100
$R16{'A7'}  = 0xE;	# 1110
$R16{'A8'}  = 0x1;	# 0001
$R16{'A9'}  = 0x3;	# 0011
$R16{'A10'} = 0x5;	# 0101
$R16{'A11'} = 0x7;	# 0111
$R16{'A12'} = 0x9;	# 1001
$R16{'A13'} = 0xB;	# 1011
$R16{'A14'} = 0xD;	# 1101
$R16{'A15'} = 0xF;	# 1111

# register index conversion table
my @A;
$A[0]  = 0x0;
$A[1]  = 0x2;
$A[2]  = 0x4;
$A[3]  = 0x6;
$A[4]  = 0x8;
$A[5]  = 0xA;
$A[6]  = 0xC;
$A[7]  = 0xE;
$A[8]  = 0x1;
$A[9]  = 0x3;
$A[10] = 0x5;
$A[11] = 0x7;
$A[12] = 0x9;
$A[13] = 0xB;
$A[14] = 0xD;
$A[15] = 0xF;

my @Reg;
$Reg[0]  = 'A0';
$Reg[1]  = 'A8';
$Reg[2]  = 'A1';
$Reg[3]  = 'A9';
$Reg[4]  = 'A2';
$Reg[5]  = 'A10';
$Reg[6]  = 'A3';
$Reg[7]  = 'A11';
$Reg[8]  = 'A4';
$Reg[9]  = 'A12';
$Reg[10] = 'A5';
$Reg[11] = 'A13';
$Reg[12] = 'A6';
$Reg[13] = 'A14';
$Reg[14] = 'A7';
$Reg[15] = 'A15';

# register symbols
use constant P   => 0x0;
use constant PC  => 0x0;
use constant A0  => 0x0;
use constant A1  => 0x2;
use constant A2  => 0x4;
use constant A3  => 0x6;
use constant A4  => 0x8;
use constant A5  => 0xA;
use constant A6  => 0xC;
use constant A7  => 0xE;
use constant A8  => 0x1;
use constant A9  => 0x3;
use constant A10 => 0x5;
use constant A11 => 0x7;
use constant A12 => 0x9;
use constant A13 => 0xB;
use constant A14 => 0xD;
use constant A15 => 0xF;

# instruction decoder tables
my %IDir;  # direct translated instruction codes
my %IDirl; # direct translated instruction codes
my %IL1;   # format 1 long constant instruction codes (r2=0)
my %ILS;   # format 1 instructions with r1=0
my %I1;    # format 1 instruction codes, using r1 and r2
my %IFPP;  # format 1 FPP instructions
my @R2;    # register strings: A0..A15
my @R3;    # register strings: A0..A7
my @CND;   # condition codes: (0)..(7)


####################### emulator controller data ############################
my $testflag       = 0;
my $StopOnIllInstr = 1;     # stop on illegal instruction
my $StopOnIllDev   = 1;     # stop on illegal device
my $StopOnMpxError = 1;     # stop on an illegal mpx condition

my $tracefile      = "";    # tracefile name in case specified
my $EmTrace = 0;			# trace on if 1, trace off 0
my $latrace = 0;            # regional trace: load address of program
my $batrace = 0;            # regional trace: region start (relative to la)
my $eatrace = 0;            # regional trace: region end   (relative to la)
my $EmBreak = 0;            # 1 in case a breakpoint exist, else 0
my %BP;                     # breakpoint hash
my $program_started = 0;    # 1 if new program is started by RUN/GO/STEP
                            # reset, by resel all command
# papertape format
my $null 	= 0;
my $rub     = 0377;         # start of tape
my $xoff    = 0224;         # end of tape

# time/event handling
my $interval;               # decreasing: number of instructions still to be executed till next event
my $currentinterval;		# fix: number of instructions to be executed till next event
my $event_queue;
my $queue_empty_time = 5000;
my $time;                   # accumulated number of executed instructions
my $ctime;                  # current time: always calculated form $time+executed part of current interval

# elapsed runtime measurement
my @t0;		                # start point elapsed time measurement during run command
my $secs;                   # run elapsed time in secs
my $cps;                    # cycles per second during run command
# real time clock synchronization
my @clk0;                   # start point of measurement period
my $laps;                   # actual measured period in second
my $pulses;                 # number of events in measured period

############################ emulator data ##################################
# cpu/memory resources and state
#===============================
my @M=();               # memory array
use constant MEMSIZE    => 0x00010000;       # 32 kW;
use constant MAXADDR    => 0x0000FFFE;
my $IR;                 # instruction register
my $PSW;                # status word
	my $PL;             # priority level 			psw bit 0:5
	my $CR;             # condition register		psw bit 6:7
	my $overflow=0;     # (aid to set CR, must be 0)
	my $RUN;            # run bit               	psw bit 8
	my $ENB;            # interrupt enable bit		psw bit 9
#	   CP               # control panel interrupt	psw bit 10
#	   PWR              # power fail interrupt      psw bit 11
#	   RTC              # real time clock interrupt psw bit 12
#	   LKM              # link monitor interrupt    psw bit 13
#						# + stack overflow 
#	   ---              # not used                  psw bit 14 
	my $UM;             # user mode indicator       psw bit 15
my @R;                  # registers A0..A15
my $pcsaved;

# Internal Interrupts (traps) P852,P856,P857
# ==========================================
# - Power Failure/Auto Restart
# - Link Monitor/Stack Overflow (A15 pointer < /100)
# - Real Time Clock
# - Control Panel
# - (Invalid Instruction = trap /7E which also works in INH mode)
# - Page Fault (for P800 with MMU; P857)
# - Floating Point Unit interrupt (for P800 with FPU; P857)

# Some internal interrupts are also part of the PSW (see above)
# Internal Interrupt register (last 5 bits are part of PSW: bits 10:14):

my $inint=0;      # internal interrupt register

# the bits in this register are allocated as follows (according PSW):
use constant CP  => 0x10;  # control panel interrupt, bit 3
use constant PWR => 0x08;  # power failure, bit 4
use constant RTC => 0x04;  # real time clock interrupt, bit 5
use constant LKM => 0x02;  # link monitor interrupt + stack overflow, bit 6
# 4 spares: the RIT instruction is able to handle 8 bits
# traps are SET by the trap subroutine, which sets the bit in $inint
# and in $intreg
# traps are RESET by the resettrap subroutine, which resets the bit in $inint
# and in $intreg

# Interrupt system
# ================
# The interrupt system is a priority system with 64 levels. The 
# interrupt lines are connected according priority to memory locations
# /0000 .. /007E, where /0000 represents the highest priority.

my $intreg=0;     # interrupt register 1st 32 bits

# The interrupt lines are linked with an interrupt flipflops collected in a
# interrupt register. The lines/flopflops are arranged in the register in 
# priority order. Highest priority is the most significant bit
#
# DOM CONFIGURATION (OF THE DOS5.IMG and p8d10) configured to boot from unit-0 (md1)
use constant IPWR   => 0x80000000;     # Power failure, auto restart (0) internal interrupt
use constant ILKM   => 0x40000000;     # Link monitor, stackoverflow (1) internal interrupt
use constant IRTC   => 0x20000000;     # Real time clock             (2) internal interrupt
use constant IPTR   => 0x08000000;     # Papertape reader            (4)
use constant IPTP   => 0x04000000;     # Papertape punch             (5)
use constant IASR   => 0x02000000;     # Console typewriter          (6)
use constant ICP    => 0x01000000;     # Panel interrupt             (7) internal interrupt
use constant IMD0   => 0x00008000;     # X1215 Moving head disk      (10)
use constant IMT    => 0x00001000;     # Mag tape                    (13)
use constant ITK    => 0x00000800;     # Cas tape                    (14)
use constant ILP    => 0x00000100;     # Line printer                (17)

# DOM4A/DOS4 (DOM from SAGHAR 4*X1215, boot from md3)
#use constant IPWR   => 0x80000000;     # Power failure, auto restart (0) internal interrupt
#use constant ILKM   => 0x40000000;     # Link monitor, stackoverflow (1) internal interrupt
#use constant IRTC   => 0x20000000;     # Real time clock             (2) internal interrupt
#use constant IPTR   => 0x08000000;     # Papertape reader            (4)
#use constant IPTP   => 0x04000000;     # Papertape punch             (5)
#use constant IASR   => 0x02000000;     # Console typewriter          (6)
#use constant ICP    => 0x01000000;     # Panel interrupt             (7) internal interrupt
#use constant IMD0   => 0x00008000;     # X1215 Moving head disk      (10) <<<< CHANGED FROM 11 => 10
#use constant IMT    => 0x00001000;     # Mag tape                    (13)
#use constant ICR    => 0x00000400;     # Card reader                 (15)
#use constant ILP    => 0x00000100;     # Line printer                (17)

# DOS5 (p8d4) boot from md1/DOS8
#use constant IPWR   => 0x80000000;     # Power failure, auto restart (0) internal interrupt
#use constant ILKM   => 0x40000000;     # Link monitor, stackoverflow (1) internal interrupt
#use constant IRTC   => 0x20000000;     # Real time clock             (2) internal interrupt
#use constant IASR   => 0x02000000;     # Console typewriter          (6)
#use constant ICP    => 0x01000000;     # Panel interrupt             (7) internal interrupt
#use constant ILP    => 0x00040000;     # Line printer                (D)
#use constant IPTP   => 0x00020000;     # Papertape punch             (E)
#use constant IPTR   => 0x00010000;     # Papertape reader            (F)
#use constant IMD0   => 0x00004000;     # X1215 Moving head disk      (11) <<<< CHANGED FROM 10 => 11
#use constant IMT    => 0x00001000;     # Mag tape                    (13)
use constant ICR    => 0x00000400;     # Card reader                 (15)

# DOST (p8d4dz) (boot md0)
#use constant IPWR   => 0x80000000;     # Power failure, auto restart (0) internal interrupt
#use constant ILKM   => 0x40000000;     # Link monitor, stackoverflow (1) internal interrupt
#use constant IRTC   => 0x20000000;     # Real time clock             (2) internal interrupt
#use constant IASR   => 0x02000000;     # Console typewriter          (6)
#use constant ICP    => 0x01000000;     # Panel interrupt             (7) internal interrupt
#use constant IPTP   => 0x00020000;     # Papertape punch             (E)
#use constant IPTR   => 0x00010000;     # Papertape reader            (F)
#use constant IMD0   => 0x00008000;     # X1215 Moving head disk      (10)

#
# The internal interrupts are linked to the priority interrupt system
# Mapping:
my @II2PI;
$II2PI[CP]  = ICP;  #intreg
$II2PI[PWR] = IPWR; #intreg
$II2PI[RTC] = IRTC; #intreg
$II2PI[LKM] = ILKM; #intreg
#
# Priority Interrupt Locations (not used by the simulator; for documentation only)
#
#                   Mem Address         Bit in IntReg
use constant AdPWR   => 0x00;           # 8000 0000 power failure interrupt location (highest priority = level 0)
use constant AdLKM   => 0x02;           # 4000 0000 LKM/Stack overflow (priority level 1) 
use constant AdRTC   => 0x04;           # 2000 0000 real time clock (priority level 2)
use constant ilres1  => 0x06;           # 1000 0000 reserved (priority level 3)
use constant AdPTR   => 0x08;           # 0800 0000 punched tape reader (priority level 4)
use constant AdPTP   => 0x0A;           # 0400 0000 tape punch (priority level 5)
use constant AdASR   => 0x0C;           # 0200 0000 operator's typewriter (priority level 6)
use constant AdCP    => 0x0E;           # 0100 0000 control panel (priority level 7)
use constant ilres2  => 0x10;           # 0080 0000 reserved (priority level 8)
use constant ilres3  => 0x12;           # 0040 0000 reserved (priority level 9)
use constant ilres4  => 0x14;           # 0020 0000 reserved (priority level A)
use constant ilres5  => 0x16;           # 0010 0000 reserved (priority level B)
use constant ilres6  => 0x18;           # 0008 0000 reserved (priority level C)
use constant ilres7  => 0x1A;           # 0004 0000 reserved (priority level D)
use constant ilres8  => 0x1C;           # 0002 0000 reserved (priority level E)
use constant ilres9  => 0x1E;           # 0001 0000 reserved (priority level F)

#use constant ilres10 => 0x20;          # 0000 8000 reserved (priority level 10)
use constant AdMD0   => 0x20;           # 0000 8000 X1215    (priority level 10)
use constant ilres11 => 0x22;           # 0000 4000 reserved (priority level 11)
#use constant AdMD0   => 0x22;          # 0000 4000 X1215    (priority level 11)

use constant ilres12 => 0x24;           # 0000 2000 reserved (priority level 12)
use constant AdMT    => 0x26;           # 0000 1000 Mag.Tape (priority level 13)
use constant AdTK    => 0x28;           # 0000 0800 Cas.Tape (priority level 14)
use constant AdCR    => 0x2A;           # 0000 0400 Card Reader (priority level 15)
use constant ilres13 => 0x2C;           # 0000 0200 reserved (priority level 16)
use constant AdLP    => 0x2E;           # 0000 0100 Line Printer (priority level 17)
use constant ilres14 => 0x30;           # 0000 0080 reserved (priority level 18)
use constant ilres15 => 0x32;           # 0000 0040 reserved (priority level 19)
use constant ilres16 => 0x34;           # 0000 0020 reserved (priority level 1A)
use constant ilres17 => 0x36;           # 0000 0010 reserved (priority level 1B)
use constant ilres18 => 0x38;           # 0000 0008 reserved (priority level 1C)
use constant ilres19 => 0x3A;           # 0000 0004 reserved (priority level 1D)
use constant ilres20 => 0x3C;           # 0000 0002 reserved (priority level 1E)
use constant ilres21 => 0x3E;           # 0000 0001 reserved (priority level 1F), 
use constant AdILL   => 0x7E;           # 0000 0001 Invalid Instruction Trap (priority level 3F)
													 # activated in case of:
													 # - Non-exsisting instructions
													 # - Privileged instructions executed in User Mode
													 # - System stack access (A15) in User Mode

# working registers for CPU, CU's, etc simulation
my $opc;                     # operation code bit1:4
my $r1;                      # r1 or cnd shifted 1 left	
my $r2;                      # r2
my $r3;                      # r3             (in format 0 instructions)
my $k;                       # short constant (in format 0 instructions)
my $sc;                      # shiftcount     (in format 0 instructions)
my $sfun;                    # shift function (in format 0 instructions)
my $md;                      # adressing mode (format 1)
my $ls;                      # load/store bit (format 1)
my $t1;                      # temporary 1 (value)
my $t2;                      # temporary 2 (value)
my $t3;                      # temporary 3 (value)
my $a;                       # temporary for effective address computation

# masks
use constant mskf10 => 0x8000;      	    		# format 1 instruction with ls bit 0

use constant SIGN    => 0x8000;               	# sign bit short
use constant DPSIGN  => 0x40000000;		        	# 31 bit double precision sign (in 32bit word)
use constant DMASK   => 0xFFFF;        	      # data mask short
use constant DPDMASK => 0x7FFFFFFF;           	# 31 bit double precision data mask (in 32bit word)
use constant MMASK   => DMASK & (~ SIGN); 	  	# magnitude mask short
use constant DPMMASK => DPDMASK & (~ DPSIGN); 	# 31 bit magnitude mask (in 32bit word)

# devices/control units
#
# The control unit STATE consists of two parts
# 1) ls 16 bits  = device status word, different for each device, but
#    bit 15 is 1 always means: device not operable
#
# 2) 4 bits before the ls 16 bits
#    
#    bit 2:    0 = output request
#              1 = input request
#    bit 3:4  00 = inactive state
#             01 = execute state
#             11 = exchange state
#             10 = wait state
#    bit 1     1 = stop io requested
#    the device is busy if not in inactive state
#
# control unit states (the 4 bits as part of STATE)
use constant INAC => 0;                  # inactive
use constant EXE  => 1;                  # execute
use constant IEXE => (0b100 + 0b001);    # input  + execute
use constant OEXE => (0b000 + 0b001);    # output + execute
use constant EXC  => 3;                  # exchange
use constant IEXC => (0b100 + 0b011);    # input  + exchange
use constant OEXC => (0b000 + 0b011);    # output + exchange
use constant WAIT => 2;                  # wait state
use constant WREQ => 0b1000;             # stop io (or wait state) requested

# device control block declaration
my @TTY;
my @PTR;
my @PTP;
my @CLK;
my @LPT;
my @CRD;
my @MD0;
my @MD1;
my @MD2;
my @MD3;

# device control block layout
use constant NEXT			=> 0;
use constant NAME   		=> 1;
use constant TIMEOUT		=> 2;
use constant SERVICE   	=> 3;
use constant STATE   	=> 4;
use constant BUFFER		=> 5;
use constant TIME			=> 6;
use constant CIO   		=> 7;
use constant INR			=> 8;
use constant OTR   		=> 9;
use constant SST			=> 10;
use constant TST   		=> 11;
use constant DSTAT      => 12;
use constant RESET      => 13;
use constant FILE       => 14;
use constant FC         => 15;
use constant RDY        => 16;

# On a 3,2 Ghz Core Duo the average instruction execution time is 4 musec. So a tick of
# the simulation clock in that case takes 4 musec. A 1 msec delay takes 250 ticks. By 
# adjusting the number of ticks in the device control blocks the timing of the peripherals
# can be set.

# tty device control block (teletype keyboard/printer - programmed channel)
# tty specific entries in device control block
use constant CSET => 14;     # 64 or 128 Ascii character set

$TTY[NEXT]  	= 0;							# next pointer field
$TTY[NAME]  	= "tty";                # device name
$TTY[TIMEOUT]  = 0;                    # event timeout value
$TTY[SERVICE]  = \&svc_tty;            # event service routine
$TTY[STATE]  	= 0;                    # control unit state
$TTY[BUFFER]  	= 0;                    # TTY buffer
$TTY[TIME]     = 12500;                # default device event time (50 msec)
$TTY[CIO]  		= \&cio_tty;            # cio processor
$TTY[INR]  		= \&inr_tty;            # inr processor
$TTY[OTR]  		= \&otr_tty;            # otr processor
$TTY[SST]  		= \&sst_tty;            # sst processor
$TTY[TST] 		= \&tst_tty;            # sst processor
$TTY[DSTAT]    = \&state_tty;          # display state
$TTY[RESET]    = \&reset_tty;          # reset device
$TTY[CSET]     = 64;                   # default is 64 character set 

# ptr device control block (paper tape reader - programmed channel )
$PTR[NEXT]     = 0;                    # next pointer field
$PTR[NAME]     = "ptr";                # device name
$PTR[TIMEOUT]  = 0;                    # event timeout value
$PTR[SERVICE]  = \&svc_ptr;            # event service routine
$PTR[STATE]  	= 0;                    # control unit state
$PTR[BUFFER]  	= 0;                    # PTR buffer
$PTR[TIME]     = 750;                  # default device event time (3 msec)
$PTR[CIO]  		= \&cio_ptr;            # cio processor
$PTR[INR]  		= \&inr_ptr;            # inr processor
$PTR[OTR]  		= 0;                    # otr processor
$PTR[SST]  		= \&sst_ptr;            # sst processor
$PTR[TST] 		= \&tst_ptr;            # sst processor
$PTR[DSTAT]    = \&state_ptr;          # display state
$PTR[RESET]    = \&reset_ptr;          # reset device
$PTR[FILE]     = "";                   # attached file (name; 0 if not attached/inoperable)
$PTR[FC];                              # file handle of attached file
$PTR[RDY];                             # device in operational state (not used yet for ptr)

# ptp device control block (paper tape punch - programmed channel)
$PTP[NEXT]     = 0;                    # next pointer field
$PTP[NAME]     = "ptp";                # device name
$PTP[TIMEOUT]  = 0;                    # event timeout value
$PTP[SERVICE]  = \&svc_ptp;            # event service routine
$PTP[STATE]  	= 0;                    # control unit state
$PTP[BUFFER]  	= 0;                    # PTP buffer
$PTP[TIME]     = 2500;                 # default device event time (10 msec)
$PTP[CIO]  		= \&cio_ptp;            # cio processor
$PTP[INR]  		= 0;                    # inr processor
$PTP[OTR]  		= \&otr_ptp;            # otr processor
$PTP[SST]  		= \&sst_ptp;            # sst processor
$PTP[TST] 		= \&tst_ptp;            # sst processor
$PTP[DSTAT]    = \&state_ptp;          # display state
$PTP[RESET]    = \&reset_ptp;          # reset device
$PTP[FILE]     = "";                   # attached file (name; 0 if not attached/inoperable)
$PTP[FC];                              # file handle of attached file
$PTP[RDY];                             # device in operational state (not used yet for ptp)

# crd device control block and buffer (card reader - multiplex channel )
# card reader specific entries in device control block
use constant RTIME      => 17;         # card moving time
use constant RCLEN      => 18;         # read number of bytes

my $crdbuf;                            # card reader buffer

$CRD[NEXT]     = 0;                    # next pointer field
$CRD[NAME]     = "crd";                # device name
$CRD[TIMEOUT]  = 0;                    # event timeout value
$CRD[SERVICE]  = \&svc_crd;            # event service routine
$CRD[STATE]  	= 0;                    # control unit state
$CRD[BUFFER]  	= 0;                    # buffer
$CRD[TIME]     = 5;                    # mpx character transfer time (20 musec)
$CRD[CIO]  		= \&cio_crd;            # cio processor
$CRD[INR]  		= \&inr_crd;            # inr processor
$CRD[OTR]  		= 0;                    # otr processor
$CRD[SST]  		= \&sst_crd;            # sst processor
$CRD[TST] 		= \&tst_crd;            # sst processor
$CRD[DSTAT]    = \&state_crd;          # display state
$CRD[RESET]    = \&reset_crd;          # reset device
$CRD[FILE]     = "";                   # attached file (name; 0 if not attached/inoperable)
$CRD[FC];                              # file handle of attached file
$CRD[RDY];                             # device in operational state (not used yet for crd)
$CRD[RTIME]    =25000;                 # time to move card to read station (100 msec)
$CRD[RCLEN]    = 0;                    # read length in bytes

# lpt device control block and buffer (line printer - multiplex channel)
# printer specific entries in printer device control block
use constant PTIME      => 17;         # buffer print time
use constant LMAX       => 18;         # lines per page printed
use constant NPAG       => 19;         # nbr of lines to skip between print pages
use constant LCUR       => 20;         # current line on page
# the following entry directs the page printing (skip page before or after printing the buffer).
use constant SKIP       => 21;          
# this entry must be set to one of the following two constants.
use constant SKIP_BP    => 1;          # skip page before printing the buffer
use constant SKIP_AP    => 2;          # skip page after  printing the buffer
use constant SKIP_NO    => 3;          # no printer control at all, store chars forwarded to lpt in the printfile

my $lptbuf;                            # lineprinter buffer

$LPT[NEXT]     = 0;                    # next pointer field
$LPT[NAME]     = "lpt";                # device name
$LPT[TIMEOUT]  = 0;                    # event timeout value
$LPT[SERVICE]  = \&svc_lpt;            # event service routine
$LPT[STATE]  	= 0;                    # control unit state
$LPT[BUFFER]  	= "";                   # buffer
$LPT[TIME]     = 5;                    # mpx character transfer time (20 musec)
$LPT[CIO]  		= \&cio_lpt;            # cio processor
$LPT[INR]  		= 0;                    # inr processor
$LPT[OTR]  		= \&otr_lpt;            # otr processor
$LPT[SST]  		= \&sst_lpt;            # sst processor
$LPT[TST] 		= \&tst_lpt;            # sst processor
$LPT[DSTAT]    = \&state_lpt;          # display state
$LPT[RESET]    = \&reset_lpt;          # reset device
$LPT[FILE]     = "";                   # attached file (name; 0 if not attached/inoperable)
$LPT[FC];                              # file handle of attached file
$LPT[RDY];                             # device in operational state (not used yet for lpt)
$LPT[PTIME]    = 25000;                # buffer print time (100 msec)
# page parameters: 66 lines per page, of which 64 being printed + 1 empty line at top and end
# (in case of no printer control chars and printer control is set by SKIP_AP or SKIP_BP)
$LPT[LMAX]     = 64;                   # 64 print lines per page
$LPT[NPAG]     = 2;                    # nbr of lines to skip between print pages
$LPT[LCUR]     = 0;                    # current line on page
$LPT[SKIP]     = SKIP_BP;              # skip page before printing the buffer
#$LPT[SKIP]     = SKIP_AP;              # skip page after  printing the buffer

# X1215 moving head disk device control blocks (1 per drive) connected to IOP subchannel 2
#
# two units per control unit cq 4 drives max. 
# - unit 0 has 2 disks: md0 (address /02 = cartridge) and md1 (address /22 = fixed disk)
# - unit 1 has 2 disks: md2 (address /12 = cartridge) and md3 (address /32 = fixed disk)
# - unit 0 is operational/ready in case both disks md0 and md1 are attached to a file
# - unit 1 is operational/ready in case both disks md2 and md3 are attached to a file
# The control unit and the disk/drive 0 data are in the md0 control block. 
# The drive specific data of the units/disks are in md1, etc, where md2 contains the unit
# specific data for unit 0. 
# Only one disk can use the control unit and subchannel at a time; even seeks can not be overlapped.
#
# moving head disk specific constants
use constant RWTIME => 17;              # latency to move to sector in cylinder
use constant TRTIME => 18;              # time to move 1 cylinder
use constant XLEN   => 19;              # transferred length in bytes
use constant DN     => 20;              # active disk/drive == drive with a pending operation
use constant CYL    => 21;              # current cylinder
use constant SEC    => 22;              # sector to read or write

# resulting state of a disk operation
#use constant DREADY   => 0x4000; # unit ready from not operable (was unready)
use constant DREADY   => 0x0000; # unit ready from not operable (was unready) <<< disk ready not supported at this moment
use constant DONCYL   => 0x0400; # on requested cylinder (seek complete)
use constant DSEEKERR => 0x0200; # seek error
use constant DDISKNBR => 0x0080; # disk/drive number: 0 for cartridge, 1 for the fixed disk
use constant DUNITNBR => 0x0040; # unit number: 0 or 1
use constant DPROGERR => 0x0010; # program error
use constant DLENGERR => 0x0008; # incorrect length
use constant DDATAERR => 0x0004; # data error
use constant DTHROERR => 0x0002; # throughput error
use constant DNOTOP   => 0x0001; # not operable

# disk parameters (X1215)
# the p800 premark (PREMDK) wants to have a disk with 204 cylinders (as the X1215 had),
# otherwise there is a seek error while executing the premark. For nomal operation also
# a disk with 203 cylinders can be used, because only 200 cylinders are used during normal
# operation. The last 4 cylinders on the x1215 pack are spare (the last 3 on an x1210 pack)
# the x1215 pack (with 204 cylinders) can be formatted like the x1210 pack with 203 cylinders.
use constant CYLS     => 204;    # ncyl 0..203 (204 tracks/cylinders)
use constant NH       => 2;      # number of heads (per disk/drive)
use constant SPT      => 16;     # 16 sectors per track
use constant LSEC     => 205;    # sector length in 16 bit words

my $md0buf;                            # internal disk buffer

# unit0 drive0 + control unit
# cartridge of unit 0 ($dn=0 00)
$MD0[NEXT]     = 0;                    # next pointer field
$MD0[NAME]     = "md0";                # device name
$MD0[TIMEOUT]  = 0;                    # event timeout value
$MD0[SERVICE]  = \&svc_md0;            # event service routine
$MD0[STATE]  	= 0;                    # control unit state
$MD0[BUFFER]  	= "";                   # buffer
$MD0[TIME]     = 5;                    # mpx transfer time (20 musec)
$MD0[CIO]  		= \&cio_md0;            # cio processor/drive selector
$MD0[INR]  		= \&inr_md0;            # inr processor
$MD0[OTR]  		= \&otr_md0;            # otr processor
$MD0[SST]  		= \&sst_md0;            # sst processor
$MD0[TST] 		= \&tst_md0;            # tst processor
$MD0[DSTAT]    = \&state_md0;          # display state
$MD0[RESET]    = \&reset_md0;          # reset device
$MD0[FILE]     = "";                   # attached file (name= "" if not attached/inoperable)
                                       # if attached, then also opened. filecode MD0
$MD0[FC];                              # not used
$MD0[RDY]      = 0;                    # 1 in case unit 0 is in operational state (READY)
$MD0[RWTIME]   = 3000;                 # latency to move to sector in cylinder (12 msec)
$MD0[TRTIME]   = 100;                  # time to move 1 cylinder (0.4 msec)
$MD0[XLEN]     = 0;                    # transferred length in bytes
$MD0[DN]       = 0;                    # active drive == drive with a pending operation
$MD0[CYL]      = 0;                    # current cylinder unit 0 
$MD0[SEC]      = 0;                    # physical sector to read or write

# unit0 drive1
# fixed disk on unit 0 ($dn=2 10)
$MD1[NAME]     = "md1";                # device name
$MD1[CIO]  		= \&cio_md1;            # cio processor/drive selector
$MD1[SST]      = \&sst_md0;
$MD1[TST]      = \&tst_md0;
$MD1[FILE]     = "";                   # attached file (name= "" if not attached/inoperable)
                                       # if attached, then also opened. filecode MD1
# unit1 drive0
# cardridge on unit 1 ($dn=1 01)
$MD2[NAME]     = "md2";                # device name
$MD2[CIO]  		= \&cio_md2;            # cio processor/drive selector
$MD2[SST]      = \&sst_md0;
$MD2[TST]      = \&tst_md0;
$MD2[FILE]     = "";                   # attached file (name= "" if not attached/inoperable)
                                       # if attached, then also opened. filecode MD2
$MD2[RDY]      = 0;                    # 1 in case unit 1 is in operational state (READY)										
$MD2[CYL]      = 0;                    # current cylinder unit 1
# unit1 drive1
# fixed disk on unit 1 ($dn=3 11)
$MD3[NAME]     = "md3";                # device name
$MD3[CIO]  		= \&cio_md3;            # cio processor/drive selector
$MD3[SST]      = \&sst_md0;
$MD3[TST]      = \&tst_md0;
$MD3[FILE]     = "";                   # attached file (name= "" if not attached/inoperable)
                                       # if attached, then also opened. filecode MD3										

# clk device control block (10Hz clock)
my $rtc_on      = 0;                   # clock activate request flag
# clock specific entries in clock device control block
use constant FREQ  		=> 14;
use constant SYSEC      => 15;
use constant SYNC       => 16;
use constant CTIME      => 17;

$CLK[NEXT]     = 0;                    # next pointer field
$CLK[NAME]     = "rtc";                # device name
$CLK[TIMEOUT]  = 0;                    # event timeout value
$CLK[SERVICE]  = \&svc_clk;            # event service routine
$CLK[STATE]  	= 0;                    # control unit state
$CLK[BUFFER]  	= 0;                    # buffer
$CLK[TIME]     = 5000;                 # default device event time (20 msec)
$CLK[CIO]  		= \&cio_clk;            # cio processor
$CLK[INR]  		= \&inr_clk;            # inr processor
$CLK[OTR]  		= \&otr_clk;            # otr processor
$CLK[SST]  		= \&sst_clk;            # sst processor
$CLK[TST] 		= \&tst_clk;            # sst processor
$CLK[DSTAT]    = \&state_clk;          # display state
$CLK[RESET]    = \&reset_clk;          # reset device
$CLK[FREQ]     = 50;                   # 50 HZ clock
$CLK[SYSEC]    = 2;                    # sync measurement period (secs) 
$CLK[SYNC]     = $CLK[SYSEC] * $CLK[FREQ]; # number sync pulses for measurement period
$CLK[CTIME]    = 0;                    # calculated synced device event time

# defined devices (device addresses)
# the single IOP has IOP address 0
my %DEV;
$DEV{0x07} = \@LPT; # IOP (subchannel 7)
$DEV{0x10} = \@TTY;
$DEV{0x20} = \@PTR;
$DEV{0x30} = \@PTP;
$DEV{0x3F} = \@CLK;
$DEV{0x06} = \@CRD; # IOP (subchannel 6)
# MD connected to IOP subchannel 2
$DEV{0x02} = \@MD0; # IOP, CU has device address 2, disk/drive 0 unit 0, $dn=0 (cartridge)
$DEV{0x22} = \@MD1; # IOP, CU has device address 2, disk/drive 1 unit 0, $dn=2 (fixed)
$DEV{0x12} = \@MD2; # IOP, CU has device address 2, disk/drive 0 unit 1, $dn=1 (cardridge)
$DEV{0x32} = \@MD3; # IOP, CU has device address 2, disk/drive 1 unit 1, $dn=3 (fixed)

# speudo device for stepping function
my @STEP;             
$STEP[NEXT] 	= 0;							# next pointer field
$STEP[NAME] 	= "step";               # device name
$STEP[TIMEOUT] = 0;                    # event timeout value
$STEP[SERVICE] = \&svc_step;           # event service routine
# stepper service routine (scheduled after stepper event)
sub svc_step {
	$RUN = 0;          # stop cpu
}

# IOPs (I/O Processors) implement the Multiplex channels
# ======================================================
# Up to 8 IOPs can be connected, but most systems have only 1 or 2 of them installed.
# Each IOP supports 8 subchannels. IOP address + subchannel number provide the
# least significant bits of the device address:
#
# Device Address           X X X X X X (6 bits)
#     Subchannel                 S S S (3 bits)
#    IOP address           I I I       (up to 3 bits)
#
# The device control unit is connected to a subchannel. The control unit may
# support a single (e.g. a line printer) or multiple devices (e.g.multiple disk drives).
# In case of multiple devices, the most significant bits of the device address specify
# a particular device.
#
# The system being simulated assumes a single IO Processor, with address 0.
# This provides 8 subchannels, numbered 0..7 (0 has the highest priority)
# The cardreader, lineprinter and moving head disks are connected to a subchannnel.
#
#                                                            |IOP address
#                                                            V
# Card Reader uses subchannel 6                   => DA = 00 0 110
# Line Printer uses subchannel 7                  => DA = 00 0 111
# Moving head disk (X1215) uses subchannel 2      => DA = 00 0 010
#  This however is the address of the disk control unit !
#  The disk control unit is able to support multiple devices: Two disk units,
#  each with 2 spindels, a cartridge and an fixed drive, so 4 drives in total.
#  The addressing of a particular drive is as follows:
#                                                    DA = 00 0 010
#                                                         ^^
#                                                         ||Unit Number 0,1
#                                                         |Drive Number 0,1
#  Drive number 0 is the cartridge and drive number 1 is the fixed disk.
#  Because the 4 drives share a singe control unit, only 1 drive can be active
#  at a particular moment.
#
# Each subchannel uses 2 external registers to control the channel.
# Register 1 specifies the transfer (char or word, in or out, nbr of chars/words,..)
# Register 2 specifies the buffer address in memory.
# The external registers are to be loaded with WER instructions and addressed
# with the Control unit address (combination IOP address + subchannel).

# External registers used for the IOP subchannels:
my @EXTR;
# 8 sets of 2 external registers for IOP 0
$EXTR[0x0] = 0;
$EXTR[0x1] = 0;
$EXTR[0x2] = 0;
$EXTR[0x3] = 0;
$EXTR[0x4] = 0; # used by X1215 moving head disk CU (CU address = 2 => reg1 address = 2 * 2)
$EXTR[0x5] = 0; # used by X1215 moving head disk CU (CU address = 2 => reg2 address = 2 * 2 + 1)
$EXTR[0x6] = 0;
$EXTR[0x7] = 0;
$EXTR[0x8] = 0;
$EXTR[0x9] = 0;
$EXTR[0xA] = 0;
$EXTR[0xB] = 0;
$EXTR[0xC] = 0; # card reader (CU address = 6)
$EXTR[0xD] = 0; # card reader
$EXTR[0xE] = 0; # linprinter  (CU address = 7)
$EXTR[0xF] = 0; # linprinter

# When an IOP/CU needs to transfer a byte or word, it sends a break. Channels breaks
# are processed in priority order: channel 0 has the  highest priority. When a transfer
# is completed, the IOP/CU sends an interrupt. The simulator collects the break 
# flipflops of the subchannels in a break register

# IOP subchannel break register
#                          bit15
# brkreg 0000 0000 1111 1111
#                  ^       ^
#                  |       channel 0 break request (highest priority)
#                  channel 7

my $brkreg;             # break register
my $mpx1;               # multiplex word 1 external register
my $mpx2;               # multiplex word 2 external register
my $tx3;						# temporary for multiplex channel processing

# Channel break flipflops in breakregister
use constant BMD0  => 0x0004;   # MH disk CU    addr 0x2 => subchannel 2 => bit 13
use constant BLP   => 0x0080;   # Line printer: addr 0x7 => subchannel 7 => bit  8
use constant BCR   => 0x0040;   # Card reader:  addr 0x6 => subchannel 6 => bit  9 

######################### Emulator controller ##############################
my $cmd;     	# command string
my $file;
my $do = "";
my $log=0;    	# default, no logging
my $n;
my $m;
my $s;
my @w; 		   # itemized command string
@w = @ARGV;  	# possible initial do string

# open default log file
if($logfile ne "")
{
	if(! open (LOG, ">$logfile"))
	{
		prlog("Cannot open the new logfile: $logfile\n");
		$logfile = "";					
	}
	else
	{
		$log = 1;
		prlog("Set LOG ON and log in $logfile\n");
	}
}
prlog("P856 Emulator, version $version\n");

# attach a default file to the lineprinter
if($filelpt ne "")
{
	if(! open (LPT, ">$filelpt"))
	{
		prlog("Cannot open lineprinter file: $filelpt\n");
		$filelpt = ""
	}
	else
	{
		$LPT[FC] = \*LPT;
		prlog("File $filelpt attached to device LPT\n");
		$LPT[FILE] = $filelpt;		  # file attached and open
		$LPT[LCUR]  = 0;             # set on line 0 current page 	
	}
}

# attach a default diskpack to md0
if($filemd0 ne "")
{
	if(! open (MD0, "+<:raw", $filemd0)) # open for update
	{
		prlog("Cannot open disk file: $filemd0 for device MD0\n");
	}
	else
	{
		prlog("File $filemd0 attached to device MD0\n");
		$MD0[FILE] = $filemd0;           # file attached and open
		$MD0[CYL]  = 0;
	}
}

# attach a default diskpack to md1
if($filemd1 ne "")
{
	if(! open (MD1, "+<:raw", $filemd1)) # open for update
	{
		prlog("Cannot open disk file: $filemd1 for device MD1\n");
	}
	else
	{
		prlog("File $filemd1 attached to device MD1\n");
		$MD1[FILE] = $filemd1;           # file attached and open
	}
}

if(($MD0[FILE] ne "") && ($MD1[FILE] ne ""))
{
	$MD0[RDY] = 1;
	prlog("X1215 unit-0 is READY\n");
	prlog("X1215 unit-1 is NOT READY\n");
}
else
{
	prlog("X1215 unit-0 is NOT READY\n");
	prlog("X1215 unit-1 is NOT READY\n");
}

# clock on
$rtc_on=1;
cio_clk(1);
prlog("RTC is active\n");
initdecoder();
if(LINUX)
{
	# open a window for thr lineprinter
	prlog("VT used for the LPT window? " . $vt);
	$cmd = <STDIN>; chop $cmd;
	$n   = length($cmd);
	$cmd = substr($cmd,0,$n);
	if($cmd)
	{
		$lptwin =  $vt . $cmd;
		if(open(LPW,(">" . $lptwin)))
		{
			print LPW "\nLPT is using VT $lptwin\n";
		}
		else
		{
			$lptwin = "";
			prlog("Wrong/No VT specified for the LPT window\n");
    	}
	}
	else
	{
		$lptwin = "";
		prlog("No VT specified for the LPT window\n");
	}
}		
prlog("\n\n");
#####################################################################################


# do proc called at startup?
if($w[0] ne "")
{
	# yes
	$w[1] = $w[0];
	$w[0] = "do";
}

while(1)
{
	# read a command
	if($do)
	{
		$cmd = <DO>;
		if($cmd eq "")
		{
			close DO || die "can't close $do: $!";			
			$do = ""; # reset after eof
		}
		else
		{
			prlog ($cmd);
			chop $cmd;
			@w = split(/ /, $cmd);			
		}
	}
	
	# in case no command from a do string, then from keyboard
	if($w[0] eq "")
	{
		prlog ("> ");
		$cmd = <STDIN>;
		cmdlog($cmd);		
		chop $cmd;
		@w = split(/ /, $cmd);
	}

	if($w[0] eq "run")
	{
		if($program_started)
		{
			prlog("Proceed with \"continue\", \"step\" or \"reset\"\n");
		}
		else
		{
			if($w[1] ne "")
			{
				$n = cvhex(UP($w[1]));
				$n = $n & 0xFFFE;
				$R[0] = $n;                    # set PC
			}
			$interval = 0;
			$currentinterval = 0;
			$time = 0;
			$event_queue = 0;
			reset_all();
			if($rtc_on)
			{
				reset_clk();
				cio_clk(1);  # start rtc
			}
			# tty active to listen to CntrP (panel interrupt) or
			#                         CntrE (stop simulation)
			activate_event(\@TTY,$TTY[TIME]); 
			@t0 = gettimeofday ();             # fix t0 to measure elapsed time
			$RUN = 1;
			ReadMode 4;                        # prepare simulated console mode
			STDOUT->autoflush(1);              # for MSWin32
			while($RUN)
			{
				instr();
			}
			ReadMode 0;                        # reset console
			STDOUT->autoflush(0);              # for MSWin32
			$secs = tv_interval (\@t0);        # elapsed run time in secs
			$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
			if($secs > 0)
			{
				$cps = int($ctime / $secs);        # cycles per second
			}
			$s = prhex(16,$R[0]);
			$program_started = 1;              # program is still in a running state
			prlog("\nRun: emulation stopped at address $s\n");
			prlog("$ctime instructions in $secs seconds: $cps instructions per second\n");
		}
		mem_check(1);
	}
	elsif($w[0] eq "tfs")
	{
		$testflag = 1;
		prlog("Testflag set\n");
	}
	elsif($w[0] eq "tfr")
	{
		$testflag = 0;
		prlog("Testflag reset\n");		
	}
	elsif($w[0] eq "go")
	{
		if($program_started)
		{
			prlog("Proceed with \"continue\", \"step\" or \"reset\"\n");
		}
		else
		{
			if($w[1] ne "")
			{
				$n = cvhex(UP($w[1]));
				$n = $n & 0xFFFE;				
				$R[0] = $n;                    # set PC
			}
			$interval = 0;
			$currentinterval = 0;
			$time = 0;
			$event_queue = 0;
			if($rtc_on)
			{
				reset_clk();
				cio_clk(1);  # start rtc
			}			
			# tty active to listen to CntrP (panel interrupt) or
			#                         CntrE (stop simulation)
			activate_event(\@TTY,$TTY[TIME]); 
			@t0 = gettimeofday ();             # fix t0 to measure elapsed time
			$RUN = 1;
			ReadMode 4;                        # prepare simulated console mode
			STDOUT->autoflush(1);              # for MSWin32
			while($RUN)
			{
				instr();
			}
			ReadMode 0;                        # reset console
			STDOUT->autoflush(0);              # for MSWin32
			$secs = tv_interval (\@t0);        # elapsed run time in secs
			$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
			if($secs>0)
			{
				$cps = int($ctime / $secs);         # cycles per second
			}
			$s = prhex(16,$R[0]);
			$program_started = 1;              # program is still in a running state
			prlog("\nRun/Go: emulation stopped at address $s\n");
			prlog("$ctime instructions in $secs seconds: $cps instructions per second\n");
		}
		mem_check(1);
	}	
	elsif($cmd eq "continue" || $cmd eq "c" || $cmd eq "C")
	{
		if(! $program_started)
		{
			prlog("Continue only after a run, go or step\n");
		}
		else
		{
			$RUN = 1;
			ReadMode 4;                        # prepare simulated console mode
			STDOUT->autoflush(1);              # for MSWin32
			while($RUN)
			{
				instr();
			}
			ReadMode 0;                        # reset console
			STDOUT->autoflush(0);              # for MSWin32
			$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
			$s = prhex(16,$R[0]);
			prlog("\nContinue: emulation stopped at address $s, $ctime instructions processed\n");
		}
		mem_check(1);
	}
	elsif($w[0] eq "step" || $w[0] eq "s" || $w[0] eq "S")
	{
		prlog("step $w[1]\n");
		$n = 1;                              # default nbr of steps
		if($w[1] ne "")
		{
			if( ($n = cvdec(UP($w[1]))) <= 1)
			{
				$n=1;
			};
		}
		if(! $program_started)
		{
			$interval = 0;
			$currentinterval = 0;
			$time = 0;
			$event_queue = 0;
			reset_all();
			if($rtc_on)
			{
				reset_clk();
				cio_clk(1);  # start rtc
			}			
			# tty active to listen to CntrP (panel interrupt) or
			#                         CntrE (stop simulation)
			activate_event(\@TTY, $TTY[TIME]);
			$program_started = 1;             # program is in a running state
		}
		activate_event(\@STEP, $n);          # set stepper event
		$RUN = 1;
		ReadMode 4;                          # prepare simulated console mode	
		STDOUT->autoflush(1);                # for MSWin32			
		while($RUN)
		{
			instr();
		}
		ReadMode 0;                          # reset console
		STDOUT->autoflush(0);                # for MSWin32
		$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
		$s = prhex(16,$R[0]);
		prlog("Step: $ctime instruction(s) (remaining interval steps $interval), now at address $s\n");
		mem_check(1);
	}
	elsif($w[0] eq "do")
	{
		$do = $w[1];
		if(! open (DO, $do))
		{
			prlog("Unable to open file: $do\n");
			$do = "";
		}
	}
	elsif($w[0] eq "breakpoint" || $w[0] eq "bp")
	{
		if($w[1] eq "set" || $w[1] eq "s")
		{
			if(($n = cvhex(UP($w[2]))) >= 0)
			{
				$n = $n & 0xFFFE;
				$BP{$n} = $n;
			}
			else
			{
				prlog("Wrong address, no breakpoint set\n");
			}
		}
		elsif($w[1] eq "clear" || $w[1] eq "c")
		{
			if($w[2] eq "all" || $w[2] eq "a")
			{
				%BP = ();
			}
			else
			{
				$n = cvhex(UP($w[2]));
				$n = $n & 0xFFFE;
				delete $BP{$n};
			}
		}
		elsif($w[1] eq "print" || $w[1] eq "p")
		{
			prlog("Defined Breakpoints\n");
			foreach $n (sort keys( %BP))
			{
				prlog(prhex(16,$n) . "\n");
			}
		}
		else
		{
			prlog("????\n");
		}
		$n = scalar(keys(%BP));
		if($n)
		{
			$EmBreak = 1;
		}
		else
		{
			$EmBreak = 0;
		}
		prlog("$n Breakpoints defined (EmBreak=$EmBreak)\n");
	}
	elsif($w[0] eq "examine" || $w[0] eq "ex")
	{
		if(UP($w[1]) =~ /^M/)
		{
			$s = UP($w[2]);
			if(($m=cvhex($s)) >= 0)
			{
				$m=$m & 0xFFFE;
				while(1)
				{
					prlog(prhex(16, $m) . "  " . prhex(16,$M[$m]) . " ");
					$s = <STDIN>; $s=UP($s);
					$m = $m + 2;
					if(($s =~ /S/) || ($m > 0xFFFF))
					{
						last;
					}
				}
			}
			else
			{
				prlog("Wrong or no hex address\n");
			}
		}
		elsif(($n=$R16{UP($w[1])}) ne undef)
		{
			prlog("$w[1] = " . prhex(16,$R[$n]) . "\n");
		}
		else
		{
			prlog("Wrong memory or register reference $w[1]\n");
		}
	}
	elsif($w[0] eq "ldhex" || $w[0] eq "ldh")
	{
		my $s0;
		my $s1;
		if($w[1] ne undef)
		{
			if(UP($w[1]) eq "S")
			{
				# show content of file while loading
				if($w[2] =~ /.hex/)
				{
					if(! open (FHX, $w[2]))
					{
						prlog("Unable to open file: $w[2]\n");
					}
					else
					{
						while($s=<FHX>)
						{
							chop $s;
							($s0,$s1) = split(/\,/,$s); # $s0=addr, $s1=content
							$s0=cvhex($s0) & 0xFFFE; 
							$s1=cvhex($s1);
							prlog(prhex(16,$s0) . " " . prhex(16,$s1) . "  octal:" . proct($s0) . " " . proct($s1) . "\n");				
							$M[$s0] = $s1;
						}
					}
				}
				else
				{
					prlog("No proper filename is specified: syntax is: ldhex|ldh [s] <fn>.hex\n");
				}					
			}
			else
			{
				if($w[1] =~ /.hex/)
				{
					if(! open (FHX, $w[1]))
					{
						prlog("Unable to open file: $w[1]\n");
					}
					else
					{
						while($s=<FHX>)
						{
							chop $s;
							($s0,$s1) = split(/\,/,$s); # $s0=addr, $s1=content
							$s0=cvhex($s0) & 0xFFFE; 
							$s1=cvhex($s1);
							$M[$s0] = $s1;
						}
					}
				}
				else
				{
					prlog("No proper filename is specified: syntax is: ldhex|ldh [s] <fn>.hex\n");
				}
			}
		}
		else
		{
			prlog("No parameters specified: syntax is: ldhex|ldh [s] <fn>.hex\n");
		}
	}	
	elsif($w[0] eq "dump" || $w[0] eq "du")
	{
		my $str;
		my $c;
		$s = UP($w[1]);		
		if(($m=cvhex($s)) >= 0)
		{
			$m=$m & 0xFFE0;
			$s = UP($w[2]);	
			if(($n=cvhex($s)) >= 0)
			{
				if($w[3] eq "")
				{
					# dump on screen
					while($m < $n)
					{
						$str="";
						$s=0;
						prlog (prhex(16,$m) . " ");
						while($s<8)
						{	
							prlog (prhex(16,$M[$m]));
							$c=$ch{($M[$m] >> 8) & 0xFF};
							if($c ne undef)
							{	
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}
							$c=$ch{$M[$m] & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}						
							$m=$m+2;
							$s=$s+1;
						}
						prlog (" ");
						$s=0;
						while($s<8)
						{
							prlog (prhex(16,$M[$m]));
							$c=$ch{($M[$m] >> 8) & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}
							$c=$ch{$M[$m] & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}						
							$m=$m+2;
							$s=$s+1;
						}					
						prlog ("  $str\n");
					}
				}
				else
				{
					# dump on an absolute papertape
		            # print "$w[3]\n";					
					if($w[3] =~ /\.abs/)
					{
						open (BIN,">:raw", $w[3]) || die "cannot open outputfile: $w[3]: $!";
						prlog ("\nPunch in absolute 8+8 papertape load format: $w[3] \n\n");
						my $byte;
						my $w;
						my $chksum=0;
                        my $low=$m;
						my $high=$n+2;
						my $i;
						
						punch($rub);                         # rubout
						$w = ($high >> 1) - ($low >> 1);     # word count (high = endaddr + 2)
						$byte = $w >> 8;
						punch($byte);
						$byte = $w & 0xFF;
						punch($byte);

						$w = $low-2;	                     # begin address - 2
						$byte = $w >> 8;
						punch($byte);
						$byte = $w & 0xFF;
						punch($byte);

						$i = $low;                           # begin address 

						while($i < $high)                    # content
						{
							$w = $M[$i];
							$chksum = $chksum ^ $w;
							$byte = $w >> 8;
							punch($byte);
							$byte = $w & 0xFF;
							punch($byte);
							$i = $i + 2;
						}

						$w=$chksum;                          # checksum
						$byte = $w >> 8;
						punch($byte);
						$byte = $w & 0xFF;
						punch($byte);
						punch($xoff);                        # xoff	
						prlog ("Papertape ready\n");
						close BIN || die "can't close $w[3]: $!";
					}
					else
					{
						prlog ("Wrong filename: no .abs extension\n");
					}
				}
			}
		}					
	}
	elsif($w[0] eq "deposit" || $w[0] eq "de")
	{
		if(UP($w[1]) =~ /^M/)
		{
			$s = UP($w[2]);                   # address
			if(($s=cvhex($s)) >= 0)
			{
				$s=$s & 0xFFFE;
				$m = UP($w[3]);               # content
				if(($m=cvhex($m)) >= 0)
				{
					$M[$s] = $m;
				}
				else
				{
					# read content from keyboard
					while(1)
					{
						prlog(prhex(16, $s) . "  " . prhex(16,$M[$s]) . " ");
						$m = <STDIN>; $m=UP($m);
						if($m =~ /S/)
						{
							last;                 # stop
						}
						elsif($m =~ /[\da-fA-F]/)
						{
							if(($m=cvhex($m)) >= 0)
							{
								$M[$s] = $m;
							}
							else
							{
								prlog("Wrong content\n");
								$s = $s - 2;       # redo current address
							}
						}
						$s = $s + 2;
					}
				}
			}
			else
			{
				prlog("Wrong or no hex address\n");
			}
		}
		elsif(($n=$R16{UP($w[1])}) ne undef)
		{
			if($w[2] eq undef)
			{
				prlog("$w[1] = ");
				$s = <STDIN>;						# input 16 bit hex number
				chomp $s;
				$s = UP($s);
				if(($m=cvhex($s)) >= 0)
				{
					$R[$n] = $m;
				}
				else
				{
					prlog("Error input hex number\n");
				}
			}
			else
			{
				$s = UP($w[2]);               # content
				if(($m=cvhex($s)) >= 0)
				{
					$R[$n] = $m;
				}
				else
				{
					prlog("Error input hex number\n");
				}
			}
		}
		else
		{
			prlog("Wrong register $w[1]\n");
		}
	}
	elsif($w[0] eq "reset")
	{
		if($w[1] eq "cpu")
		{
			prlog("CPU reset\n");
			reset_cpu();
		}
		elsif($w[1] eq "tty")
		{
			prlog("TTY reset\n");
			reset_tty();
		}
		elsif($w[1] eq "rtc")
		{
			prlog("Realtime clock reset\n");
			reset_clk();
		}
		elsif($w[1] eq "ptr")
		{
			prlog("Papertape reader reset\n");
			reset_ptr();
		}	
		elsif($w[1] eq "ptp")
		{
			prlog("Papertape punch reset\n");
			reset_ptp();
		}
		elsif($w[1] eq "lpt")
		{
			prlog("Line printer reset\n");
			reset_lpt();
		}	
		elsif($w[1] eq "crd")
		{
			prlog("Card reader reset\n");
			reset_crd();
		}
		elsif($w[1] eq "md0" || $w[1] eq "md1"|| $w[1] eq "md2"|| $w[1] eq "md3")
		{
			prlog("Moving head disk reset\n");
			reset_md0();
		}		
		elsif($w[1] eq "mem")
		{
			prlog("Memory reset\n");
			@M=();
		}
		elsif($w[1] eq "power")
		{
		    reset_all();
			@M=();
			$R[0]=0;
			prlog("Power reset\n"); 
		}
		else
		{
			prlog("Reset All, except A0/P and Memory\n"); 
			reset_all();
		}
	}
    elsif($w[0] eq "trace" || $w[0] eq "tr")
	{
		$latrace = 0; # reset rtrace
		$batrace = 0;
		$eatrace = 0;
		if($w[1] eq "on")
		{
			$EmTrace = 1;
			if($w[2] eq "")
			{
				# no tracefile specified
				prlog("Set trace $w[1]\n");
			}
			else
			{
				# tracefile specified
				if($tracefile ne "")
				{
					close TRACE || die "can't close $tracefile: $!";
					prlog("Previous tracefile $tracefile is closed\n");
				}
				$tracefile = $w[2];
				if(! open (TRACE, ">$tracefile"))
				{
					$EmTrace = 0;
					prlog("Cannot open the new tracefile: $tracefile (trace off)\n");
				}
				else
				{
					prlog("set trace $w[1], trace dumped in $tracefile\n");
				}					
			}
		}
		elsif($w[1] eq "off")
		{
			$EmTrace = 0;
			prlog("Set trace $w[1]\n");
		}
		else
		{
			$EmTrace = 1;
			prlog( "Set trace on\n");
		}		
	}
	elsif($w[0] eq "rtrace" || $w[0] eq "rt")
	{
		$EmTrace = 0;
		$latrace = 0;
		$batrace = 0;
		$eatrace = 0;
		if($w[1] eq "on")
		{
		    $s = UP($w[2]);		
		    if(($latrace=(cvhex($s) & 0xFFFE)) >= 0)
		    {
				if($w[3] ne "")
				{
					$s = UP($w[3]);
					$m=cvhex($s) & 0xFFFE;
				}
				else
				{
					$m=0;
				}
				$batrace = ($m+$latrace) & 0xFFFE;
				if($w[4] ne "")
				{
					$s = UP($w[4]);	
					$n=cvhex($s) & 0xFFFE;
					$eatrace = ($n+$latrace) & 0xFFFE;
				}
				else
				{
					$eatrace = 0xFFFE;
				}
				
				if($batrace<=$eatrace && ($batrace>=0))
				{
					$EmTrace = 1;
					prlog("Set rtrace on: la=" . prhex(16,$latrace) . " ba=" . prhex(16,$batrace) . " ea=" . prhex(16,$eatrace) . "\n");
				}
				else
				{
					prlog("Wrong <la>/<ba>/<ea>, rtrace off\n");
					$latrace = 0;
					$batrace = 0;
					$eatrace = 0;
				}
			}
			else
			{
				prlog("Wrong <la>/<ba>/<ea>, rtrace off\n");
				$latrace = 0;
				$batrace = 0;
				$eatrace = 0;
			}
		}
		elsif($w[1] eq "off")
		{
			prlog("Set rtrace $w[1]\n");
		}
		else
		{
			prlog("Wrong rt command, rtrace off\n");
		}
	}
	elsif($w[0] eq "log")
	{
		if($w[1] ne "")
		{
			if($w[1] eq "on" || $w[1] eq "off")
			{
				if($w[1] eq "on")
				{
					if($logfile ne "")
					{
						$log=1;
					    prlog("Set LOG ON, LOG in existing $logfile\n");						
					}
					else
					{
						prlog("No logfile specified\n");
					}
				}
				if($w[1] eq "off")
				{
					$log=0;
				}
			}
			else
			{
				if($logfile ne "")
				{
					close LOG || die "can't close $logfile: $!";
					prlog("Previous logfile $logfile is closed\n");	
				}
				$logfile = $w[1];
				if(! open (LOG, ">$logfile"))
				{
					prlog("Cannot open the new logfile: $logfile\n");
					$logfile = "";					
				}
				else
				{
					$log = 1;
					prlog("Set LOG ON, LOG dumped in $logfile\n");
				}
			}																			
		}
		else
		{
			prlog("Invalid parameter; use: log [on|off] | <file>\n");
		}
	}	
	elsif($w[0] eq "rtc")
	{
		my $stat;
		if($w[1] eq "on")
		{
			$rtc_on=1;
			$stat = $CLK[STATE] >> 16;
			if($stat == INAC)
			{
				cio_clk(1);
			}
		}
		elsif($w[1] eq "off")
		{
			$rtc_on=0;
			cio_clk(0); # stop clock			
		}

		state_clk();
		$stat = $CLK[STATE] >> 16;
		if($stat == INAC)
		{
			prlog("RTC is stopped\n");
		}
		elsif($stat == (EXE | WREQ))
		{
			prlog("RTC is requested to be stopped\n");
		}
		else
		{
			prlog("RTC is active\n");				
		}		
	}
	elsif($w[0] eq "load")
	{
		my $leader = "false";
		my $c;
		my $w;
		my $err=0;
		my $ba=0;     # begin address of load area
		my $wc=0;     # word count
		my $l;
		my $mode = ""; 

		$file = $w[1];
		if ($file eq "" || (( $file !~ /\.abs/ ) && ( $file !~ /\.rel/ ) && ( $file !~ /\.LM/ ) && ( $file !~ /\.lm/ )))
		{
			prlog("No filename, or wrong extension (must be .abs or .rel/.LM/.lm)\n");
			$err=1;
		}	
				
		if($w[1] =~ /\.abs/)
		{
			$mode = 'a';
		}
		
		if($w[1] =~ /\.rel/ || $w[1] =~ /\.LM/ || $w[1] =~ /\.lm/)
		{
			if($w[2] ne "" && $w[2] =~ /^[0-9a-f]/)
			{
				# load a relocatable load module
				$ba = cvhex(UP($w[2]));
				$ba = $ba & 0xFFFE;	
				$mode = 'r';
			}
			elsif($w[2] ne "" && $w[2] eq "mon")
			{
				# load a relocatable monitor load module at address 0
				$ba = -8;
				$mode = 'm';				
			}			
			else
			{
				prlog("Load address missing\n");
				$err = 1;
			}			
		}

		if($err == 0)
		{
			if(! open (IN, $file))
			{
				prlog( "Unable to open requested load file: $file\n");
				$err = 1;
			}
		}
		
		if($err == 0 && $mode eq 'a')
		{
			binmode IN, ":raw";			
			prlog("Load $w[1]\n");
			# Read bootable 8+8 formatted paper tape
			# skip leader
			while( 1 )
			{
				$c = nextchar();
				# skip leader in case there
				if ($leader eq "false")
				{
					next if $c eq $null;
					$leader = "true";
					prlog("Leader skipped\n");
					last;
				}
			}
			$wc = nextword();                       # word count
			$ba = (nextword() + 2) & 0xFFFE;        # begin address of load area
			my $low = $ba;
			my $high = $ba + $wc*2;
			my $chksum = 0;
			prlog("Load " . prhex(16,$wc) . " words\; start loading at" . prhex(16,$ba) . "\n");
			$l = 0;
			while($l < $wc)
			{
				$w = nextword();
				$chksum = $chksum ^ $w;
				$M[$ba] = $w;
				$ba = $ba + 2;
				$l = $l + 1;
			}	
			my $chk = nextword();	
			if( ($chksum - $chk) != 0)
			{
				prlog("Checksum error\n");
			}
			$c =nextchar();
			if($c != $xoff)
			{
				prlog("No x-off marking end of tape\n");
			}
			close (IN) || die "can't close $file: $!";
			prlog("LOADING COMPLETE\; 1st free address is" . prhex(16,$high) . "\n");
			$R[0] = $low;		# default start address
			prlog("Default start address (P) is:" . prhex(16,$low) . "\n");
		}
		
		if($err == 0 && $mode eq 'r')
		{		
			# load relocatable load module
			binmode IN, ":raw";			
			prlog("Load $w[1]\n");
			nextword(); # header word 1
			nextword(); # header word 2
            my $sa = nextword();    # start address
			my $ns = nextword();    	# number of sectors
			$wc    = nextword();    	# length in words, counted in bytes
			$wc    = $wc >> 1;      	# length in words
			nextword();             	# sym table pointer for debugging: not used at this moment
			prlog("Load " . prhex(16,$wc) . " words\; start loading at" . prhex(16,$ba) . "\n");
            # $ba is the relocator
		    $M[$ba]   = $sa;
			$M[$ba+2] = $ns;
			$M[$ba+4] = $wc << 1;
			$M[$ba+6] = 0;          	# sym table pointer for debugging: not used at this moment
				
			my $k = 0;              	# total number of words relocated
			$l=4;								# total number of words loaded
			
            my $i=4;						# number of code/data words retrieved from record
			while($l<$wc)
			{
				if($i == 0)
				{
					# skip header
			        nextword(); # header word 1
			        nextword(); # header word 2
                }					
				# load record
				while(($i<188) && ($l<$wc))
				{
					$w=nextword();
					$M[$ba+($l*2)] = $w;
					$i=$i+1;
					$l=$l+1;
				}
				# skip to relocation table, if not there yet
				while($i<188)
				{
					nextword();
					$i=$i+1;
				}
				# relocate this record
				$i=0;
				while($i<12)    			# scan the relocation table
				{
					my $rw = nextword();
					my $j = 0;
					while(($j<16) && ($k<$l))
					{
						if($rw & 0x8000)
						{
							$M[$ba+($k*2)] = ($M[$ba+($k*2)] + $ba) & 0xFFFF ; # relocate
						}
						$rw = $rw << 1;
						$j=$j+1;
						$k=$k+1;
					}
					$i=$i+1;
				}
				# skip to end of record
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				$i=0;
			}
			close (IN) || die "can't close $file: $!";
			$i=$ba+($l*2);      # first free address 
			prlog("LOADING COMPLETE\; 1st free address is" . prhex(16,$i) . "\n");
			$R[0] = $M[$ba];		# start address
			prlog("Start address (P) is: " . prhex(16,$R[0]) . "\n");			
		}
		if($err == 0 && $mode eq 'm')
		{		
			# load relocatable monitor load module at address 0
			binmode IN, ":raw";			
			prlog("Load monitor $w[1]\n");
			nextword(); # header word 1
			nextword(); # header word 2
            my $sa = nextword()-8;  # start address (relocated)
			my $ns = nextword();    # number of sectors
			$wc    = nextword();    # length in words, counted in bytes
			$wc    = $wc >> 1;      # length in words
			nextword();             # sym table pointer for debugging: not used at this moment
			prlog("Load " . prhex(16,$wc) . " -4 words\; start loading at" . prhex(16,0) . "\n");
            # $ba is the relocator (-8 !)
			my $sec0=1; # sector 0 indicator
			my $low=0;
			my $high=0;	
			my $k = 4;              # total number of words relocated
			$l=4;							# total number of words loaded
            my $i=4;					# number of code/data words retrieved from record
			while($l<$wc)
			{
				if($i == 0)
				{
					# skip header
			        nextword(); # header word 1
			        nextword(); # header word 2
                }					
				# load record
				while(($i<188) && ($l<$wc))
				{
					$w=nextword();
					$M[$ba+($l*2)] = $w;
					$i=$i+1;
					$l=$l+1;
				}
				# skip to relocation table, if not there yet
				while($i<188)
				{
					nextword();
					$i=$i+1;
				}
				# relocate this record
				$i=0;
				while($i<12)    		# scan the relocation table
				{
					my $rw = nextword();
					my $j = 0;
					if($sec0)
					{
						# skip 1st 4 words of sector 0
						$rw = $rw << 4;
						$j  = 4;
						$sec0=0;
					}					
					while(($j<16) && ($k<$l))
					{
						if($rw & 0x8000)
						{
							$M[$ba+($k*2)] = ($M[$ba+($k*2)] + $ba) & 0xFFFF ; # relocate
						}
						$rw = $rw << 1;
						$j=$j+1;
						$k=$k+1;
					}
					$i=$i+1;
				}
				# skip to end of record
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				$i=0;
			}
			close (IN) || die "can't close $file: $!";
			$i=$ba+($l*2);      # first free address 
			prlog("LOADING COMPLETE\; 1st free address is" . prhex(16,$i) . "\n");
			$high=$ba+($l*2)-2;
			$i = "Used memory area is: low=" . prhex(16,$low) . " high=" . prhex(16,$high);
			prlog("$i\n");					
			$R[0] = $sa;		# start address
			prlog("Start address (P) is:" . prhex(16,$R[0]) . "\n");			
		}		
	}
	elsif($w[0] eq "boot")
	{
		if($w[1] eq "ptr")
		{
			if($PTR[FILE] eq "")
			{
				prlog("No file attached to paper tape reader\n");
			}
			else
			{
				reset_all();
				@M=();
				# put papertape 8+8 bootstrap in memory
				#                 * CODE   EQU   *+/2E
				#                   PR     EQU   /20
				#                   S      EQU   1
				$M[0]  = 0x20BF;  #        INH
				$M[2]  = 0x41E0;  #        CIO   A1,S,PR
				$M[4]  = 0x4A20;  # SEARCH INR   A2,0,PR
				$M[6]  = 0x5C04;  #        RB(4) *-2
				$M[8]  = 0x227F;  #        ANK   A2,/7F
				$M[10] = 0x5808;  #        RB(0) SEARCH              skip leader
				$M[12] = 0x0704;  #        LDK   A7,4                read 4 bytes:
				$M[14] = 0x0540;  #        LDK   A5,/40              word count, and
				#                   INPUT  EQU   *
				$M[16] = 0x4A20;  #        INR   A2,0,PR             begin address - 2
				$M[18] = 0x5C04;  #        RB(4) *-2
				$M[20] = 0xE235;  #        SCR   A2,A5
				$M[22] = 0x1501;  #        ADK   A5,1
				$M[24] = 0x1F01;  #        SUK   A7,1
				$M[26] = 0x590C;  #        RB(1) INPUT               > 0
				$M[28] = 0x5210;  #        RF(2) CODE                < 0
				$M[30] = 0xBAC0;  #        ML    5,/3A               = 0
				$M[32] = 0x003A;  #
				$M[34] = 0x1502;  #        ADK   A5,2
				$M[36] = 0x9508;  #        ADR   A5,A2
				$M[38] = 0x3C41;  #        SLL   A4,1
				$M[40] = 0x8614;  #        LDR   A6,A5               A6 default start address
				$M[42] = 0x0100;  #        LDK   A1,0
				$M[44] = 0x5F1E;  #        RB    INPUT
				#                   CODE   EQU   *
				$M[46] = 0xE104;  #        ECR   A1,A1
				$M[48] = 0xB108;  #        XRR   A1,A2
				$M[50] = 0x1C01;  #        SUK   A4,1
				$M[52] = 0x5C26;  #        RB(4) INPUT               ne 0
				$M[54] = 0x8F18;  #        ABR   A6                  jump to start address (or HLT)           
				$M[56] = 0x0000;  #        DATA  0 UNUSED
				$M[58] = 0x0000;  # (3A)   DATA  0 UNUSED            A1
				$M[60] = 0x0000;  #        DATA  0 BASE ADDRESS      A2
				$M[62] = 0x0000;  #        DATA  0 UNUSED            A3
				########################################
				$M[64] = 0x0000;  # (40)   DATA  0 WORD COUNT        A4
				$M[66] = 0x0000;  #        DATA  0 BEGIN ADDRESS -2  A5
				#
				if($w[2] eq "h")
				{
					$M[54] = 0x207F;               # HALT, after loading complete
				}
				if($w[2] eq "r")
				{
					$M[54] = 0x8F18;               # ABR A6 => RUN starting at begin 1st address
				}
				$R[0] = 0;                        # PC = 0, start at address 0
				$interval = 0;
				$currentinterval = 0;
				$time = 0;
				$event_queue = 0;
				if($rtc_on)
				{
					reset_clk();
					cio_clk(1);  # start rtc
				}				
				activate_event(\@TTY,$TTY[TIME]); 
				@t0 = gettimeofday ();             # fix t0 to measure elapsed time
				$RUN = 1;
				ReadMode 4;                        # prepare simulated console mode
			    STDOUT->autoflush(1);             # for MSWin32
				while($RUN)
				{
					instr();
				}
				ReadMode 0;                        # reset console
			    STDOUT->autoflush(0);             # for MSWin32
				$secs = tv_interval (\@t0);        # elapsed run time in secs
				$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
				$cps = int($ctime / $secs);        # cycles per second
				$s = prhex(16,$R[0]);
				$program_started = 1;              # program is still in a running state
				prlog("\nRun/Boot: emulation stopped at address $s\n");
				prlog("$ctime instructions in $secs seconds: $cps instructions per second\n");				
			}
		}
		elsif($w[1] eq "md0" || $w[1] eq "md1" || $w[1] eq "md2" || $w[1] eq "md3")
		{
			if(($w[1] eq "md0" && $MD0[FILE] eq "") || ($w[1] eq "md1" && $MD1[FILE] eq "") ||
			   ($w[1] eq "md2" && $MD2[FILE] eq "") || ($w[1] eq "md3" && $MD3[FILE] eq ""))
			{
				prlog("No disk image file attached to disk drive\n");
			}
			else
			{
				reset_all();
				@M=(); # reset memory
                # X1215 BOOT LOADS ONE RECORD INTO LOCATION /80 ... THEN STARTS AT /84
				# THE RECORD IS THE SECTOR 1 ON DISK (=physical sector 3) , OR 254 CHARACTERS FROM THE 
				# INPUT DEVICE (LEADING NULL CHARACTERS ARE IGNORED). SECTOR 0 IN CASE THE FLOPPY IS USED.
                #
				# At boot start, A15 must contain the boot parameters:
                # BIT 0    = 0 : CHARACTER EXCHANGE ON PROGRAMMED CHANNEL
                #            1 : WORD         "      "     "        =
                # 
                #     1    = 0 : IPL FROM A NON-DISK DEVICE
                #            1 : IPL FROM DISK
                #
                #     2    = 0 : FIXED HEADS DISK
                #            1 : MOVING HEADS DISK
                #
                #     3    = 0 : DEVICE ON I/O PROCESSOR (MX)
                #            1 : DEVICE ON PROGRAMMED CHANNEL (CP)
                #
                #     4 TO 7   =  BOU LINES, USED TO QUALIFY I/O INSTRUCTION (CIO START NON-DISK DEVICE)
                #
                #     8    = 0 : SINGLE DEVICE CONTROLLER
                #            1 : MULTI DEVICE CONTROLLER
                #
                #     9    = 0 : NON X1215 DISK IS USED
                #            1 : X1215 DISK IS USED
                #
                #     10 TO 15 = DEVICE ADDRESS
                #
                #        WARNING: NOTE THAT THE BITS 0 AND 3 ARE COMPLEMENTARY				
				#
				# A15 = /63C2 boots from the X1215 cartridge  (drive 0 unit 0)
				# A15 = /63E2 boots from the X1215 fixed disk (drive 1 unit 0)
				# A15 = /63D2 boots from the X1215 cartridge  (drive 0 unit 1)
				# A15 = /63F2 boots from the X1215 fixed disk (drive 1 unit 1)				
				#
				# Example A15 boot parameter setting: boot from X1215 cartridge
				#    0110 0011 1100 0010
				#     ||  ---- |||| |--- subchannel 2 
				#     ||     | |||| |IOP 0
				#     ||     | ||||disk unit 0
				#     ||     | |||drive 0
				#     ||     | ||boot from an X1215
				#     ||     | |a multi device controller is used
				#     ||     |load physical sector 3 (logical 1) which is the sector with IPL
				#     ||moving heads disk
				#     |ipl from disk
				#
				if(  $w[1] eq "md0")
				{
					$R[A15] = 0x63C2; # boot from unit 0 drive 0 (address 02)
				}
				elsif($w[1] eq "md1")
				{
					$R[A15] = 0x63E2; # boot from unit 0 drive 1 (address 22)
				}
				elsif($w[1] eq "md2")
				{
					$R[A15] = 0x63D2; # boot from unit 0 drive 1 (address 12)
				}				
				elsif($w[1] eq "md3")
				{
					$R[A15] = 0x63F2; # boot from unit 0 drive 1 (address 32)
				}				
                #				
                                    # * BASE REGISTER INITIALIZATION
				$M[0x00] = 0x0254;  # BOOT   LDK         A2,INR               A2 = ADDRESS INR INSTRUCTION
				$M[0x02] = 0x0336;  #        LDK         A3,CIO               A3 = ADDRESS CIO INSTRUCTION
				$M[0x04] = 0x0472;  #        LDK         A4,SST               A4 = ADDRESS SST INSTRUCTION
                                    # * SET DEVICE ADDRESS TO INITIALIZE THE I/O COMMANDS
				$M[0x06] = 0x861E;  #        LDR         A6,A15               A6 = VALUE OF THE KEYS
				$M[0x08] = 0x263F;  #        ANK         A6,/3F               KEEP JUST THE DEVICE ADDRESS
				$M[0x0A] = 0xAE29;  #        ORRS        A6,A2                INIT INR WITH DEVICE ADDRESS
				$M[0x0C] = 0xAE2D;  #        ORRS        A6,A3                INIT CIO WITH DEVICE ADDRESS
				$M[0x0E] = 0xAE41;  #        ORS         A6,HIO               INIT THE HALT IO WITH DEVICE ADDRESS
				$M[0x10] = 0x0070;  
                                    # * SET CONTROLLER TYPE AND ADDRESS
				$M[0x12] = 0x871E;  #        LDR         A7,A15               MULTI OR SINGLE DEVICE CU?
				$M[0x14] = 0x3FC8;  #        SLC         A7,8
				$M[0x16] = 0x5602;  #        RF(6)       INIT20               JUMP IF SINGLE
				$M[0x18] = 0x260F;  #        ANK         A6,/F                MULTI: KEEP CONTROLLER ADDRESS (CA)
                                    # * INITIALIZE THE WER INSTRUCTIONS AND SST ONE
				$M[0x1A] = 0xAE31;  # INIT20 ORRS        A6,A4                INIT SST WITH DEVICE OR CONTROLLER ADDR
				$M[0x1C] = 0x3E41;  #        SLL         A6,1                 COMPUTE EXTERNAL REGISTER ADDRESS
				$M[0x1E] = 0xAE41;  #        ORS         A6,WER1              UPDATE WER INSTRUCTIONS
				$M[0x20] = 0x0042;  
				$M[0x22] = 0xAE41;  #        ORS         A6,WER2
				$M[0x24] = 0x0044;  
				$M[0x26] = 0x811C;  #        LDR         A1,A7                A1 = BOU LINES
				$M[0x28] = 0x0550;  #        LDK         A5,/50               BLOCK LENGTH TO BE LOADED
				$M[0x2A] = 0x0680;  #        LDK         A6,/80               IN LOCATION /80.....
                                    # * CHECK DEVICE TYPE
				$M[0x2C] = 0x3FE7;  #        SRC         A7,7                 IS THE DEVICE ADISK OR NOT ?
				$M[0x2E] = 0x5612;  #        RF(6)       WER1                 NO
				$M[0x30] = 0x3FC1;  #        SLC         A7,1                 IS IT A FIXED HEADS DISK ?
				$M[0x32] = 0x5604;  #        RF(6)       NOSEEK               YES => NO SEEK TO ZERO
				$M[0x34] = 0x0103;  #        LDK         A1,3                 EXECUTE A SEEK TO ZERO OPERATION
				$M[0x36] = 0x41C0;  # CIO    CIO         A1,1,0               ON THE MOVING HEADS DISK
                                    # * PREPARE THE NEXT LOGICAL IO FOR A DISK
				$M[0x38] = 0x811E;  # NOSEEK LDR         A1,A15               COMPUTE THE 
				$M[0x3A] = 0x3966;  #        SRL         A1,6                 SECTOR
				$M[0x3C] = 0x213C;  #        ANK         A1,/3C               NUMBER
				$M[0x3E] = 0x8520;  #        LDKL        A5,/80CD             1ST WORD FOR MX FOR A DISK
				$M[0x40] = 0x80CD; 
                                    # * EXECUTE THE LOGICAL IO OPERATION FOR ANY DEVICE, ON MX OR CP
				$M[0x42] = 0x7500;  # WER1   WER         A5,0                 FOR ANY DEVICE (DISK OR NOT) EXECUTE THE WER,
				$M[0x44] = 0x7601;  # WER2   WER         A6,1                 THE CHANNEL IS
				$M[0x46] = 0xF031;  #        EXR*        A4                   EXECUTE SST A7,..
				$M[0x48] = 0xF02D;  #        EXR*        A3                   EXECUTE CIO A1,..
				$M[0x4A] = 0x5C06;  #        RB(4)       *-4                  REFUSED !  TRY AGAIN
                                    # * CHECK WHETHER THE CHANNEL WANTED TO TAKE THE ACTION
				$M[0x4C] = 0x8194;  #        LDR         A9,A5                A9 = NBR OF CHARS TO READ
				$M[0x4E] = 0x871E;  # MX::CP LDR         A7,A15               IS THE DEVICE CONNECTED TO
				$M[0x50] = 0x3F43;  #        SLL         A7,3                 THE MULTIPLEX CHANNEL ?
				$M[0x52] = 0x561E;  #        RF(6)       SST                  YES, SEND AN SST
				                    # * NO, CONNECTED TO CP => INPUT ONE CHARACTE/WORD FROM THE DEVICE
				$M[0x54] = 0x4F00;  # INR    INR         A7,0,0               READ INTO A7
				$M[0x56] = 0x541A;  #        RF(4)       SST                  REFUSED !, SEND AN SST TO KNOW WHY
				$M[0x58] = 0xE994;  #        CWR         A9,A5                LEADING CHAR ?
				$M[0x5A] = 0x5404;  #        RF(4)       INR10                NO
				$M[0x5C] = 0x871C;  #        LDR         A7,A7                YES
				$M[0x5E] = 0x580C;  #        RB(0)       INR
				                    # * CHECK WHETHER WORD OR CHARACTER MODE
				$M[0x60] = 0x879E;  # INR10  LDR         A15,A15              CHARACTER MODE ?
				$M[0x62] = 0x5604;  #        RF(6)       STORE                YES
				$M[0x64] = 0x8739;  #        STR         A7,A6                STORE WORD JUST READ FROM CP
				$M[0x66] = 0x1601;  #        ADK         A6,1                 USE SAME ROUTINE FOR WORD AND CHAR MODE
                                    # * STORE CHAR JUST READ AND CHECK FOR END OF TRANSFER
				$M[0x68] = 0xE739;  # STORE  SCR         A7,A6                STORE ONE CHAR TO /80 + (A6)
				$M[0x6A] = 0x1601;  #        ADK         A6,1                 PREPARE FOR NEXT LOCATION TO STORE
				$M[0x6C] = 0x1D01;  #        SUK         A5,1                 OTHER CHAR TO READ ?
				$M[0x6E] = 0x5C1C;  #        RB(4)       INR                  YES, LOOP TO INPUT
                                    # * END OF TRANSFER, EXECUTE A CIO HALT
				$M[0x70] = 0x4180;  # HIO    CIO         A1,0,0               TO EXCHANGE
                                    # * EXECUTE AN SST AND ANALYZE THE STATUS 
				$M[0x72] = 0x4FC0;  # SST    SST         A7,0                 
				$M[0x74] = 0x5C28;  #        RB(4)       MX::CP               REFUSED ! => THIS HANDLES BOTH CASES: WAIT ON CP OR MX
				$M[0x76] = 0xA720;  #        ANKL        A7,/4007             
				$M[0x78] = 0x4007; 
                                    # * TEST: BIT 1  = DEVICE BECAME READY
                                    # *       BIT 13 = DATA FAULT
                                    # *       BIT 14 = THROUGHPUT ERROR
                                    # *       BIT 15 = NOT OPERABLE
				$M[0x7A] = 0x0C00;  #        AB(4)       /00                  ONE OF THESE CONDITIONS => RESTART THE BOOTSTRAP
				$M[0x7C] = 0x0F84;  #        AB          /84                  NORMAL END => JUMP TO IPL
                                    #        END         BOOT
				if($w[2] eq "h")
				{
					$M[0x7C] = 0x207F;             # HALT, after loading complete
					prlog("Boot stops after loading IPL\n");
				}
				if($w[2] eq "r")
				{
					$M[0x7C] = 0x0F84;             # AB /84 => RUN starting at begin 1st address of IPL
					prlog("Boot starts IPL after loading IPL\n");			
				}
				$R[0] = 0;                        # PC = 0, start at address 0
				$interval = 0;
				$currentinterval = 0;
				$time = 0;
				$event_queue = 0;
				if($rtc_on)
				{
					reset_clk();
					cio_clk(1);  # start rtc
				}				
				activate_event(\@TTY,$TTY[TIME]); 
				@t0 = gettimeofday ();             # fix t0 to measure elapsed time
				$RUN = 1;
				ReadMode 4;                        # prepare simulated console mode
			    STDOUT->autoflush(1);             # for MSWin32
				while($RUN)
				{
					instr();
				}
				ReadMode 0;                        # reset console
			    STDOUT->autoflush(0);             # for MSWin32
				$secs = tv_interval (\@t0);        # elapsed run time in secs
				$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
				$cps = int($ctime / $secs);        # cycles per second
				$s = prhex(16,$R[0]);
				$program_started = 1;              # program is still in a running state
				prlog("\nRun/Boot: emulation stopped at address $s\n");
				prlog("$ctime instructions in $secs seconds: $cps instructions per second\n");
			}
			
		}		
		else
		{
			prlog( "Unsupported boot device $w[1]\n");
		}
	}
	elsif($w[0] eq "attach" || $w[0] eq "att")
	{
		if($w[1] eq "ptr")
		{
			$file = $w[2];
			if($file eq "")
			{
				prlog("No proper filename\n");
			}
			else
			{
				if($PTR[FILE] ne "")
				{
					close ($PTR[FC]) || die "can't close $PTR[FILE]: $!";
					prlog("File $PTR[FILE] detached from device PTR and closed\n");
					$PTR[FILE] = "";         		# file detached and closed; device inoperable	
				}
				if(! open (PTR, $file))
				{
					prlog("Cannot open papertape input file: $file\n");
				}
				else
				{
					$PTR[FC] = \*PTR;
					binmode $PTR[FC], ":raw";
					prlog("File $file attached to device PTR\n");
					$PTR[FILE] = $file;           # file attached and open
				}
			}
		}
		elsif($w[1] eq "crd")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			if($file eq "" || $fy ne "crd")
			{
				prlog("No filename, or wrong extension (must be a .crd file)\n");
			}
			else
			{
				if($CRD[FILE] ne "")
				{
					close ($CRD[FC]) || die "can't close $CRD[FILE]: $!";
					prlog("File $CRD[FILE] detached from device CRD and closed\n");
					$CRD[FILE] = "";         		# file detached and closed; device inoperable	
				}
				if(! open (CRD, $file))
				{
					prlog("Cannot open card reader input file: $file\n");
				}
				else
				{
					$CRD[FC] = \*CRD;
					binmode $CRD[FC], ":raw";
					prlog("File $file attached to device CRD\n");
					$CRD[FILE] = $file;           # file attached and open
				}
			}
		}		
		elsif($w[1] eq "ptp")
		{
			$file = $w[2];
			if($file eq "")
			{
				prlog("No proper filename\n");
			}
			else
			{
				if($PTP[FILE] ne "")
				{
					close ($PTP[FC]) || die "can't close ($PTP[FILE]: $!";
					prlog("File $PTP[FILE] detached from device PTP and closed\n");
					$PTP[FILE] = "";         		# file detached and closed; device inoperable	
				}				
				if(! open (PTP, ">:raw", $file))
				{
					prlog("Cannot open papertape punch file: $file\n");
				}
				else
				{
					$PTP[FC] = \*PTP;
					prlog("File $file attached to device PTP\n");
					$PTP[FILE] = $file;		      # file attached and open
				}
			}				
		}
		elsif($w[1] eq "lpt")
		{
			$file = $w[2];
			if($file eq "")
			{
				prlog("No proper filename\n");
			}
			else
			{
				if($LPT[FILE] ne "")
				{
					close ($LPT[FC]) || die "can't close ($LPT[FILE]: $!";
					prlog("File $LPT[FILE] detached from device LPT and closed\n");
					$LPT[FILE] = "";         		# file detached and closed; device inoperable	
				}				
				if(! open (LPT, ">$file"))
				{
					prlog("Cannot open lineprinter file: $file\n");
				}
				else
				{
					$LPT[FC] = \*LPT;
					prlog("File $file attached to device LPT\n");
					$LPT[FILE] = $file;		      # file attached and open
					$LPT[LCUR]  = 0;              # set on line 0 current page 	
				}
			}					
		}
		elsif($w[1] eq "md0")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			my $stat = $MD0[STATE] >> 16;   		# get status: cu state
			if($file eq "" || $fy ne "img")
			{
				prlog("No filename, or wrong extension (must be a .img file)\n");
			}
			else
			{
				if($MD0[FILE] ne "")
				{
					if($stat != INAC)
					{
						if($MD0[DN] == 0)
						{
							prlog("MD0 is active and its attached file can not be closed\n");
						}
						else
						{
							close (MD0) || die "can't close $MD0[FILE]: $!";
							prlog("File $MD0[FILE] detached from device MD0 and closed\n");
							$MD0[FILE] = ""; 			# file detached and closed; device inoperable
						}
					}
					else
					{
						close (MD0) || die "can't close $MD0[FILE]: $!";
						prlog("File $MD0[FILE] detached from device MD0 and closed\n");
						$MD0[FILE] = ""; 				# file detached and closed; device inoperable
					}					
				}
				if($MD0[FILE] eq "")
				{					
					if(! open (MD0, "+<:raw", $file)) # open for update
					{
						prlog("Cannot open disk file: $file\n");
					}
					else
					{
						prlog("File $file attached to device MD0\n");
						$MD0[FILE] = $file;        # file attached and open
						$MD0[CYL]  = 0;
					}
					if($MD0[FILE] ne "" && $MD1[FILE] ne "")
					{
						$MD0[RDY] = 1;
						prlog("X1215 unit-0 is READY\n");
					}
					else
					{
						$MD0[RDY] = 0;
						prlog("X1215 unit-0 is NOT READY\n");
						prlog("To make unit-0 ready, both MD0 and MD1 must be attached to a diskimage file\n");
						#state_md0();
					}							
				}
			}
		}			
		elsif($w[1] eq "md1")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			my $stat = $MD0[STATE] >> 16;   # get status: cu state			
			if($file eq "" || $fy ne "img")
			{
				prlog("No filename, or wrong extension (must be a .img file)\n");
			}
			else
			{
				if($MD1[FILE] ne "")
				{
					if($stat != INAC)
					{
						if($MD0[DN] == 2)
						{
							prlog("MD1 is active and its attached file can not be closed\n");
						}
						else
						{
							close (MD1) || die "can't close $MD1[FILE]: $!";
							prlog("File $MD1[FILE] detached from device MD1 and closed\n");
							$MD1[FILE] = ""; # file detached and closed; device inoperable
						}
					}
					else
					{
						close (MD1) || die "can't close $MD1[FILE]: $!";
						prlog("File $MD1[FILE] detached from device MD1 and closed\n");
						$MD1[FILE] = ""; # file detached and closed; device inoperable
					}					
				}
				if($MD1[FILE] eq "")
				{				
					if(! open (MD1,"+<:raw", $file)) # open for update
					{
						prlog("Cannot open disk file: $file\n");
					}
					else
					{
						prlog("File $file attached to device MD1\n");
						$MD1[FILE] = $file;           # file attached and open
						$MD0[CYL]  = 0;
					}
					if($MD0[FILE] ne "" && $MD1[FILE] ne "")
					{
						$MD0[RDY] = 1;
						prlog("X1215 unit-0 is READY\n");
					}
					else
					{
						$MD0[RDY] = 0;
						prlog("X1215 unit-0 is NOT READY\n");
						prlog("To make unit-0 ready, both MD0 and MD1 must be attached to a diskimage file\n");
						#state_md0();
					}							
				}
			}
		}
		elsif($w[1] eq "md2")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			my $stat = $MD0[STATE] >> 16;   # get status: cu state			
			if($file eq "" || $fy ne "img")
			{
				prlog("No filename, or wrong extension (must be a .img file)\n");
			}
			else
			{
				if($MD2[FILE] ne "")
				{
					if($stat != INAC)
					{
						if($MD0[DN] == 1)
						{
							prlog("MD2 is active and its attached file can not be closed\n");
						}
						else
						{						
							close (MD2) || die "can't close $MD2[FILE]: $!";
							prlog("File $MD2[FILE] detached from device MD2 and closed\n");
							$MD2[FILE] = ""; # file detached and closed; device inoperable
						}
					}
					else
					{						
						close (MD2) || die "can't close $MD2[FILE]: $!";
						prlog("File $MD2[FILE] detached from device MD2 and closed\n");
						$MD2[FILE] = ""; # file detached and closed; device inoperable
					}					
				}
				if($MD2[FILE] eq "")
				{				
					if(! open (MD2,"+<:raw", $file)) # open for update
					{
						prlog("Cannot open disk file: $file\n");
					}
					else
					{
						prlog("File $file attached to device MD2\n");
						$MD2[FILE] = $file;           # file attached and open
						$MD2[CYL]  = 0;
					}
					if($MD2[FILE] ne "" && $MD3[FILE] ne "")
					{
						$MD2[RDY] = 1;
						prlog("X1215 unit-1 is READY\n");
					}
					else
					{
						$MD2[RDY] = 0;
						prlog("X1215 unit-1 is NOT READY\n");
						prlog("To make unit-1 ready, both MD2 and MD3 must be attached to a diskimage file\n");
						#state_md0();
					}							
				}
			}
		}		
		elsif($w[1] eq "md3")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			my $stat = $MD0[STATE] >> 16;   # get status: cu state				
			if($file eq "" || $fy ne "img")
			{
				prlog("No filename, or wrong extension (must be a .img file)\n");
			}
			else
			{
				if($MD3[FILE] ne "")
				{
					if($stat != INAC)
					{
						if($MD0[DN] == 3)
						{
							prlog("MD3 is active and its attached file can not be closed\n");
						}
						else
						{						
							close (MD3) || die "can't close $MD3[FILE]: $!";
							prlog("File $MD3[FILE] detached from device MD3 and closed\n");
							$MD3[FILE] = "";  # file detached and closed; device inoperable	
						}
					}
					else
					{						
						close (MD3) || die "can't close $MD3[FILE]: $!";
						prlog("File $MD3[FILE] detached from device MD3 and closed\n");
						$MD3[FILE] = "";  # file detached and closed; device inoperable	
					}					
				}
				if($MD3[FILE] eq "")
				{				
					if(! open (MD3,"+<:raw", $file)) # open for update
					{
						prlog("Cannot open disk file: $file\n");
					}
					else
					{
						prlog("File $file attached to device MD3\n");
						$MD3[FILE] = $file;           # file attached and open
						$MD2[CYL]  = 0;
					}
					if($MD2[FILE] ne "" && $MD3[FILE] ne "")
					{
						$MD2[RDY] = 1;
						prlog("X1215 unit-1 is READY\n");
					}
					else
					{
						$MD2[RDY] = 0;
						prlog("X1215 unit-1 is NOT READY\n");
						prlog("To make unit-1 ready, both MD2 and MD3 must be attached to a diskimage file\n");
						#state_md0();
					}						
				}
			}
		}		
		else
		{
			prlog("Unsupported device $w[1] requested for file attachement\n");
		}
	}
	elsif($w[0] eq "detach" || $w[0] eq "det")
	{
		if($w[1] eq "ptr")
		{
			$file = $PTR[FILE];
			if($file eq "")
			{
				prlog("Device PTR not attached\n");
			}
			else
			{
				close ($PTR[FC]) || die "can't close $file: $!";
				prlog("File $file detached from device PTR and closed\n");
				$PTR[FILE] = "";         # file detached and closed; device inoperable	
			}
		}
		elsif($w[1] eq "cdr")
		{
			$file = $CRD[FILE];
			if($file eq "")
			{
				prlog("Device CRD not attached\n");
			}
			else
			{
				close ($CRD[FC]) || die "can't close $file: $!";
				prlog("File $file detached from device CRD and closed\n");
				$CRD[FILE] = "";         # file detached and closed; device inoperable	
			}
		}	
		elsif($w[1] eq "ptp")
		{
			$file = $PTP[FILE];
			if($file eq "")
			{
				prlog("Device PTP not attached\n");
			}
			else
			{
				close ($PTP[FC]) || die "can't close $file: $!";
				prlog("File $file detached from device PTP and closed\n");			
				$PTP[FILE] = "";			# file detached and closed; device inoperable		
			}		
		}
		elsif($w[1] eq "lpt")
		{
			$file = $LPT[FILE];
			if($file eq "")
			{
				prlog("Device LPT not attached\n");
			}
			else
			{
				close ($LPT[FC]) || die "can't close $file: $!";
				prlog("File $file detached from device LPT and closed\n");			
				$LPT[FILE] = "";			# file detached and closed; device inoperable		
			}			
		}
		elsif($w[1] eq "md0")
		{
			$file = $MD0[FILE];
			my $stat = $MD0[STATE] >> 16;   # get status: cu state					
			if($file eq "")
			{
				prlog("Device MD0 not attached\n");
			}
			else
			{
				if($stat != INAC)
				{
					if($MD0[DN] == 0)
					{
						prlog("MD0 is active and its attached file can not be closed\n");
					}
                }
				else
				{					
					close (MD0) || die "can't close $file: $!";
					prlog("File $file detached from device MD0 and closed\n");			
					$MD0[FILE] = ""; # file detached and closed; device inoperable
					$MD0[RDY] = 0;
					prlog("X1215 unit-0 is NOT READY\n");						
				}
			}
		}
		elsif($w[1] eq "md1")
		{
			$file = $MD1[FILE];
			my $stat = $MD0[STATE] >> 16;   # get status: cu state					
			if($file eq "")
			{
				prlog("Device MD1 not attached\n");
			}
			else
			{
				if($stat != INAC)
				{
					if($MD0[DN] == 2)
					{
						prlog("MD1 is active and its attached file can not be closed\n");
					}
                }
				else
				{					
					close (MD1) || die "can't close $file: $!";
					prlog("File $file detached from device MD1 and closed\n");			
					$MD1[FILE] = ""; # file detached and closed; device inoperable	
					$MD0[RDY] = 0;						
					prlog("X1215 unit-0 is NOT READY\n");						
				}						
			}				
		}
		elsif($w[1] eq "md2")
		{
			$file = $MD2[FILE];
			my $stat = $MD0[STATE] >> 16;   # get status: cu state					
			if($file eq "")
			{
				prlog("Device MD2 not attached\n");
			}
			else
			{
				if($stat != INAC)
				{
					if($MD0[DN] == 1)
					{
						prlog("MD2 is active and its attached file can not be closed\n");
					}
                }
				else
				{						
					close (MD2) || die "can't close $file: $!";
					prlog("File $file detached from device MD2 and closed\n");			
					$MD2[FILE] = ""; # file detached and closed; device inoperable	
					$MD2[RDY] = 0;						
					prlog("X1215 unit-1 is NOT READY\n");
				}
			}				
		}
		elsif($w[1] eq "md3")
		{
			$file = $MD3[FILE];
			my $stat = $MD0[STATE] >> 16;   # get status: cu state					
			if($file eq "")
			{
				prlog("Device MD3 not attached\n");
			}
			else
			{
				if($stat != INAC)
				{
					if($MD0[DN] == 3)
					{
						prlog("MD3 is active and its attached file can not be closed\n");
					}
                }
				else
				{						
					close (MD3) || die "can't close $file: $!";
					prlog("File $file detached from device MD3 and closed\n");			
					$MD3[FILE] = ""; # file detached and closed; device inoperable
					$MD2[RDY] = 0;						
					prlog("X1215 unit-1 is NOT READY\n");
				}						
			}	
		}
		else
		{
			prlog("Unsupported device $w[1]\n");
		}
	}
	elsif($w[0] eq "state")
	{
		if($w[1] eq "cpu")
		{
			prlog("CPU state\n");
			trace($R[0]);
		}
		elsif($w[1] eq "tty")
		{
			prlog("TTY state: ");
			state_tty();
		}
		elsif($w[1] eq "rtc")
		{
			prlog("RTC state: ");
			state_clk();
		}	
		elsif($w[1] eq "ptr")
		{
			prlog("PTR state: ");
			state_ptr();
		}
		elsif($w[1] eq "ptp")
		{
			prlog("PTP state: ");
			state_ptp();
		}
		elsif($w[1] eq "lpt")
		{
			prlog("LPT state: ");
			state_lpt();
		}	
		elsif($w[1] eq "crd")
		{
			prlog("CRD state: ");
			state_crd();
		}
		elsif($w[1] eq "md0" || $w[1] eq "md1" || $w[1] eq "md2" || $w[1] eq "md3")
		{
			prlog("MD* state: ");
			state_md0();
		}		
		else
		{
			prlog("CPU state\n");
			trace($R[0]);
			prlog("TTY state: ");
			state_tty();
			prlog("RTC state: ");
			state_clk();
			prlog("PTR state: ");
			state_ptr();
			prlog("PTP state: ");
			state_ptp();
   			prlog("LPT state: ");
			state_lpt();
   			prlog("CRD state: ");
			state_crd();
			prlog("MD* state: ");
			state_md0();            			
		}
	}
	elsif($cmd eq "queue" || $cmd eq "q")
	{
		show_event_queue();
	}
	elsif($w[0] eq "echo")
	{	
		prlog(eval($w[1]));
	}
	elsif($w[0] eq "console")
	{
		if($w[1] eq "64")
		{
			$TTY[CSET] = 64;
		}
		elsif($w[1] eq "128")	
		{
			$TTY[CSET] = 128;	
		}	
		else
		{
			prlog("Wrong console character set selected, set to 64\n");
			$TTY[CSET] = 64;
		}
	}
	elsif($cmd eq "exit")
	{
		# close still open/attached files
		if($PTR[FILE] ne "")
		{
			close ($PTR[FC]) || die "can't close $PTR[FILE]: $!";
			prlog("File $PTR[FILE] detached from device PTR and closed\n");	
		}
		if($PTP[FILE] ne "")
		{
			close ($PTP[FC]) || die "can't close $PTP[FILE]: $!";
			prlog("File $PTP[FILE] detached from device PTP and closed\n");
		}
		if($LPT[FILE] ne "")
		{
			close ($LPT[FC]) || die "can't close $LPT[FILE]: $!";
			prlog("File $LPT[FILE] detached from device LPT and closed\n");
		}
		if($CRD[FILE] ne "")
		{
			close ($CRD[FC]) || die "can't close $CRD[FILE]: $!";
			prlog("File $CRD[FILE] detached from device CRD and closed\n");
		}
		if($MD0[FILE] ne "")
		{
			close (MD0) || die "can't close $MD0[FILE]: $!";
			prlog("File $MD0[FILE] detached from device MD0 and closed\n");
		}
		if($MD1[FILE] ne "")
		{
			close (MD1) || die "can't close $MD1[FILE]: $!";
			prlog("File $MD1[FILE] detached from device MD1 and closed\n");
		}
		if($MD2[FILE] ne "")
		{
			close (MD2) || die "can't close $MD2[FILE]: $!";
			prlog("File $MD2[FILE] detached from device MD2 and closed\n");
		}
		if($MD3[FILE] ne "")
		{
			close (MD3) || die "can't close $MD3[FILE]: $!";
			prlog("File $MD3[FILE] detached from device MD3 and closed\n");
		}		
		if($do ne "")
		{
			close DO || die "can't close $do: $!";
			prlog("Dofile $do closed\n");
		}
		if($tracefile ne "")
		{
			close TRACE || die "can't close $tracefile: $!";
			prlog("Tracefile $tracefile closed\n");
		}
		if($lptwin ne "")
        {
            close LPW || die "can't close LPW: $!";
            prlog("LPW closed\n");
        }            
		prlog ("EXIT!\n");
		if($logfile ne "")
		{
			close LOG || die "can't close $logfile: $!";
			prlog("Logfile $logfile is closed\n");	
		}		
		exit;
	}
	else
	{
		prlog("Unknown command\n");
	}
	@w = (); # reset
}

###########################################################################
# logging
sub prlog {
	my $s=shift;
	print $s;
	if($log)
	{
		print LOG $s;
	}
}
# commands not echo'd to console
sub cmdlog{
 my $s=shift;
	if($log)
	{
		print LOG $s;
	}
}

############################## Emulator #################################
# execute a single instruction at address $R[0]
#
sub instr {

  $interval = $interval - 1;
	if($interval <= 0)
	{
		process_events();
	}
	if($brkreg)
	{
		# Break (operates currently only for a single IOP with address 0 !)
		$t1 = $brkreg;
		$t2 = 0;
		while(! ($t1 & 1))
		{
			$t1 = $t1 >> 1;
			$t2 = $t2 + 1;
		}
		# $t2 = subchannel number requesting a break
		# the control unit must be in EXCHANGE state
		$t1 = 1 << $t2;
		$brkreg = $brkreg & (~ $t1);    # reset break request
		$t1 = $DEV{$t2};                # cu on IOP 0 requesting the break
		$t3 = $$t1[STATE] >> 16;        # cu state
		$a  = $t2 << 1;                 # index of external register used by subchannel		
		$mpx1 = $EXTR[$a];	           # w1: function: bit0 0=chr 1=wrd bit1 0=in 1=out; bit5-15 length in chr/wrd
		$mpx2 = $EXTR[$a+1];            # w2: buffer start cq pointer to byte/word to be read/written
		$tx3 = $mpx1 & 0xFFF;           # length to tranfer (still)
	
		if($mpx1 & 0x4000)	# check IO-bit in register 1
		{	
			# IO-bit is 1 => OUTPUT
			if($t3 == OEXC)
			{
				# indeed the CU is in outputmode
				if($tx3 == 0)
				{
					# end of range; transfer ready
					# switch control unit to the execute state
					$$t1[STATE] = $$t1[STATE] & 0xFFFF;           # reset cu state bits
					$$t1[STATE] = $$t1[STATE] | (OEXE << 16);     # state = output + execute										
				}				
				else
				{
					$t3 = $$t1[OTR];            # call output service routine					
					if($mpx1 & 0x8000) # check C/W-bit in register 1
					{
						# C/W-bit is 1 => word transfer			
						$$t1[BUFFER] = $M[$mpx2 & 0xFFFE];	
						$mpx1 = $mpx1 - 1;		# update length to transfer (length counted in words)
						$mpx2 = $mpx2 + 2;      # update position
						&$t3(1);                # transfer a word							
					}
					else
					{
						# C/W-bit is 0 => character transfer
						if($mpx2 & 1)
						{
							$$t1[BUFFER] = $M[$mpx2 & 0xFFFE] & 0xFF;        # righthand byte
						}
						else
						{
							$$t1[BUFFER] = ($M[$mpx2 & 0xFFFE] >> 8) & 0xFF; # lefthand byte
						}
						$mpx1 = $mpx1 - 1;		# update length to transfer
						$mpx2 = $mpx2 + 1;      # update position
						&$t3(0);                # transfer a byte	
					}
					$EXTR[$a]   = $mpx1;       # update register 1 (length)
					$EXTR[$a+1] = $mpx2;       # update register 2 (position)					
				}
			}
			else
			{
				# control unit state error		
				prlog("Error Output Multiplex channel, DEV:" . prhex(8,$t2) . " STATE:" . prhex(20,$t3) . "\n");
				mpxerror();
			}
		}
		else
		{
			# IO-bit is 0 => INPUT
			if($t3 == IEXC)
			{
				# indeed the CU is in inputmode
				if($tx3 == 0)
				{
					# end of range; transfer ready
					# switch control unit to the wait request state
					$$t1[STATE] = $$t1[STATE] & 0xFFFF;           # reset cu state bits
					$$t1[STATE] = $$t1[STATE] | (WREQ << 16);    									
				}
				else
				{
					$t3 = $$t1[INR];            # call input service routine				
					if($mpx1 & 0x8000) # check C/W-bit in register 1
					{
						# C/W-bit is 1 => word transfer
						&$t3(1);	# ask for a word
						$M[$mpx2 & 0xFFFE] = $$t1[BUFFER] & 0xFFFF;
						$mpx1 = $mpx1 - 1;		# update length (length counted in words)
						$mpx2 = $mpx2 + 2;      # update position							
					}
					else
					{
						&$t3(0);	# ask for a byte
						if($mpx2 & 1)
						{
							# righthand byte
							$M[$mpx2 & 0xFFFE] = ($M[$mpx2 & 0xFFFE] & 0xFF00) | ($$t1[BUFFER] & 0xFF);
						}
						else
						{
							# lefthand byte
							$M[$mpx2 & 0xFFFE] = ($M[$mpx2 & 0xFFFE] & 0xFF) | (($$t1[BUFFER] & 0xFF) << 8);
						}
						$mpx1 = $mpx1 - 1;		# update length
						$mpx2 = $mpx2 + 1;      # update position					
					}
					$$t1[BUFFER] = 0;           # reset buffer
					$EXTR[$a]   = $mpx1;        # update register 1 (length)
					$EXTR[$a+1] = $mpx2;        # update register 2 (position)							
				}
			}
			else
			{
				# control unit state error	
				prlog("Error Input Multiplex channel, DEV:" . prhex(8,$t2) . " STATE:" . prhex(20,$t3) . "\n");
				mpxerror();				
			}
		}
	}
	
    #if($ENB || $UM) # implicit ENB in User mode
    if($ENB) # implicit ENB in User mode
	{
		# interrupt?
		if($intreg)
		{		
			$t1 = $intreg;
			$t2 = 0;		# t2 is interrupt level
			while(! ($t1 & 0x80000000))
			{
				$t1 = $t1 << 1;
				$t2 = $t2 + 1;
			}
			# print "Interrupt: current PL:$PL new PL:$t2\n";
			if($t2 < $PL)   # if interrupt level < current level => interrupt
			{
			    # P, PSW on stack
			    $M[$R[A15]] = $R[A0];              # save P
			    $R[A15] = $R[A15] - 2;             # decrease link stack
			    if($R[A15] < 0x100)
			    {
					trap(LKM);		                  # A15 stack overflow
				}				
				$M[$R[A15]] = Fpsw();          		# save PSW
				$R[A15] = $R[A15] - 2;              # decrease link stack
				if($R[A15] < 0x100)
				{
					trap(LKM);		                  # A15 stack overflow
				}											
			    $ENB = 0;                          # inhibit
                $UM  = 0;                       # system mode
				$PL = $t2;		                     # set new level
				$R[A0] = $M[$t2+$t2];				   # fetch address of interrupt routine
			}
		}
	}
	$R[A0] = $R[A0] & 0xFFFE;            			# even address and 16 bit
	$IR = $M[$R[A0]];                    			# fetch instruction
    $pcsaved = $R[A0];                   			# save pc for trace
	$R[A0] = ($R[A0] + 2);               			# incr location counter
	if(! ($IR & mskf10))
	{
		# format 0
		# - short constant instructions
		# - shift instructions
		# - i/o instructions
		# - miscellanious instructions
		# execute a format 0 instruction		
		format0();
	}	
	elsif(!($IR & 1))
	{
		# format 1, ls bit 0
		format10();
	}
	else
	{
		# format 1, ls bit 1
		format11();
	}
	
	if($EmBreak)
	{
		# breakpoint?
		if($BP{$pcsaved} ne undef)
		{
			$RUN = 0;
			prlog("\nBreakpoint, last instruction executed and result:\n");
			if(! $EmTrace)
			{
				trace($pcsaved);
			}
		}
	}
	if($EmTrace)
	{
		# display the result of the executed instruction
		if($batrace == 0)
		{
			trace($pcsaved); # trace all
		}
		else
		{
			rtrace($pcsaved); # trace instructions a region
		}
	}
}

# emulate the format 1 instructions with the ls bit == 0:	
sub format10 {

$opc = ($IR >> 11) & 0xF;                 # opcode
$r1  = ($IR >> 7)  & 0xF;                 # r1 or condition shifted 1 left	
$r2  = ($IR >> 1)  & 0xF;                 # r2
$md  = ($IR >> 5)  & 0x3;                 # addressing mode

if($opc == 0)
{
	# load
	if($md == 0)				            	# register operand
	{
		# T1 LDR r1,r2
		$R[$r1] = $R[$r2];
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3A/B LDR* r1,r2, T2 LDKL r1,lk
		if($r2 == 0xF)                      # A15 ?
		{
			# A15, the stack pointer is addressed
			$R[0xF] = ($R[0xF] + 2) & 0xFFFE;
		}
		$R[$r1] = $M[($R[$r2] & 0xFFFE)];
		if($r2 == 0)
		{
			# T2 indirect via A0: LDKL r1,lk
			$R[0] = ($R[0] + 2);         		# increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 LD r1,m or LD r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);        	    	# increment A0 (PC)
		if($r2 != 0)			            	# indexing ?
		{
			$a = ($a + $R[$r2]);            	# yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$R[$r1] = $M[$a];                   # load content of effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 LD* r1,m or LD* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);         	    	# increment A0 (PC)
		if($r2 != 0)			            	# indexing ?
		{
			$a = ($a + $R[$r2]);   				# yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
		$R[$r1] = $M[$a];                   # load content of effective address
	}
	setCR1($r1);
}
elsif($opc == 1)
{
	$r1=$r1>>1;								    	# position condition (added in 0.8)
	# absolute branche
    if($md == 0)				               # register operand
	{
		# T1 ABR(cnd) r2
		# $r1 is the condition code in the operation
		if(evalcond($r1))
		{
			$R[0] = $R[$r2];
		}
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3 ABR(cnd)* r2
		# $r1 is the condition code in the operation
		if(evalcond($r1))
		{
			$R[0] = $M[$R[$r2] & 0xFFFE];    # jump
		}
		elsif($r2 == 0)                     # no jump, skip address in case of an ABL
		{
			# T2 indirect via A0: ABL(cnd) lk
			$R[0] = $R[0] + 2;           	   # increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{	
		# T4/T5 ABI(cnd) m or ABI(cnd) m,r2
		# $r3 is the condition code in the operation
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);         	      # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $a & 0xFFFE;                      # effective address
		if(evalcond($r1))
		{
			$R[0] = $M[$a];                    # load P with content of effective address
		}
	}
	elsif($md == 3)				               # indirect memory/indexed memory operand
	{
		# T6/T7 ABI(cnd)* m or ABI(cnd)* m,r2
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);         	       # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;         # memory content = effective address
		if(evalcond($r1))
		{
			$R[0] = $M[$a];                    # load P with content of effective address
		}
	}		
}
elsif($opc == 2)
{
	# add, increment memory
	if($md == 0)				            # register operand
	{
		# T1 ADR r1,r2
		$t1 = $R[$r1]; $t2 = $R[$r2]; $t3 = $t1 + $t2;
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3 ADR* r1,r2, T2 ADKL r1,lk
		$t1 = $R[$r1]; $t2 = $M[($R[$r2] & 0xFFFE)]; $t3 = $t1 + $t2;
		if($r2 == 0)
		{
			# T2 indirect via A0: ADKL r1,lk
			$R[0] = ($R[0] + 2);        	# increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 AD r1,m or AD r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$t1 = $R[$r1]; $t2 = $M[$a]; 
		$t3 = $t1 + $t2;                    # add content of effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 AD* r1,m or AD* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
		$t1 = $R[$r1]; $t2 = $M[$a];        
		$t3 = $t1 + $t2;                    # add content of effective address
	}
	# overflow? (+)
	if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & SIGN)
	{
		$overflow = 1;
	}
	$R[$r1] = setCRvalue($t3);
}
elsif($opc == 3)
{
	# subtract, two complement
	if($md == 0)				            # register operand
	{
		# T1 SUR r1,r2
		$t1 = $R[$r1]; $t2 = $R[$r2]; $t3 = $t1 - $t2;
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3 SUR* r1,r2, T2 SUKL r1,lk
		$t1 = $R[$r1]; $t2 = $M[($R[$r2] & 0xFFFE)]; $t3 = $t1 - $t2;
		if($r2 == 0)
		{
			# T2 indirect via A0: ADKL r1,lk
			$R[0] = ($R[0] + 2);        	# increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 SU r1,m or SU r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$t1 = $R[$r1]; $t2 = $M[$a]; 
		$t3 = $t1 - $t2;                    # subtract content of effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 SU* r1,m or SU* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
		$t1 = $R[$r1]; $t2 = $M[$a];        
		$t3 = $t1 - $t2;                    # subtract content of effective address
	}
	# underflow? (-)
	if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
	{
		$overflow = 1;
	}	
	$R[$r1] = setCRvalue($t3);
}
elsif($opc == 4)
{
	# AND
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 ANR r1,r2
			$t3 = $R[$r1] & $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 ANR* r1,r2, T2 ANKL r1,lk
			$t3 = $R[$r1] & $M[($R[$r2] & 0xFFFE)];
			if($r2 == 0)
			{
				# T2 indirect via A0: ADKL r1,lk
				$R[0] = ($R[0] + 2);	        # increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 AN r1,m or AN r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = $R[$r1] & $M[$a];             # and content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 AN* r1,m or AN* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = $R[$r1] & $M[$a];             # and content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "AND10");
	}			
}
elsif($opc == 5)
{
	# OR
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 ORR r1,r2
			$t3 = $R[$r1] | $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 ORR* r1,r2, T2 ORKL r1,lk
			$t3 = $R[$r1] | $M[($R[$r2] & 0xFFFE)];
			if($r2 == 0)
			{
				# T2 indirect via A0: ADKL r1,lk
				$R[0] = ($R[0] + 2);	# increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 OR r1,m or OR r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = $R[$r1] | $M[$a];             # or content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 OR* r1,m or OR* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = $R[$r1] | $M[$a];             # or content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "OR10");
	}	
}
elsif($opc == 6)
{
	# XOR/TNM
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 XRR r1,r2
			$t3 = $R[$r1] ^ $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 XRR* r1,r2, T2 XRKL r1,lk
			$t3 = $R[$r1] ^ $M[($R[$r2] & 0xFFFE)];
			if($r2 == 0)
			{
				# T2 indirect via A0: ADKL r1,lk
				$R[0] = ($R[0] + 2);	        # increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 XR r1,m or XR r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = $R[$r1] ^ $M[$a];             # xr content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 XR* r1,m or XR* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = $R[$r1] ^ $M[$a];             # xr content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "XOR10");
	}			
}
elsif($opc == 7)
{
	# multiple load
    if($md == 1)				                # indirect register operand
	{
		# T3A/B MLR n,r2, T2 MLK n
		# $r1 is n
		$t1 = 1;
		if($r2 == 0xF)
		{
			# A15, the stack pointer is addressed
			$a = $R[0xF] = ($R[0xF] + $r1*2) & 0xFFFE;
			while($t1 <= $r1)
			{
				$R[$A[$t1]] = $M[$a];
				$a = $a - 2;
				$t1 = $t1 + 1;
			}
		}
		else
		{
			$a = $R[$r2] & 0xFFFE;              # effective address
			while($t1 <= $r1)
			{
				$R[$A[$t1]] = $M[$a];
				$a = $a + 2;
				$t1 = $t1 + 1;
			}
		}
		if($r2 == 0)
		{
			# T2 indirect via A0: MLK n ($r1 = n)
			$R[0] = ($R[0] + $r1*2);	       # increment A0 (PC)
		}
	}
	elsif($md == 2)				               # memory/indexed memory operand
	{
		# T4/T5 ML n,m or ML n,m,r2
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);	               # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $a & 0xFFFE;                      # effective address
		# $r1 is n
		$t1 = 1;
		while($t1 <= $r1)
		{
			$R[$A[$t1]] = $M[$a];              # load content of effective address
			$a = $a + 2;
			$t1 = $t1 + 1;
		}			
	}
	elsif($md == 3)				               # indirect memory/indexed memory operand
	{
		# T6/T7 ML* n,m or ML* n,m,r2
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);	               # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;         # memory content = effective address
		# $r1 is n
		$t1 = 1;
		while($t1 <= $r1)
		{
			$R[$A[$t1]] = $M[$a];              # load content of effective address
			$a = $a + 2;
			$t1 = $t1 + 1;
		}
	}
	else
	{
		# $md == 0
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "ML10");
	}
	$R[2] = setCRvalue($R[2]); # condition register set by A1		
}
elsif($opc == 8)
{
	# Multiply
	if($r1 == 0)
	{
		my $s = 0;		                               # collect sign result
		$t2 = $R[4];                                   # A2				
		if($md == 0)
		{
			# T1 MUR r2
			$t1 = $R[$r2];                             # operand
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 MUR* r2
				$t1 = $M[($R[$r2] & 0xFFFE)];          # operand
			}
			else
			{
				# T2 MUK lk
				$t1 = $M[$R[0]];                       # lk, operand
				$R[0] = ($R[0] + 2);	               # increment A0 (PC)
			}
		}
		elsif($md == 2)
		{
			# T4/T5 MU m or MU m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t1 = $M[$a];                       # operand
		}
		elsif($md == 3)
		{
			# T6/T7 MU* m or MU* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t1 = $M[$a];                       # operand	
		}
		
		$t1 = $t1 & DMASK; 
		$t2 = $t2 & DMASK; 
		if($t1 & SIGN)
		{
			$t1 = ((~ $t1) + 1) & DMASK;        # sign - => 2-compl
			$s = $s ^ 1;
		}
		if($t2 & SIGN)
		{
			$t2 = ((~ $t2) + 1) & DMASK;        # sign - => 2-compl
			$s = $s ^ 1;
		}
		$t3 = ($t1 * $t2) & DPDMASK;            # A2 * operand 
		if($t3 != 0)
		{
			if($s)
			{
				$t3 = ((~ $t3) & DPDMASK) + 1;  # sign - => 2-compl
				if( ! ($t3 & DPSIGN))
				{
					$overflow = 1;
				}
			}
			else
			{
				if($t3 & DPSIGN)
				{
					$overflow = 1;
				}
			}
		}
		setCR2s($t3);
		putdbls($t3);
	}
	else
	{	
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"MUL10");
	}
}
elsif($opc == 9)
{
	if($r1 == 0)
	{
		# Divide
		my $s  = 0;		                               # collect sign result (quotient)
		my $sr = 0;                                  # sign remainder (gets sign of dividend)
		$t1 = getdbls();					                   # A1,A2 dividend				
		if($md == 0)
		{
			# T1 DVR r2
			$t2 = $R[$r2];                             # operand, divisor
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 DVR* r2
				$t2 = $M[($R[$r2] & 0xFFFE)];          # operand, divisor
			}
			else
			{
				# T2 DVK lk
				$t2 = $M[$R[0]];                       # lk, operand, divisor
				$R[0] = ($R[0] + 2);	               # increment A0 (PC)
			}
		}
		elsif($md == 2)
		{
			# T4/T5 DV m or DV m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t2 = $M[$a];                       # operand, divisor
		}
		elsif($md == 3)
		{
			# T6/T7 DV* m or DV* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t2 = $M[$a];                       # operand, divisor
		}
		if($t1 & DPSIGN)
		{
			$t1 = (((~ $t1) & DPDMASK) + 1) & DPDMASK; # make positive (2-compl)
			$s = $sr = $s ^ 1;                         # and save sign / sign remainder
		}
		if($t2 & SIGN)
		{
			$t2 = (((~ $t2) & DMASK) + 1) & DMASK;
			$s = $s ^ 1;
		}

		#my $tstr = "t1" . prhex(32,$t1) . " t2" . prhex(32,$t2) . " s" . prhex(16,$s) . " sr" . prhex(16,$sr) . "\n" ;
		#print $tstr;

        	# 1st test divisor then dividend
		if($t2 == 0)
		{
			$overflow = 1;       # divisor = 0
			$t3 = 0x7FFF;
			$t1 = 0;
		}
		elsif($t1 == 0)
		{
			$t3 = 0;	     # dividend = 0 => result = 0
		}
		else
		{
			$t3 = int($t1 / $t2);    # quotient
			#if($t3 & 0xFFFF8000)
			#{
			#	$overflow = 1; # quotient bigger than 15 bit
			#}
			
			#my $tstr = "t3(Q)" . prhex(32,$t3) . "\n" ;
			#print $tstr;
			
			#if($s)
			#{
			#	$t3 = (((~ $t3) & DMASK) + 1) & DMASK;
			#}
            
            		if($s==0)               # positive quotient requested
            		{
                		if($t3 > 0x7FFF)
                		{
                    			$overflow = 1;  # + quotient bigger than 15 bit
                		}
            		}
            		if($s==1)           # negative quotient requested
            		{
                		if($t3 > 0x8000)				
                		{
                    			$overflow = 1;   # for negative quotients, the lnn is allowed 
                		}
                		else
                		{
                    			# no overflow
                    			$t3 = (((~ $t3) & DMASK) + 1) & DMASK; # 2-comp
                		}
            		}            
			$t1 = $t1 % $t2;	# remainder
			
			#my $tstr = "t1(R)" . prhex(32,$t1) . "\n" ;	
			#print $tstr;
			
			if($sr)              # remainder negative ?
			{
				$t1 = (((~ $t1) & DMASK) + 1) & DMASK;
                
                                #### Remainder op overflow testen ??????? zie div3.pl
                
				#$t1 = ~ $t1 & DMASK;
				#my $tstr = "t1(R)" . prhex(32,$t1) . "\n" ;	
				#print $tstr;
				#$t1 = $t1 + 1;
				#my $tstr = "t1(R)" . prhex(32,$t1) . "\n" ;	
				#print $tstr;
				#$t1 = $t1 & DMASK;
				#my $tstr = "t1(R)" . prhex(32,$t1) . "\n" ;	
				#print $tstr;
			}
		}
		
		#my $tstr = "t3 A2 Q" . prhex(32,$t3) . " t1 A1 R" . prhex(32,$t1) . "\n" ;	
		#print $tstr;
		
		$R[4] = setCRvalue($t3);                # A2 = quotient
		$R[2] = $t1;                            # A1 = remainder
	}
	else
	{	
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"DIV10");
	}
}
elsif($opc == 10)
{
	# double add
	if($r1 == 0)
	{
		$t1 = getdbls();								# A1,A2 double accumulator
		if($md == 0)
		{
			# T1 DAR r2
			$t2 = $R[$r2] << 15;
			if($r2 == 0xE)
			{
				$r2 = 1;
			}
			else
			{
				$r2 = $r2 + 2;
			}
			if($r2 > 0xF)
			{
				# illegal instruction
				ilcerror($R[0]-2, $IR, $opc,"DADD10a");
			}
			$t2 = $t2 | ($R[$r2] & MMASK);				# operand
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 DAR* r2
				$t2 = $M[($R[$r2] & 0xFFFE)] << 15;
				$t2 = $t2 | ($M[(($R[$r2]+2) & 0xFFFE)] & MMASK);
			}
			else
			{
				# T2 DAK lk1,lk2
				$t2 = $M[$R[0]];         
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t3 = $M[$R[0]];		
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t2 = ($t2 << 15) | ($t3 & MMASK);		# operand
			}
		}
		elsif($md == 2)
		{
			# T4/T5 DA m or DA m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | ($M[($a + 2) & 0xFFFE] & MMASK); # operand
		}
		elsif($md == 3)
		{
			# T6/T7 DA* m or DA* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | ($M[($a + 2) & 0xFFFE] & MMASK); # operand		
		}
		$t3 = $t1 + $t2;
		# overflow ?
		if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & DPSIGN)
		{
			$overflow = 1;
		}
		setCR2s($t3);
		putdbls($t3);		
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"DADD10");
	}
}
elsif($opc == 11)
{
	# double sub
	if($r1 == 0)
	{
		$t1 = getdbls();								# A1,A2 double accumulator
		if($md == 0)
		{
			# T1 DSR r2
			$t2 = $R[$r2] << 15;
			if($r2 == 0xE)
			{
				$r2 = 1;
			}
			else
			{
				$r2 = $r2 + 2;
			}
			if($r2 > 0xF)
			{
				# illegal instruction
				ilcerror($R[0]-2, $IR, $opc,"DSUB10a");
			}
			$t2 = $t2 | ($R[$r2] & MMASK);				# operand
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 DSR* r2
				$t2 = $M[($R[$r2] & 0xFFFE)] << 15;
				$t2 = $t2 | ($M[(($R[$r2]+2) & 0xFFFE)] & MMASK);
			}
			else
			{
				# T2 DSK lk1,lk2
				$t2 = $M[$R[0]];         
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t3 = $M[$R[0]];		
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t2 = ($t2 << 15) | ($t3 & MMASK);		# operand
			}
		}
		elsif($md == 2)
		{
			# T4/T5 DS m or DS m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | ($M[($a + 2) & 0xFFFE] & MMASK); # operand
		}
		elsif($md == 3)
		{
			# T6/T7 DS* m or DS* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | ($M[($a + 2) & 0xFFFE] & MMASK); # operand		
		}
		$t3 = $t1 - $t2;
		# underflow ? (-)
		if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & DPSIGN)
		{
			$overflow = 1;
		}	
		setCR2s($t3);
		putdbls($t3);
	}
	else
	{	
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"DSUB10");
	}
}
elsif($opc == 12)
{
	# character transfer
	if($r1 != 0)
	{
		# load
		if($md == 0)				            # register operand
		{
			# T1 ECR r1,r2
			$t1 = ($R[$r2] >> 8) & 0xFF;
			$R[$r1] = (($R[$r2] << 8) | $t1) & 0xFFFF;
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 LCR r1,r2, T2 LCK r1,lk
			$a = $R[$r2];
			$t1 = $M[$a & 0xFFFE];
			if($a & 1)
			{
				$t1 = $t1 & 0xFF;               # 1 = right hand byte
			}
			else
			{
				$t1 = ($t1 >> 8) & 0xFF;        # 0 = left hand byte
			}
	
			if($r2 == 0)
			{
				# T2 indirect via A0: LCK r1,lk
				$R[0] = ($R[0] + 2);	        # increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 LC r1,m or LC r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			# $a is byte address !
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = $t1 & 0xFF;               # 1 = right hand byte
			}
			else
			{
				$t1 = ($t1 >> 8) & 0xFF;        # 0 = left hand byte
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 LC* r1,m or LC* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE];               # memory content = effective byte address
			# $a is byte address !
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = $t1 & 0xFF;               # 1 = right hand byte
			}
			else
			{
				$t1 = ($t1 >> 8) & 0xFF;        # 0 = left hand byte
			}		
		}
		$R[$r1] = ($R[$r1] & 0xFF00) | $t1;	
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "LC10");	
	}		
}
elsif($opc == 13)
{
	# compare word
	if($md == 0)				                # register operand
	{
		# T1 CWR r1,r2
		$t1 = $R[$r1]; $t2 = $R[$r2];
	}
	elsif($md == 1)				                # indirect register operand
	{
		# T3 CWR* r1,r2, T2 CWK r1,lk
		$t2 = $M[($R[$r2] & 0xFFFE)];
		if($r2 == 0)
		{
			# T2 indirect via A0: CWK r1,lk
			$R[0] = ($R[0] + 2);	# increment A0 (PC)
		}
		$t1 = $R[$r1];
	}
	elsif($md == 2)				                # memory/indexed memory operand
	{
		# T4/T5 CW r1,m or CW r1,m,r2
		$a = $M[$R[0]];			                # memory operand address
		$R[0] = ($R[0] + 2);	                # increment A0 (PC)
		if($r2 != 0)			                # indexing ?
		{
			$a = ($a + $R[$r2]);                # yes
		}
		$a = $a & 0xFFFE;                       # effective address
		$t1 = $R[$r1]; $t2 = $M[$a]; 
	}
	elsif($md == 3)				                # indirect memory/indexed memory operand
	{
		# T6/T7 CW* r1,m or CW* r1,m,r2
		$a = $M[$R[0]];			                # memory operand address
		$R[0] = ($R[0] + 2);	                # increment A0 (PC)
		if($r2 != 0)			                # indexing ?
		{
			$a = ($a + $R[$r2]);                # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;          # memory content = effective address
		$t1 = $R[$r1]; $t2 = $M[$a];        
	}
	# direct subtraction of t1 and t2 is not possible; prevent over/underflow
	# over/undeflow with subtraction is only possible with t1 and t2 having 
	# different signs (and when comparing 0x8000)
	if(($t1 ^ $t2) & SIGN)
	{
		# sign t1 and t2 are different, so underflow is possible
		if($t1 & SIGN)
		{
			# sign t1 is - (and so sign t2 is +)
			$t3 = -1;		                    # t1 < t2
		}
		else
		{
			# sign t1 is + (and so sign t2 is -)
			$t3 = 1;                            # t1 > t2
		}
	}
	elsif($t1 == $t2)                           # t1 == t2
	{
		# t1 and t2 can be 0x8000, the smallest neg integer. subtraction
		# (making 2complement) gives overflow
		$t3 = 0;
	}
	else
	{
		$t3 = $t1 - $t2;                       
	}
	setCRvalue($t3);                            # set condition register
}
elsif($opc == 14)
{
	# return
	if($r2 != 0)
	{
		# return from function, RTN r2
		$R[$r2] = ($R[$r2] + 2) & 0xFFFE;
		$t1 = $M[$R[$r2]];                            # PSW
		$CR = ($t1 >> 8) & 0x3;                       # restore saved CR
		
		  ##### p852 always restores PSW ??
			#$PL =  ($t1 >> 10) & 0x3F;                # restore saved PL		
			#$ENB = ($t1 >> 6) & 1;
			#####
			
		$R[$r2] = ($R[$r2] + 2) & 0xFFFE;	
		$R[0] = $M[$R[$r2]];                          # restore saved P
		if($r2 == 0xF)                                # A15 ?
		{
			$PL =  ($t1 >> 10) & 0x3F;                  # restore saved PL
			$ENB = ($t1 >> 6) & 1;					            # restore saved interrupt mode (P856,P857)
			#$ENB = 1;                                  # P852, set bit ***************************
			$UM  = $t1 & 1;                             # restore saved user/system mode
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "RTN10");			
	}
}
elsif($opc == 15)
{
	# 1-complement
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 C1R r1,r2
			$t3 = ~ $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 C1R* r1,r2
			$t3 = ~ $M[($R[$r2] & 0xFFFE)];
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 C1 r1,m or C1 r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = ~ $M[$a];                     # complement content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 C1* r1,m or C1* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = ~ $M[$a];                     # complement content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "C110");
	}	
}
else
{
	# illegal instruction
	ilcerror($R[0]-2, $IR, $opc,"OPC10");
}
} # end format10

# emulate the format 1 instructions	with the ls bit == 1:
sub format11 {

$opc = ($IR >> 11) & 0xF;   # opcode
$r1  = ($IR >> 7)  & 0xF;   # r1 or cnd shifted 1 left	
$r2  = ($IR >> 1)  & 0xF;   # r2
$md  = ($IR >> 5)  & 0x3;   # adressing mode

if($opc == 0)
{
	# store
	if($md == 1)				             # indirect register operand
	{
		# T3A/B STR r1,r2
		$a=$R[$r2] & 0xFFFE;
		if($r2 == 0xF)
		{
			# A15, the stack pointer is addressed
			$R[0xF] = ($R[0xF] - 2) & 0xFFFE;
			if($R[0xF] < 0x100) 
			{
				trap(LKM);		            # A15 stack overflow
			}
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 ST r1,m or ST r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 ST* r1,m or ST* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "ST11");
	}
	$M[$a] = $R[$r1] & DMASK;             # store content r1 in effective address	
}
elsif($opc == 2)
{
	# add, increment memory
	if($r1 != 0)
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 ADRS r1,r2
			$a  = $R[$r2];
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 ADS r1,m or ADS r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 ADS* r1,m or ADS* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		    $a = $M[$a & 0xFFFE];               # memory content = effective address
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "ADD11");
		}
		$a = $a & 0xFFFE;                       # effective address
		$t1 = $R[$r1]; $t2 = $M[$a];
		$t3 = $t1 + $t2;                        # add content of effective address		
		# overflow? (+)
		if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}		
		$M[$a] = setCRvalue($t3);               # store result in effective address
	}
	else
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 IMR r2
			$a = $R[$r2];
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 IM m or IM m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 IM* m or IM* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "IM11");
		}
		$a = $a & 0xFFFE;                       # effective address		
		$t1 = 1; $t2 = $M[$a];             
		$t3 = $t1 + $t2;                        # add 1 to content of effective addres		
		# overflow? (+)
		if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}		
		$M[$a] = setCRvalue($t3);               # store result in effective address		
	}
} 
elsif($opc == 3)
{
	# subtract, two complement
	if($r1 != 0)
	{
		if($md == 0) # not on P855/P860
		{
			# T1 NGR r1,r2
			$t1 = 0; $t2 = $R[$r2];
			$t3 = $t1 - $t2;                    # subtract content of effective address
			# underflow? (-)
			if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
			{
				$overflow = 1;
			}	
			$R[$r1] = setCRvalue($t3);		
		}
        else
        {
            if($md == 1)				            # indirect register operand
            {
                # T3 SURS r1,r2
                $a  = $R[$r2];                      # effective address
            }
            elsif($md == 2)				            # memory/indexed memory operand
            {
                # T4/T5 SUS r1,m or SUS r1,m,r2
                $a = $M[$R[0]];			            # memory operand address
                $R[0] = ($R[0] + 2);	            # increment A0 (PC)
                if($r2 != 0)			            # indexing ?
                {
                    $a = ($a + $R[$r2]);            # yes
                }
            }
            elsif($md == 3)				            # indirect memory/indexed memory operand
            {
                # T6/T7 SUS* r1,m or SUS* r1,m,r2
                $a = $M[$R[0]];			            # memory operand address
                $R[0] = ($R[0] + 2);	            # increment A0 (PC)
                if($r2 != 0)			            # indexing ?
                {
                    $a = ($a + $R[$r2]);            # yes
                }
                $a =  $M[$a & 0xFFFE];              # memory content = effective address
            }
            $a = $a & 0xFFFE;		                # effective address
            $t1 = $R[$r1]; $t2 = $M[$a];
            $t3 = $t1 - $t2;                        # subtract content of effective address
            # underflow? (-)
            if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
            {
                $overflow = 1;
            }		
            $M[$a] = setCRvalue($t3);               # store result in effective address
        }
	}
	else
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 C2R r2
			$a  = $R[$r2];                      # effective address
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 C2 m or C2 m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}	
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 C2* m or C2* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address			
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc,"C211");
		}
		$a = $a & 0xFFFE;
		$t1 = 0; $t2 = $M[$a];
		$t3 = $t1 - $t2;                        # subtract content of effective address			
		# underflow? (-)
		if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}			
		$M[$a] = setCRvalue($t3);               # store result in effective address
	}
}
elsif($opc == 4)
{
	# AND/TM
	if($r1 != 0)
	{
		if($md == 0)				                # register operand
		{
			# T1 TM r1,r2
			setCRvalue($R[$r1] & $R[$r2]);                         
		}
		else
		{
			if($md == 1)				            # indirect register operand
			{
				# T3 ANRS r1,r2
				$a  = $R[$r2] & 0xFFFE;             # effective address
			}
			elsif($md == 2)				            # memory/indexed memory operand
			{
				# T4/T5 ANS r1,m or ANS r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a = $a & 0xFFFE;                   # effective address                		
			}
			elsif($md == 3)				            # indirect memory/indexed memory operand
			{
				# T6/T7 ANS* r1,m or ANS* r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address           			
			}
			$M[$a] = setCRvalue($R[$r1] & $M[$a]);  # store result in effective address
		}
	}
	else
	{
		#CM clear memory
		if($md == 1)
		{
			# T3 CMR r2
			$a  = $R[$r2] & 0xFFFE;             # effective address
		}
		elsif($md == 2)
		{
			# T4/T5 CM m or CM m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address         			
		}
		elsif($md == 3)
		{
			# T6/T7 CM* m or CM* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address   				
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "CM11");
		}			
		$M[$a] = 0;                             # clear memory cell at effective address
	}	
}
elsif($opc == 5)
{
	# OR
	if($r1 != 0)
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 ORRS r1,r2
			$a  = $R[$r2] & 0xFFFE;             # effective address
        }
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 ORS r1,m or ORS r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address                		
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 ORS* r1,m or ORS* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address           			
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "OR11b");
		}	
		$M[$a] = setCRvalue($R[$r1] | $M[$a]);  # store result in effective address
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "OR11a");
	}	
}	
elsif($opc == 6)
{
	# XOR/TNM
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 TNM r1,r2
			setCRvalue($R[$r1] ^ $R[$r2]);                         
		}
		else
		{
			if($md == 1)				            # indirect register operand
			{
				# T3 XRRS r1,r2
				$a  = $R[$r2] & 0xFFFE;             # effective address
			}
			elsif($md == 2)				            # memory/indexed memory operand
			{
				# T4/T5 XRS r1,m or XRS r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a = $a & 0xFFFE;                   # effective address                		
			}
			elsif($md == 3)				            # indirect memory/indexed memory operand
			{
				# T6/T7 XRS* r1,m or XRS* r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address
			}
			$M[$a] = setCRvalue($R[$r1] ^ $M[$a]);  # store result in effective address
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "XOR11");
	}	
}
elsif($opc == 7)
{
	# multiple store
	$t1 = 1;
	if($md == 1)				            # indirect register operand
	{
		# T3A/B MSR n,r2
		# $r1 is n
		if($r2 == 0xF)
		{
			# A15, the stack pointer is addressed (adapt maybe for R[0xF] is not even **************
			$a = $R[0xF] & 0xFFFE; # effective address (save start value of $R[0xF] because a15 can be put in the stack)
			while($t1 <= $r1)
			{
				$M[$a] = $R[$A[$t1]];
				$a=$a-2;
				if($a < 0x100)
				{
					$R[0xF]=$a;
					trap(LKM);		        # A15 stack overflow
				}					
				$t1 = $t1 + 1;
			}
			$R[0xF]=$a;
		}
		else
		{
			$a = $R[$r2] & 0xFFFE;              # effective address			
			while($t1 <= $r1)
			{
				$M[$a] = $R[$A[$t1]];
				$a = $a + 2;
				$t1 = $t1 + 1;
			}
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 MS n,m or MS n,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		
		while($t1 <= $r1)
		{
			$M[$a] = $R[$A[$t1]];           # store content if effective address
			# print "MS $t1 $r1 " . prhex(16,$a) . prhex(16,$M[$a]) . "\n";
			$a = $a + 2;
			$t1 = $t1 + 1;
		}			
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 MS* n,m or MS* n,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address
		while($t1 <= $r1)
		{
			$M[$a] = $R[$A[$t1]];           # store content if effective address
			$a = $a + 2;
			$t1 = $t1 + 1;
		}				
	}
	else
	{
		# md == 0
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "MS11");
	}
}	
elsif($opc == 12)
{
	# character transfer
	if($r1 !=0)
	{
		# store
		if($md == 1)				            # indirect register operand
		{
			# T3 SCR r1,r2
			$a = $R[$r2];
			# $a is byte address
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = ($t1 & 0xFF00) | ($R[$r1] & 0xFF); # right hand byte
			}
			else
			{
				$t1 = ($t1 & 0x00FF) | ($R[$r1] << 8);   # left hand byte
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 SC r1,m or SC r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			# $a is effective byte address
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = ($t1 & 0xFF00) | ($R[$r1] & 0xFF); # right hand byte
			}
			else
			{
				$t1 = ($t1 & 0x00FF) | ($R[$r1] << 8);   # left hand byte
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 SC* r1,m or SC* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective byte address
			# $a is effective byte address
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = ($t1 & 0xFF00) | ($R[$r1] & 0xFF); # right hand byte
			}
			else
			{
				$t1 = ($t1 & 0x00FF) | ($R[$r1] << 8);   # left hand byte
			}
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "SC11b");
		}
		$M[$a & 0xFFFE] = $t1 & 0xFFFF;              # replace word
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "SC11a");	
	}
}
elsif($opc == 13)
{
	# compare characters
	if($r1 !=0)
	{	
		if($md == 1)				                # indirect register operand
		{
			# T3 CCR r1,r2, T2 CCK r1,lk
			$a  = $R[$r2];
			$t1 = $R[$r1] & 0xFF; $t2 = $M[($a & 0xFFFE)];
			if($a & 1)
			{
				$t2 = $t2 & 0xFF;                   # 1 = right hand byte
			}
			else
			{
				$t2 = ($t2 >> 8) & 0xFF;            # 0 = left hand byte
			}		
			if($r2 == 0)
			{
				# T2 indirect via A0: CCK r1,lk
				$R[0] = ($R[0] + 2);	# increment A0 (PC)
			}
		}
		elsif($md == 2)				                # memory/indexed memory operand
		{
			# T4/T5 CC r1,m or CC r1,m,r2
			$a = $M[$R[0]];			                # memory operand address
			$R[0] = ($R[0] + 2);	                # increment A0 (PC)
			if($r2 != 0)			                # indexing ?
			{
				$a = ($a + $R[$r2]);                # yes
			}
			$t1 = $R[$r1] & 0xFF; $t2 = $M[($a & 0xFFFE)]; 
			if($a & 1)
			{
				$t2 = $t2 & 0xFF;                   # 1 = right hand byte
			}
			else
			{
				$t2 = ($t2 >> 8) & 0xFF;            # 0 = left hand byte
			}				
		}
		elsif($md == 3)				                # indirect memory/indexed memory operand
		{
			# T6/T7 CC* r1,m or CC* r1,m,r2
			$a = $M[$R[0]];			                # memory operand address
			$R[0] = ($R[0] + 2);	                # increment A0 (PC)
			if($r2 != 0)			                # indexing ?
			{
				$a = ($a + $R[$r2]);                # yes
			}
			$a = $M[$a & 0xFFFE];                   # memory content = effective address
			$t1 = $R[$r1] & 0xFF; $t2 = $M[($a & 0xFFFE)];  
			if($a & 1)
			{
				$t2 = $t2 & 0xFF;                   # 1 = right hand byte
			}
			else
			{
				$t2 = ($t2 >> 8) & 0xFF;            # 0 = left hand byte
			}				
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "CC11b");			
		}		
		# t1 and t2 contain the characters to compare
		$t3 = $t1 - $t2; 
		setCRvalue($t3);                        # set condition register
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "CC11a");			
	}		
}
elsif($opc == 14)
{
	if($r1 != 0)
	{
		# call function
		if($md == 0)				                    # register operand
		{
			# T1 CFR r1,r2
			if($r1 != 0xF)                              # not A15
			{
				$M[$R[$r1]] = $R[0];                    # save P
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $R[$r2] & 0xFFFE;
			}
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $R[$r2] & 0xFFFE;				
			}
		}
		elsif($md == 1)				                    # indirect register operand
		{
			# T2/T3 CF r1,lk   CFR* r1,r2
			$t1 =  $M[$R[$r2]& 0xFFFE] & 0xFFFE;        # save new P value
			if($r2 == 0)
			{
				# T2 CF r1,lk
				$R[0] = ($R[0] + 2);                    # increment A0 (PC)
			}
			if($r1 != 0xF)    			
			{
				$M[$R[$r1]] = $R[0];                    # save P on stack
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW on stack
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $t1;                            # jump
			}				
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P on stack
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW on stack
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $t1;		                    # jump		
			}
		}
		elsif($md == 2)				                    # memory/indexed memory operand
	    {	
			# T4/T5 CFI r1,m or CFI r1,m,r2
			$a = $M[$R[0]];		                        # memory operand address
			$R[0] = ($R[0] + 2);                        # increment A0 (PC)
			if($r2 != 0)			                    # indexing ?
			{
				$a = ($a + $R[$r2]);                    # yes
			}
			$a = $a & 0xFFFE;                           # effective address
			if($r1 != 0xF)    			
			{
				$M[$R[$r1]] = $R[0];                    # save P
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $M[$a];
			}				
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $M[$a];				
			}		
		}
		elsif($md == 3)				                    # indirect memory/indexed memory operand
		{
			# T6/T7 CFI* r1,m or CFI* r1,m,r2
			$a = $M[$R[0]];			                    # memory operand address
			$R[0] = ($R[0] + 2);	                    # increment A0 (PC)
			if($r2 != 0)			                    # indexing ?
			{
				$a = ($a + $R[$r2]);                    # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;              # memory content = effective address			
			if($r1 != 0xF)    			
			{
				$M[$R[$r1]] = $R[0];                    # save P
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $M[$a];
			}				
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $M[$a];				
			}		
		}
	}
	else
	{
		# execute instructions
		my $t1;
		my $exerr=0;
		my $il=2;    # default instruction length
		# execute
		if($md == 0)
		{
			$t1=$R[$r2];						# instruction to execute
		}
		elsif($md == 1)
		{
			# T3 EXR* r2, T2 EXK lk
			$t1 = $M[($R[$r2] & 0xFFFE)];		# instruction to execute
			if($r2 == 0)
			{
				# T2 indirect via A0: EXK lk
				$R[0] = ($R[0] + 2);         	# increment A0 (PC)	
				$il=4;
			}
		}
		elsif($md == 2)		
		{
			# T4/T5 EX m or EX m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			$il=4;
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t1 = $M[$a]; 						# instruction to execute	
		}
		elsif($md == 3)
		{
			# T6/T7 EX* m or EX* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			$il=4;
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		    $a = $M[$a & 0xFFFE];               # memory content = effective address
			$a = $a & 0xFFFE;                   # effective address
			$t1 = $M[$a]; 						# instruction to execute				
		}
		
		# execute t1 if:
		# - it is a single word instruction
		# - it is not a CF.., RTN, EX..
		#
		# check instruction to execute (format0 needs no checking)
		if($t1 & 0x8000)
		{
			# format 1, check whether it is an allowed single word instruction 
			my $opc = ($t1 >> 11) & 0xF;            # opcode
			my $r1  = ($t1 >> 7)  & 0xF;            # r1 or condition shifted 1 left	
			my $r2  = ($t1 >> 1)  & 0xF;            # r2
			my $md  = ($t1 >> 5)  & 0x3;            # addressing mode
			if($opc == 0xE || $md == 2 || $md == 3 || $r2 == 0)
			{
				# illegal instruction
				ilcerror($R[0]-$il, $t1, $opc, "EX11");
				$exerr=1;
				# in case the illegal instruction is not trapped it
				# operates as a no operation to ease debugging
			}
		}
		#print "Execute1" . prhex(16,$t1) . "\n";
		# execute $t1
		if($exerr ==0)
		{
			$IR=$t1;
			if(! ($t1 & mskf10))
			{
				#print "format 0\n";
				format0();
			}
			elsif(!($t1 & 1))
			{
				#print "format 10\n";				
				format10();
			}
			else
			{
				#print "format 11\n";				
				format11();
			}
		}
	}
}
elsif($opc == 15)
{
	# 1-complement
	if($r1 == 0)
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 C1RS r2
			$a  = $R[$r2];                      # effective address
        }
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 C1S m or C1S m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}               		
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 C1S* m or C1S* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address           			
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "C111b");
		}
		$a = $a & 0xFFFE;                       # effective address 
		$M[$a] = setCRvalue(~ $M[$a]);          # store result in effective address
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "C111a");
	}	
}	
else
{
	# illegal instruction
	ilcerror($R[0]-2, $IR, $opc,"OPC11");
}

} # end format11

sub format0 {
#
# emulate the format 0 instructions:
# 	- short constant instructions
# 	- shift instructions
# 	- i/o instructions
# 	- miscellanious instructions
$opc = ($IR >> 11) & 0xF;	# opcode
$r3  = ($IR >> 7)  & 0xE;   # r3 or cnd (shifted 1 left) 
$k   = $IR & 0xFF;          # shortconstant
				
if($opc == 0)
{
	# load constant LDK
	$R[$r3] = $k;			# CR unchanged !!!!
}
elsif($opc == 1)
{
	# absolute branche AB
	# $r3 is the condition code in operation
	$r3 = $r3 >> 1; # position the condition code	
	if(evalcond($r3))
	{
		$R[0] = $k;
	}
}
elsif($opc == 2)
{
	# add constant ADK
	$t1 = $R[$r3];
	$t2 = $t1 + $k;
	# overflow? (+)
	if((((~$t1) ^ $k) & ($t1 ^ $t2)) & SIGN)
	{
		$overflow = 1;
	}
	$R[$r3] = setCRvalue($t2);
}
elsif($opc == 3)
{
	# subtract constant SUK
	$t1 = $R[$r3];
	$t2 = $t1 - $k;
	# underflow? (-)
	if((($t1 ^ $k) & ((~$k) ^ $t2)) & SIGN)
	{
		$overflow = 1;
	}
	$R[$r3] = setCRvalue($t2);
}
elsif($opc == 4)
{
	# logical and constant ANK - MIS: HLT, INH, RIT
	if($r3 != 0)
	{
		# ANK
		$R[$r3] = $R[$r3] & $k;
		setCR1($r3);
	}
	else
	{
		if($k == 0x7F)
		{
			# HLT
			$RUN = 0;
			prlog("\nExecution stopped with a HLT\n");
		}
		elsif($k == 0xBF)
		{
			# INH
			$ENB = 0;
		}
		else
		{
			my $t1 = $k & 0xC1;
			if($t1 == 0xC1) # is it an RIT
			{
				# RIT
				# resetting more internal interrupts in a single instruction is allowed
				$t1 = ($k & 0x3F) >> 1; # 5 bits
				$t1 = ~ $t1;
				if($t1 & 4) # 0x1B
				{
					resettrap(RTC);     # reset clock interrupt
				}
				if($t1 & 2) # 0x1D
				{
					resettrap(LKM);     # reset LKM interrupt
				}
				if($t1 & 16) # 0x0F
				{
					resettrap(CP);      # reset control panel interrupt
				}
				if($t1 & 8) # 0x17
				{
					resettrap(PWR);     # reset power failure interrupt
				}
				# ignore other bits
			}
			else
			{
				# illegal instruction
				ilcerror($R[0]-2, $IR, $opc, "F04");
			}
		}
	}
}
elsif($opc == 5)
{
	# logical or constant ORK - MIS: LKM ENB SMD
	if($r3 != 0)
	{
		# ORK
		$R[$r3] = $R[$r3] | $k;
		setCR1($r3);
	}
	else
	{
		if($k & 0x04)
		{
			# LKM
            if($ENB)
            {
                trap(LKM);
            }
            else
            {
                $R[0] = ($R[0] + 2); # skip DATA to execute next instruction
            }
		}
		if($k & 0x40)
		{
			# ENB
			$ENB = 1;
		}
		if($k & 0x01)
		{
			# SMD
			$UM  = 1;  # Bit is set, but no usermode implemented
		}
        # ignore setting of other bits
	}
}
elsif($opc == 6)
{
	# excl or constant XRK
	if($r3 != 0)
	{
		# XRK
		$R[$r3] = $R[$r3] ^ $k;
		setCR1($r3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "F06");
	}
}
elsif($opc == 7)
{
	# shift: single/double word shift, normalize
	my $tlng;
	my $i;
	$sc   = $k & 0x1F;                      # shift count (5 bit)
	$sfun = ($k & 0xE0) >> 5;               # shift function (3 bit)
	$t1   = $sc;
	if($r3 != 0)
	{
		# single register shifts
				
		if($sfun == 0)
		{
			# SLA: single left arithmetic shift
			if($t1 > 16) { $t1 = 16 };      # limit to 16 
			$t2 = $R[$r3] & SIGN;           # save sign
			$i = 0;
			while($i < $t1)
			{
				$R[$r3] = $R[$r3] << 1;
				if(($R[$r3] & SIGN) != $t2)
				{
					$overflow = 1;
				}
				$i = $i + 1;
			}
			setCR1($r3);
		}
		elsif($sfun == 1)
		{
			# SRA: single right arithmetic shift
			if($t1 > 16) { $t1 = 16 };      # limit to 16 
			$t2 = $R[$r3] & SIGN;           # save sign
			$i=0;
			while($i < $t1)
			{
				$R[$r3] = $R[$r3] >> 1;
				$R[$r3] = $R[$r3] | $t2;
				$i = $i + 1;
			}
			setCR1($r3);
		}
		elsif($sfun == 2)
		{
			# SLL: single left logical shift
		    if($t1 > 16) 
			{ 
				$R[$r3] = 0;               # > 16 ? all 0 
			}
			else
			{
				$R[$r3] = ($R[$r3] << $t1) & DMASK;
			}
			setCR1($r3);
		}
		elsif($sfun == 3)
		{
			# SRL: single right logical shift
			if($t1 > 16) 
			{ 
				$R[$r3] = 0;               # > 16 ? all 0 
			}
			else
			{
				$R[$r3] = ($R[$r3] >> $t1) & DMASK;
			}
			setCR1($r3);
		}
		elsif($sfun == 6)
		{
			# SLC: single left circular shift
			$t2 = $t1 % 16;               # mod 16, t2 is remainder
			$R[$r3] = (($R[$r3] << $t2) | ($R[$r3] >> (16 -$t2)) & DMASK);
			setCR1($r3);
		}
		elsif($sfun == 7)
		{
			# SRC: single right circular shift
			$t2 = $t1 % 16;               # mod 16, t2 is remainder
			$R[$r3] = (($R[$r3] >> $t2) | ($R[$r3] << (16 -$t2)) & DMASK);
			setCR1($r3);
		}
		elsif($sfun == 4)
		{
			# SLN: single left normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$t2 = $R[$r3] & SIGN;          # save sign (bit 0)
			$t3 = $t2 >> 1;                # sign shifted 1 right for comparison
			$i = 0;
			while($i < 16)
			{
				$t1 = $R[$r3] & 0x4000;    # extract bit 1
				if($t3 == $t1)             # bit 0 equal to bit 1?
				{
					$R[$r3] = $R[$r3] << 1; # yes => shift left
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$R[$r3] = ($R[$r3] & MMASK) | $t2;
			$R[$r2] = $i;
		}
		elsif($sfun == 5)
		{
			# SRN: single right normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$t2 = $R[$r3] & SIGN;          # save sign
			$i = 0;
			while($i < 16)
			{
				$t1 = $R[$r3] & 1;
				if($t1 == 0)
				{
					$R[$r3] = ($R[$r3] >> 1) | $t2;
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$R[$r3] = ($R[$r3] & MMASK) | $t2;
			$R[$r2] = $i;
		}
	}
	else
	{
		# double register shifts (A1-A2)
				
		if($sfun == 0)
		{
			# DLA: double left arithmetic shift
			$tlng = getdbls();				# get A1, A2 as long
			$t2 = $tlng & DPSIGN;           # save sign
			$i = 0;
			while($i < $t1)
			{
				$tlng = $tlng << 1;
				if(($tlng & DPSIGN) != $t2)
				{
					$overflow = 1;
				}
				$i = $i + 1;
			}
			setCR2s($tlng);	                 # set condition code
			putdbls($tlng);                  # restore A1, A2
		}
		elsif($sfun == 1)
		{
			# DRA: double right arithmetic shift
			$tlng = getdbls();				# get A1, A2 as long
			$t2 = $tlng & DPSIGN;           # save sign
			$i=0;
#			if($t1 != 0) # special case for 0 shifts: keep A1/A2 as is (from testprogram CPB8K)
#           the P852 in case of 0 shifts did not change bit-0 of A2 (if it is 1, it remains 1)
#           the P854 (and later models) makes bit-0 of A2 always 0 also in case of zero shifts.
#           This implementation follows approach (at this moment): bit-0 A2 is always made 0
#			{
				while($i < $t1)
				{
					$tlng = $tlng >> 1;
					$tlng = $tlng | $t2;
					$i = $i + 1;
				}		
				putdbls($tlng);                  # restore A1, A2
#			}
			setCR2s($tlng);
		}
		elsif($sfun == 2)
		{
			# DLL: double left logical shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = $tlng << $t1;
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 3)
		{
			# DRL: double right logical shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = $tlng >> $t1;
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 6)
		{
			# DLC: double left circular shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = ($tlng << $t1) | ($tlng >> (32 - $t1));
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 7)
		{
			# DRC: double right circular shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = ($tlng >> $t1) | ($tlng << (32 - $t1));
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 4)
		{
			# DLN: double left normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$tlng = getdbls();				 # get A1, A2 as long
			$t2 = $tlng & DPSIGN;            # save sign
			$t3 = $t2 >> 1;                  # sign shifted 1 right for comparison
			$i = 0;
			while($i < 31)
			{
				$t1 = $tlng & 0x20000000;
				if($t3 == $t1)
				{
					$tlng = $tlng << 1;
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$tlng = ($tlng & DPMMASK) | $t2;
			putdbls($tlng);                  # restore A1, A2
			$R[$r2] = $i;
		}
		elsif($sfun == 5)
		{
			# DRN: double right normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$tlng = getdbls();				 # get A1, A2 as long
			$t2 = $tlng & DPSIGN;            # save sign
			$i = 0;
			while($i < 31)
			{
				$t1 = $tlng & 1;
				if($t1 == 0)
				{
					# $R[$r3] = ($R[$r3] >> 1) | $t2; # fout
					$tlng = ($tlng >> 1) | $t2;
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$tlng = ($tlng & DPMMASK) | $t2;
			putdbls($tlng);                  # restore A1, A2
			$R[$r2] = $i;
		}
	}		
}
elsif($opc == 8)
{
	# IO: CIO OTR
	# 
	if($r3 !=0)
	{
		my $bit8 = $k & 0x80;
		my $bit9 = $k & 0x40;                 # function bit
		my $devaddr = $k & 0x3F;
		my $d = $DEV{$devaddr};               # pointer to device block
		if($d eq undef)
		{
			# unknown device address
			$CR = 3;
		}
		elsif($bit8 == 0x80)
		{
			# CIO
			$n = $$d[CIO];                    # CIO routine of device
			&$n($bit9);                       # call
		}
		else
		{
			# OTR
			$n = $$d[OTR];                    # OTR routine of device
			&$n($bit9);                       # call
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "IO1");
	}
}
elsif($opc == 9)
{
	# IO: INR SST TST 
	if($r3 !=0)
	{
		my $bit8 = $k & 0x80;
		my $bit9 = $k & 0x40;
		my $devaddr = $k & 0x3F;
		my $d = $DEV{$devaddr};               # pointer to device block
		if($d eq undef)
		{
			# unknown device address
			$CR = 3;
		}
		elsif($bit8 == 0)
		{
			# INR
			$n = $$d[INR];                    # INR routine of device
			&$n($bit9);                       # call			
		}
		elsif($bit9 == 0)
		{
			# TST
			$n = $$d[TST];                    # TST routine of device
			&$n();                            # call			
		}
		else
		{
			# SST
			$n = $$d[SST];                    # SST routine of device
			&$n();                            # call	
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "IO2");
	}
}
elsif($opc == 0xA)
{
	# RF
	# $r3 (4 bit) is the condition code in operation >> 1
	$r3 = $r3 >> 1;
	if(evalcond($r3))
	{
		$R[0] = ($R[0] + $k);
	}
}
elsif($opc == 0xB)
{
	# RB
	# $r3 (4 bit) is the condition code in operation >> 1
	$r3 = $r3 >> 1;
	if(evalcond($r3))
	{
		$R[0] = ($R[0] - $k);
	}
}
elsif($opc == 0xE)
{
	# WER
	if($r3 != 0)
	{
		$EXTR[$k]=$R[$r3];
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "F0E");
	}	
}
elsif($opc == 0xF)
{
	# RER
	if($r3 != 0)
	{
		$R[$r3]=$EXTR[$k];
		$CR = $R[$r3] >> 8;
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "F0F");
	}	
}
else
{
	# illegal instruction
	ilcerror($R[0]-2, $IR, $opc, "OPC00");	
}
} # end format0

########################### emulator subroutines #########################
# Evaluate the Condition code in an operation against the current CR value
# If true, the function returns a 1, else 0.
sub evalcond {
	my $cd = shift;  	# condition code (3 bit) from operation
	my $not;
	if($cd == 7)
	{
		return 1;	# unconditional
	}
	$not  = $cd & 4;
	$cd   = $cd & 3;
	if($not)        # not bit
	{
		if($cd == $CR)
		{
			return 0;
		}
		else
		{
			return 1;
		}
	}
	else
	{
		if($cd == $CR)
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}
}

# set Condition Register according value of r1 or r3 (or temporary register in @R)
sub setCR1 {
	my $r = shift;
	my $t = $R[$r];
	if($overflow)
	{
		$CR = 3;                 # overflow
	}
	elsif($t & 0x8000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($t & 0x7FFF)
	{
		$CR = 1;                 # result GT 0
	}
	else
	{
		$CR = 0;                 # result = 0
	}
	$R[$r] = $t & 0xFFFF;
	$overflow = 0;
}

# set Condition Register according  a given value
# the masked 16-bit value is reurned
sub setCRvalue {
	my $t = shift;
	if($overflow)
	{
		$CR = 3;                 # overflow
	}
	elsif($t & 0x8000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($t & 0x7FFF)
	{
		$CR = 1;                 # result GT 0
	}
	else
	{
		$CR = 0;                 # result = 0
	}
	$overflow = 0;
    return $t & 0xFFFF;
}

# set Condition Register according value of A1, A2 as 31 bit signed long integer
# bit 1 (A1 bit 0) is sign bit (bit 0 A2 excluded)
# to be used in combination of getdbls/putdbls
sub setCR2s {
	my $l = shift;
	if($overflow)               # DLA + arith ops like double add, etc.
	{
		$CR = 3;                 # overflow
	}	
	elsif($l & 0x40000000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($l & 0x3FFFFFFF)
	{
		$CR = 1;                 # result GT 0
	}
	else
	{
		$CR = 0;                 # result = 0
	}
	$overflow = 0;
}

# set Condition Register according value of A1, A2 as 32 bit signed long integer
# bit 0 (A1 bit 0) is sign bit
# to be used in combination of getdblu/putdblu
sub setCR2u {
	my $l = shift;
	if($l & 0x80000000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($l & 0x7FFFFFFF)
	{
		$CR = 1;                 # result GT 0
	}
	else
	{
		$CR = 0;                 # result = 0
	}
}

# In the following 4 routines A1 an A2 are accessed as long accumulator.
# !!!!!! A1 = $R[2] A2=$R[4] !!!!!!!!
# get double signed
# returns A1, A2 as signed 31 value
sub getdbls {
	return (($R[2] << 15) | ($R[4] & MMASK));
}

# get double unsigned
# returns A1, A2 as unsigned 32 value
sub getdblu {
	return (($R[2] << 16) | $R[4] );
}

# put double signed
# stores signed 31 bit value in A1, A2
sub putdbls {
	my $x = shift;
	$x = $x & 0x7FFFFFFF;
	$R[2] = ($x >> 15) & DMASK;
	$R[4] = ($x & MMASK);
}

# put double unsigned
# stores unsigned 32 bit value in A1, A2
sub putdblu {
	my $x = shift;
	$R[2] = ($x >> 16) & DMASK;
	$R[4] = $x & DMASK;
}

# set internal interrupt (which must also be reset lateron by the program with an RIT , like LKM)
sub trap {
	my $int = shift;
	my $pc = $R[0] - 2;
	$inint = $inint | $int; 				# set internal interrupt
	$intreg = $intreg | $II2PI[$int];		# set priority interrupt
	if($DEBUG)
	{
		prlog("\nTrap" . prhex(8,$int) . " at" . prhex(16,$pc) . "\n");
		prlog("Opcode =" . prhex(4,$opc) . " Instruction =" . prhex(16,$IR) . "\n");
		trace($pc);
	}
}

# reset internal interrupt (which is not automatically reset)
sub resettrap {
	my $int = shift;
	$inint  = $inint & (~ $int);          # reset internal interrupt
	$intreg = $intreg & (~ $II2PI[$int]); # reset priority interrupt
}

# Invalid instruction trap
# this trap should also be generated in case of:
#  - a privileged instruction is executed in User Mode (not implemented)
#  - A15 stack is used in User Mode (not implemented)
#
# execution of the instruction is stopped; PSW and the address of the invalid 
# instruction are stored in the stack.
# the internal interrupt is automatically reset and an indirect jump is made
# to address /7E
#
sub trap7E {
	my $pc=shift; # address of trapped instruction

	if($DEBUG)
	{
		prlog("\nTrap-7E at" . prhex(16,$pc) . "\n");
		trace($pc);
	}
	# instruction address and PSW on stack
	$M[$R[A15]] = $pc;                      # save address
	$R[A15] = $R[A15] - 2;                  # decrease link stack
	if($R[A15] < 0x100)
	{
		trap(LKM);		                    # A15 stack overflow
	}				
	$M[$R[A15]] = Fpsw();                   # save PSW
	$R[A15] = $R[A15] - 2;                  # decrease link stack
	if($R[A15] < 0x100)
	{
		trap(LKM);		                    # A15 stack overflow
	}											
	$ENB = 0;                               # inhibit
	$UM  = 0;                               # system mode
    # $PL=0x3F;                             # level is NOT changed
	$R[A0] = $M[0x7E];                      # continue with the trap routine
}

# Illegal instruction 
sub ilcerror {
	my $lc    = shift;
	my $instr = shift;
	my $opc   = shift;
	my $ident = shift;
	if($DEBUG)
	{
		prlog("\nERROR: Illegal Instruction at" . prhex(16,$lc) . " IR:" . prhex(16,$instr) . " OPCODE:" . prhex(4,$opc) . " $ident\n");
    }
	if($StopOnIllInstr)
	{
		prlog("\nERROR: Illegal Instruction at" . prhex(16,$lc) . " IR:" . prhex(16,$instr) . " OPCODE:" . prhex(4,$opc) . " $ident\n");
		$RUN = 0;
	}
	else
	{
		trap7E($lc);      # trap illegal instruction
	}
}

# Illegal IOP channel condition
sub mpxerror {
	if($StopOnMpxError)
	{
		trace($pcsaved);
		$RUN=0;
	}
}

# format PSW
sub Fpsw {
#	my $PL;             # priority level 			psw bit 0:5
#	my $CR;             # condition register		psw bit 6:7
#	my $overflow=0;     # (aid to set CR, must be 0)
#	my $RUN;            # run bit               	psw bit 8
#	my $ENB;            # interrupt enable bit		psw bit 9
#	my $CP;             # control panel interrupt	psw bit 10      X
#	my $PWR;            # power fail interrupt      psw bit 11      X
#	my $RTC;            # real time clock interrupt psw bit 12      X
#	my $LKM;            # link monitor interrupt    psw bit 13      X
#						# + stack overflow
#	-------                                         psw bit 14      X (not used)
#	my $UM;             # user mode indicator       psw bit 15
#                       ########### X bits are part of $inint and $intreg

my $i = $inint & 0x1F;      # isolate the 5 internal interrupts represented in the PSW
$PSW = ((((((((($PL << 2) | $CR) << 1) | $RUN) << 1) | $ENB) << 5) | $i) << 1) | $UM;
return $PSW;
}

sub reset_cpu {
	$PL  = 0x3F;     #??? or is the reset value 0 ??? NO, on P856 /3F is OK
	$CR  = 0;
	$overflow = 0;
	$RUN = 0;
#####################################################################################
	$ENB = 0;        #??? should be 1 after reset (however for testing it is set to 0)#
									 #in case ENB == 1 after reset, system does not boot dos5 from md1#
#####################################################################################
	$UM  = 0;        # system mode
	$PSW = 0;
	$intreg = 0;  # priority interrupts
	$inint  = 0;  # internal interrupts
	$brkreg = 0;  # channel breaks
	# all registers, except A0 to 0
	$R[1]  = 0;
	$R[2]  = 0;	
	$R[3]  = 0;	
	$R[4]  = 0;
	$R[5]  = 0;	
	$R[6]  = 0;		
	$R[7]  = 0;
	$R[8]  = 0;	
	$R[9]  = 0;	
	$R[10] = 0;
	$R[11] = 0;	
	$R[12] = 0;	
	$R[13] = 0;
	$R[14] = 0;	
	$R[15] = 0;	
#########################################################
# as a debugging aid, A15 is set to a predefined value: #
	$R[15] = 0x1FE;                                     #
# for the same reason ENB is set to 0 and PL to /3F     #
#########################################################
}

# the memory array M should only be used for even addresses
# in case an oneven address contains a value, there is an error
# in the simulators memory access
# on even address, also 1-bits outside the lower 16 part of each
# word should not be there
sub mem_check {
	my $fl = shift;
	my $b = 0;
	my $i = 0;
	while($i < MEMSIZE)
	{
		if($i & 1)
		{
			# uneven address
			if($M[$i] ne undef)
			{
				prlog("ERROR! value" . prhex(16,$M[$i]) . " is found at the uneven address" . prhex(16,$i) . "\n");
				$b=1;
			}
		}
		else
		{
			# even address
			if($M[$i] & 0xFFFF0000)
			{
				prlog("ERROR! wrong bits" . prhex(32,$M[$i]) . " found at the even address" . prhex(16,$i) . "\n");
				$b=1;
			}
		}
		$i = $i + 1;
	}
	$i=0;
	while($i < 0x10)
	{
		if($R[$i] & 0xFFFF0000)
		{
			prlog("ERROR! wrong bits" . prhex(32,$R[$i]) . " found in Register $Reg[$i] \n");
			$b=1;
		}
		$i = $i + 1;
	}
	if($fl)
	{
		prlog("Memory and Registers Checked\n");
	}
	return $b;
}

###################### Controller subroutines #################################

sub reset_all {
	$program_started = 0;
	reset_cpu();
	reset_clk();
	reset_tty();
	reset_ptr();
	reset_ptp();
	reset_lpt();
	reset_crd();
	reset_md0();
}

# trace can be used in two ways:
# 1) with pc=A0/P it displays the state of the cpu
# 2  with pc=location of an executed instruction it provides the state of
#    the cpu after the execution of that instruction
my $pctrace = 0;
my $trrep   = 0;
sub trace {
	my $pc = shift;
	my $psw = Fpsw();
    if($tracefile) # dump trace in file ?
	{
		if($pc != $pctrace)
		{
			$pctrace=$pc;
			$trrep=0;
			print TRACE decode() . "\n";
			print TRACE "  P" . prhex(16,$pc) . "    M" . prhex(16,$M[$pc]) . "  PSW" . prhex(16,$psw);
			print TRACE " CR" . prhex(4,$CR); 
			print TRACE " IR" . prhex(32,$intreg) . " IN" . prhex(16,$inint) . "   BR" . prhex(16,$brkreg);
			print TRACE " A1";
			print TRACE prhex(16,$R[2]);
			print TRACE prhex(16,$R[4]);
			print TRACE prhex(16,$R[6]);
			print TRACE prhex(16,$R[8]);
			print TRACE prhex(16,$R[10]);
			print TRACE prhex(16,$R[12]);
			print TRACE prhex(16,$R[14]);
			
			print TRACE prhex(16,$R[1]);
			print TRACE prhex(16,$R[3]);
			print TRACE prhex(16,$R[5]);
			print TRACE prhex(16,$R[7]);
			print TRACE prhex(16,$R[9]);
			print TRACE prhex(16,$R[11]);
			print TRACE prhex(16,$R[13]);
			print TRACE prhex(16,$R[15]);
			print TRACE "\n";
		}
		else
		{
			if($trrep < 5)
			{
				print TRACE "=\n";
			}
			$trrep=$trrep+1;
		}
	}
	prlog(decode() . "\n");
	prlog("  P" . prhex(16,$pc) . "    M" . prhex(16,$M[$pc]) . "  PSW" . prhex(16,$psw));
	prlog(" CR" . prhex(4,$CR)); 
	prlog(" IR" . prhex(32,$intreg) . " IN" . prhex(16,$inint) . "   BR" . prhex(16,$brkreg) . "\n");
	prlog(" A1");
	prlog(prhex(16,$R[2]));
	prlog(prhex(16,$R[4]));
	prlog(prhex(16,$R[6]));
	prlog(prhex(16,$R[8]));
	prlog(prhex(16,$R[10]));
	prlog(prhex(16,$R[12]));
	prlog(prhex(16,$R[14]));

	prlog(prhex(16,$R[1]));
	prlog(prhex(16,$R[3]));
	prlog(prhex(16,$R[5]));
	prlog(prhex(16,$R[7]));
	prlog(prhex(16,$R[9]));
	prlog(prhex(16,$R[11]));
	prlog(prhex(16,$R[13]));
	prlog(prhex(16,$R[15]));
	prlog("\n");
}

# trace instructions in a region
sub rtrace {
	my $pc = shift;
	if($pc >= $batrace && $pc <= $eatrace)
	{
		trace($pc);
	}
}

######################### Simulator event handler
sub activate_event {
	my $d = shift;		# device for which event is started
	my $t = shift;      # event time out value
	my $cur;
	my $prev = 0;
	my $acc = 0;        # acccumulated time of active events
	
    if($t < 0) { return -1 };                # error
	if(is_active($d)) { return 0 };          # already active
	
	$time = $time + ($currentinterval - $interval);
	# scan event queue (= linked activated devices)
	for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
	{
		if($t < ($acc + $$cur[TIMEOUT]))
		{
			last;
		}
		$acc = $acc + $$cur[TIMEOUT];
		$prev = $cur;
	}
	if($prev == 0)       # insert new event at head of the queue
	{
		$cur = $$d[NEXT] = $event_queue;   # = 0
		$event_queue = $d;                 # = head
	}
	else                 # insert at previous in the queue
	{
		$cur = $$d[NEXT] = $$prev[NEXT];   # = 0
        $$prev[NEXT] = $d;
	}
	$$d[TIMEOUT] = $t - $acc;
	if($cur != 0)
	{
		# make time of next element relative to new one
		$$cur[TIMEOUT] = $$cur[TIMEOUT] - $$d[TIMEOUT];	
	}
	$interval = $$event_queue[TIMEOUT];
	$currentinterval = $interval;
	return 0;
}

sub cancel_event {
    my $d = shift;                # event t o cancel
	my $cur;
	my $next = 0;
	
    if($event_queue == 0) { return 0 };	
	$time = $time + ($currentinterval - $interval);
	if($event_queue == $d)
	{
		$next = $event_queue = $$d[NEXT];
	}
	else
	{
		for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
		{
			if($$cur[NEXT] = $d)
			{
				$next = $$cur[NEXT] = $$d[NEXT];
				last;
			}
		}
	}
	if($next != 0)
	{
		$$next[TIMEOUT] = $$next[TIMEOUT] + $$d[TIMEOUT];
	}
	$$d[0] = 0;
	if($event_queue != 0)
	{
		$interval = $$event_queue[TIMEOUT];
	}
	else
	{
		$interval = $queue_empty_time;
	}
	$currentinterval = $interval;
	return 0;
}

sub is_active {
	my $d = shift;
	my $acc = 0;
	my $cur;
	for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
	{
		if($cur == $event_queue)
		{
			if($interval > 0)
			{
				$acc = $acc + $interval;
			}
		}
		else
		{
			$acc = $acc + $$cur[TIMEOUT];
		}
		if($cur == $d)
		{
			return $acc + 1;
		}
	}
	return 0;
}

sub process_events {
	my $d;
	my $err;
	
	$time = $time + $currentinterval;
	if($event_queue == 0)
	{
		$interval = $queue_empty_time;
		$currentinterval = $interval;
		$err = 0;
	}
	else
	{
		do
		{
			$d = $event_queue;          # remove 1st element from queue
			$event_queue = $$d[NEXT];   # set next
			$$d[NEXT] = 0;
			if($event_queue != 0)
			{
				$interval = $$event_queue[TIMEOUT]; # set new interval
			}
			else
			{
				$interval = $queue_empty_time;
			}
			$currentinterval = $interval;
			if($$d[SERVICE] != 0)
			{
				my $sub = $$d[SERVICE];
				$err = &$sub;           # execute event service routine
			}
			else
			{
				$err = 0;
			}
			# print "p";
		}
		while(($err == 0) && ($interval == 0));
	}
	return $err;
}

sub show_event_queue {
	my $cur;
	my $d;
	my $t;
	prlog("Time: $time Current interval: $currentinterval Actual interval: $interval\n");
	for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
	{
		$d = $$cur[NAME];
		$t = $$cur[TIMEOUT];
		prlog("Device $d Current event time $t\n");
	}
}

######################### CLK device routines
# Real time clocks were delivered with several frequencies (1 ms, 10 ms, etc.),
# depending on the applied cristal. The default was a line frequency clock(20 ms or
# 16.6 ms). 
# On P856 the clock is started/stopped with a key on the panel (RTC ON/OFF)

sub cio_clk {
	my $f = shift;                  # function bit
	my $stat = $CLK[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if($stat == EXE)
		{
			$CLK[STATE] = $CLK[STATE] | (WREQ << 16);     # state = wait state requested
		}
		else
		{
			 $CLK[STATE] = INAC;
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START, accepted if state == inactive
		# F=1 => CIO start clock
		$CLK[STATE] = (EXE << 16);                    # state = execute	
		$CLK[BUFFER] = 0;                             # reset sync period
		$CLK[CTIME] =  $CLK[TIME];                    # start with default interval
		activate_event(\@CLK,$CLK[CTIME]);            # schedule event
		$CR = 0;		
	}
	else
	{
		$CR = 1;                    # not accepted                    
	}	
}

# Clock service routine 
# processes the 'execute state' of the simulated clock device.

sub svc_clk {
	my $stat = ($CLK[STATE] >> 16);	
	if($stat == EXE)
	{
		trap(RTC);                                     # set the ready flipflop
		###### clock sync code
        if($CLK[BUFFER] == 0)                          # new sync period ?
		{
			@clk0 = gettimeofday ();                   # fix start point
		}
		elsif($CLK[BUFFER] == $CLK[SYNC])
		{
		#	# sync period ended
			$laps = tv_interval (\@clk0);              # elapsed sync time in secs
			$pulses = $CLK[CTIME] * $CLK[SYNC];
			# print "$laps secs $pulses pulses\n";
		#	# new interval time
			$CLK[CTIME] = int(($pulses / $laps) / $CLK[FREQ]);
		#	# start a new sync period
			@clk0 = gettimeofday ();                   # fix new start point
			$CLK[BUFFER] = 0;
	    }
		$CLK[BUFFER] = $CLK[BUFFER] + 1;               # sync  period
		activate_event(\@CLK,$CLK[CTIME]);             # schedule a new event
		
	}
	else
	{
		prlog("Clock switched off\n");
		# switch directly to inactive state 
		$CLK[STATE] = INAC;                  # cu = inactive
	}
}

sub inr_clk {
	return 0; # nop
}

sub otr_clk {
	return 0; # nop
}

sub sst_clk {
	return 0; # nop
}

sub tst_clk {
	if( ($CLK[STATE] >> 16) )
	{
		$R[$r3] = 1;                                   # busy, not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}	
	$CR = 0;
}

sub state_clk {
		prlog("CU" . prhex(20,$CLK[STATE]) . " INTERVAL $CLK[CTIME]\n");
}

sub reset_clk {
	resettrap(RTC);                                    # reset the ready flipflop
	$CLK[STATE]  = INAC << 16;                         # state = inactive + 0
	$CLK[BUFFER] = 0;                                  # reset buffer
	$CLK[CTIME]  = 0;
}
		
######################### TTY device routines

sub cio_tty {
	my $f = shift;                  # function bit
	my $stat = $TTY[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == IEXE) || ($stat == OEXE) || ($stat == IEXC) || ($stat == OEXC))
		{
			$TTY[STATE] = $TTY[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START, accepted if state == inactive
		if(($R[$r3] & 1) == 0)
		{
			# OUTPUT
			$TTY[STATE] = (OEXC << 16);        # state = reset + output + exchange
			$intreg = $intreg | IASR;          # set the ready flipflop
		}
		else
		{
			# INPUT
			# tty event is already scheduled
		    $TTY[STATE] = (IEXE << 16);        # state = reset + input + execute
			$TTY[BUFFER]  = 0;                 # reset tty buffer
		}
		$CR = 0;
	}
	else
	{
		$CR = 1;                               # not accepted                    
	}
}
	
# tty service routine 
# processes the 'execute state' of the simulated tty cu+device, and
# - simulated panel interrupt (key CntrlP)
# - stopping of the simulator by pressing the stop key CntrE
#
sub svc_tty {
	my $key;
	my $stat = ($TTY[STATE] >> 16);
	my $k;
	
	if(defined ($key = ReadKey(-1)))
	{
		$k = ord ($key);
		#print "ttyin:" . prhex(16,$k) . "\n";
		#if($k == 0xA || $k == 0xD)     # ENTER provides 0A under Unix and 0D under Windows
		#{
		#	print "\r\n";
		#}
		#else
		#{
		#	if($TTY[CSET] == 64)
		#	{
		#		print UP($key);              # Echo, suppress LF(ENTER), which is converted to CR
		#	}
		#	else
		#	{
		#		print $key;
		#	}
		#}
		
		# ENTER provides 0A under Unix and 0D under Windows
		# Always respond with a 0D.
		if($k == 0xA)
		{
			$k = 0xD;
		}
		if($k == 0xD)     
		{
			prlog("\r\n");
		}
		else
		{
			if($TTY[CSET] == 64)
			{
				prlog(UP($key));              # Echo, suppress LF(ENTER), which is converted to CR
			}
			else
			{
				prlog($key);
			}
		}
		
		
		if($k == 16)	# CntrP
		{
			prlog("\nPanel interrupt\n");
			trap(CP);                                 # set interrupt
		}
		elsif($k == 5)  # CntrE
		{
			prlog("\nKeyboard interrupt (Stop execution)\n");
			$RUN = 0;
		}
		elsif( $stat == IEXE )                            # input + execute
		{
			if($TTY[BUFFER] == 0)                     # buffer reset?
			{
				if($TTY[CSET] == 64)
				{
					$TTY[BUFFER] = ord(UP($key)) & 0x7F;     # save character in buffer
				}
				else
				{
					$TTY[BUFFER] = $k;                # save character in buffer	
				}			
				$intreg = $intreg | IASR;                 # set the ready flipflop
				$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
				$TTY[STATE] = $TTY[STATE] | (IEXC << 16);     # state = input + exchange
				#print "tysvc: " . prhex(8,$k) . "\n";
                # wait for INR
			}
			else
			{
				prlog("Error: TTY-INR did not reset the buffer\n");				
			}
		}
		elsif( $stat == IEXC )                        # input + exchange
		{
			# throughput error (INR did not process)
			$intreg = $intreg | IASR;                 # set the ready flipflop
		    $TTY[STATE] = $TTY[STATE] & 0xFFFF;       # reset cu state bits
		    $TTY[STATE] = $TTY[STATE] | (WAIT << 16); # cu state = wait
			$TTY[STATE] = $TTY[STATE] | 2;            # set throughput error bit in state	
		}
		else
		{
			; # ignore key
		}
	}
	
	if( $stat == OEXE )                               # output + execute
	{
		$k = $TTY[BUFFER];                       # print buffered character

		if($k == 0xD)
		{
			prlog("\r\n");
		}
		elsif($k == 0xA)
		{
			;
		}
		else
		{
			prlog(chr($k));
		}
		$intreg = $intreg | IASR;                     # set the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (OEXC << 16);     # state = output + exchange
		# wait for OTR
	}
	
	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
		if($stat == (OEXE + WREQ))
		{
			if(!($intreg & IASR))
			{
				# not ready, so still a buffered output character pending
				$k = $TTY[BUFFER]; # print buffered character
				if($k == 0xD)
				{
					prlog("\r\n");
				}
				elsif($k == 0xA)
				{
					;
				}
				else
				{
					prlog(chr($k));
				}				                       
			}
		}
		$intreg = $intreg | IASR;                     # set the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (WAIT << 16);     # cu state = wait
	}
	# schedule next tty event
	activate_event(\@TTY, $TTY[TIME]);
}

sub inr_tty {
	my $k;
	
	if( ($TTY[STATE] >> 16) == IEXC )
	{
		# cu in exchange state
		$k = $TTY[BUFFER] & 0x7F;                     # buffered character
		if($k == 0xA)                                 # buffered character to r3
		{
			$R[$r3] = 0xD;
		}
		else
		{
			$R[$r3] = $k;
		}               
		$intreg = $intreg & (~ IASR);                 # reset the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (IEXE << 16);     # state = input + execute
        #print "tyinr: " . prhex(8,$R[$r3]) . "\n";
		$TTY[BUFFER] = 0;                             # reset the buffer
		$CR = 0;
	}
	else
	{
		$CR = 1;
	}	
}

sub otr_tty {
	if( ($TTY[STATE] >> 16) == OEXC )
	{
		# cu in exchange state
		$TTY[BUFFER] = $R[$r3] & 0x7F;                # buffer character
		$intreg = $intreg & (~ IASR);                 # reset the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (OEXE << 16);     # state = output + execute
		$CR = 0;	
	}
	else
	{
		$CR = 1;
	}
}

sub sst_tty {
	if( ($TTY[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $TTY[STATE];                        # r3 = status
		$intreg = $intreg & (~ IASR);                 # reset the ready flipflop
        $TTY[BUFFER] = 0;
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}
}

sub tst_tty {
	if( ($TTY[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}	
	$CR = 0;
}

sub state_tty {
	prlog("CU" . prhex(20,$TTY[STATE]) . " CR" . prhex(4,$CR) . " BUFFER" . prhex(8,$TTY[BUFFER]) . " CSET $TTY[CSET]" . "\n");
}

sub reset_tty {
	$intreg = $intreg & (~ IASR);                     # reset the ready flipflop
	$TTY[STATE]  = INAC << 16;                        # state = inactive + 0
	$TTY[BUFFER] = 0;                                 # reset tty buffer
}

######################### PTR device routines

sub cio_ptr {
	my $f = shift;                  # function bit
	my $stat = $PTR[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == IEXE) || ($stat == IEXC))
		{
			$PTR[STATE] = $PTR[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START INPUT, accepted if state == inactive
	    $PTR[STATE] = (IEXE << 16);                   # state = reset + input + execute
		$PTR[BUFFER]  = 0;                            # reset buffer
		$CR = 0;
		# schedule a PTR event
		activate_event(\@PTR, $PTR[TIME]);
	}
	else
	{
		$CR = 1;                                      # not accepted                    
	}
}
	
# ptr service routine 
# processes the 'execute state' of the simulated ptr cu+device.
#
sub svc_ptr {
	my $stat = ($PTR[STATE] >> 16);  
	my $c;
	$PTR[STATE] = $PTR[STATE] & 0xFFFF;                       # reset cu state bits

	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
		$PTR[STATE] = $PTR[STATE] | (WAIT << 16);             # cu state = wait
	}	
	elsif( $stat == IEXE )                                    # input + execute
	{
		if($PTR[FILE] eq "")
		{
			# not attached, so not operable
	        $PTR[STATE] = $PTR[STATE] | (WAIT << 16);         # cu state = wait
			$PTR[STATE] = $PTR[STATE] | 1;                    # set not operable error bit in state
		}
		elsif($PTR[BUFFER] == 0)                              # buffer reset?
		{
			$c = read_ptr();
			if($c)            
			{
				$PTR[BUFFER] = $c;                            # save character in buffer
				$PTR[STATE]  = ($PTR[STATE] | (IEXC << 16));  # state = input + exchange
				# wait for INR
				# schedule next PTR event
				activate_event(\@PTR, $PTR[TIME]);
			}
			else
			{
				# end of tape
				$PTR[STATE] = $PTR[STATE] | (WAIT << 16);     # cu state = wait
				$PTR[STATE] = $PTR[STATE] | 0x20;             # set eot bit in state	
			}
		}
		else
		{
			prlog("Error: PTR-INR did not reset the buffer\n");		
		}
	}
	elsif( $stat == IEXC )                                    # input + exchange
	{
		# throughput error (INR did not process)
	    $PTR[STATE] = $PTR[STATE] | (WAIT << 16);             # cu state = wait
		$PTR[STATE] = $PTR[STATE] | 2;                        # set throughput error bit in state
	}
	else
	{
		prlog("Error: PTR service request with CU in wrong state: $stat\n");	
	}
	$intreg = $intreg | IPTR;                                # set the ready flipflop
}

sub inr_ptr {
	
	if( ($PTR[STATE] >> 16) == IEXC )
	{
		# cu in exchange state
		$R[$r3] = $PTR[BUFFER] & 0xFF;                # buffered character
		$intreg = $intreg & (~ IPTR);                 # reset the ready flipflop
		$PTR[STATE] = $PTR[STATE] & 0xFFFF;           # reset cu state bits
		$PTR[STATE] = $PTR[STATE] | (IEXE << 16);     # state = input + execute
		$PTR[BUFFER] = 0;                             # reset the buffer
		$CR = 0;
	}
	else
	{
		$CR = 1;
	}	
}

sub sst_ptr {
	if( ($PTR[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$PTR[STATE] = $PTR[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $PTR[STATE];                        # r3 = status
		$intreg = $intreg & (~ IPTR);                 # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_ptr {
	if( ($PTR[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}	
	$CR = 0;
}

sub state_ptr {
	prlog("CU" . prhex(20,$PTR[STATE]) . " CR" . prhex(4,$CR) . " BUFFER" . prhex(8,$PTR[BUFFER]) . "\n");
    if($PTR[FILE] ne "")
	{
		prlog("Device PTR is attached to file: $PTR[FILE]\n");
	}
	else
	{
		prlog("Device PTR is not attached to a file\n");
	}
}

sub reset_ptr {
	$intreg = $intreg & (~ IPTR);                     # reset the ready flipflop
	$PTR[STATE]  = INAC << 16;                        # state = inactive + 0
	$PTR[BUFFER] = 0;                                 # reset buffer
	# if attached, reattach (set file at load point)
	if($PTR[FILE] ne "")
	{
		$file = $PTR[FILE];
		close ($PTR[FC]) || die "can't close $file: $!";
		open (PTR, $file) || die "Cannot open papertape inputfile $file: $!";
		$PTR[FC] = \*PTR;
		binmode $PTR[FC], ":raw";	
		prlog("File $file reattached to device PTR\n");
	}
}

######################### PTP device routines

sub cio_ptp {
	my $f = shift;                  # function bit
	my $stat = $PTP[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == OEXE) || ($stat == OEXC))
		{
			$PTP[STATE] = $PTP[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START OUTPUT, accepted if state == inactive
	    $PTP[STATE] = (OEXC << 16);                       # state = reset + output + exchange
		$intreg = $intreg | IPTP;                         # set the ready flipflop
		activate_event(\@PTP,$PTP[TIME]);                 # schedule a new event
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # not accepted                    
	}
}

# ptp service routine 
# processes the 'execute state' of the simulated ptp cu+device.
#
sub svc_ptp {
	my $stat = ($PTP[STATE] >> 16);  
    #print "1 service ptp called $stat\n";

	if( $stat == OEXE )                                       # output + execute
	{
		if($PTP[FILE] eq "")
		{
			# not attached, so not operable
		    $PTP[STATE] = $PTP[STATE] & 0xFFFF;               # reset cu state bits			
	        $PTP[STATE] = $PTP[STATE] | (WAIT << 16);         # cu state = wait
			$PTP[STATE] = $PTP[STATE] | 1;                    # set not operable error bit in state
			$intreg = $intreg | IPTP;                         # set the ready flipflop
		}
		else
		{
			punch_ptp($PTP[BUFFER]);
			$PTP[STATE] = $PTP[STATE] & 0xFFFF;               # reset cu state bits	
			$PTP[STATE]  = ($PTP[STATE] | (OEXC << 16));      # state = output + exchange
			$intreg = $intreg | IPTP;                         # set the ready flipflop
			# wait for next OTR or CIO HALT
			# schedule next PTP event
			activate_event(\@PTP, $PTP[TIME]);
		}
	}
	elsif( $stat & WREQ)
	{
		# stop io requested (CIO stop)
		if($stat == (OEXE + WREQ))
		{
			if(!($intreg & IPTP))
			{
				# not ready, so still a buffered output character pending
				punch_ptp($PTP[BUFFER]);
			}
		}
		$PTP[STATE] = $PTP[STATE] & 0xFFFF;                   # reset cu state bits
		$PTP[STATE] = $PTP[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | IPTP;                             # set the ready flipflop
	}		
	else
	{
		# do nothing: wait for OTR or CIO HALT
		activate_event(\@PTP, $PTP[TIME]);
	}
}

sub otr_ptp {
	
	if( ($PTP[STATE] >> 16) == OEXC )
	{
		# cu in exchange state
		$PTP[BUFFER] = $R[$r3];                       # buffered character
		$intreg = $intreg & (~ IPTP);                 # reset the ready flipflop
		$PTP[STATE] = $PTP[STATE] & 0xFFFF;           # reset cu state bits
		$PTP[STATE] = $PTP[STATE] | (OEXE << 16);     # state = output + execute
		$CR = 0;
	}
	else
	{
		$CR = 1;
	}	
}

sub sst_ptp {
	if( ($PTP[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$PTP[STATE] = $PTP[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $PTP[STATE];                        # r3 = status
		$intreg = $intreg & (~ IPTP);                 # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_ptp {
	if( ($PTP[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}	
	$CR = 0;
}

sub state_ptp {
	prlog("CU" . prhex(20,$PTP[STATE]) . " CR" . prhex(4,$CR) . " BUFFER" . prhex(8,$PTP[BUFFER]) . "\n");
    if($PTP[FILE] ne "")
	{
		prlog("Device PTP is attached to file: $PTP[FILE]\n");
	}
	else
	{
		prlog("Device PTP is not attached to a file\n");
	}
}

sub reset_ptp {
	$intreg = $intreg & (~ IPTP);                     # reset the ready flipflop
	$PTP[STATE]  = INAC << 16;                        # state = inactive + 0
	$PTP[BUFFER] = 0;                                 # reset buffer
	# if attached, reattach
	if($PTP[FILE] ne "")
	{
		$file = $PTP[FILE];
		close ($PTP[FC]) || die "can't close $file: $!";
		open (PTP, ">:raw", $file) || die "Cannot open paper punch file $file: $!";	
		$PTP[FC] = \*PTP;
		prlog("File $file reattached to device PTP\n");
	}
}

####################### X1215 Moving Head Disk routines ############################
# The control unit is wired to an IOP subchannel; so only an interrupt when the
# cu switches to wait state. Data is transferred to/from memory from/to the internal 
# disk control unit buffer by the IOP/controlunit by means of breaks
# The controlunit supports 2 disk units, each unit with 2 disks so 4 disks/drives in total.
# The control unit is connected to IOP subchannel 2.
# 
# The addressing of a particular disk/drive is as follows:
#                                                    DA = 00 0 010
#                                                         ^^
#                                                         ||Unit Number 0,1
#                                                         |Drive Number 0,1
# Drive number 0 is the cartridge and drive number 1 is the fixed disk.
# Because the 4 drives share a singe control unit, only 1 drive can be active
# at a particular moment.
# The implementation supports two unit2: 
# unit 0; 2 drives. Drive 0 has address /02 and drive 1 has address /22.
# unit 1; 2 drives. Drive 0 has address /12 and drive 1 has address /32.

# unit 0
sub cio_md0 {
	my $f  = shift;	
	my $dn = 0;     # drive/unit number (address 2)
	cio_md($f,$dn);
}

sub cio_md1 {
	my $f  = shift;	
	my $dn = 2;     # drive/unit number	(address 22)
	cio_md($f,$dn);		
}

# unit 1
sub cio_md2 {
	my $f  = shift;	
	my $dn = 1;     # drive/unit number (address 12)
	cio_md($f,$dn);
}

sub cio_md3 {
	my $f  = shift;	
	my $dn = 3;     # drive/unit number	(address 32)
	cio_md($f,$dn);	
}

# common CIO processor
sub cio_md {
	my $f  = shift;
	my $dn = shift;
	my $stat = $MD0[STATE] >> 16;   # get status: cu state
	my $r;
	my $d;
	my $c;
	my $p;
	if($f == 0)
	{	
	  	# CIO STOP, alsways accepted
		if(($stat == IEXE) || ($stat == IEXC) || ($stat == OEXC)|| ($stat == OEXC))
		{
			$MD0[STATE] = $MD0[STATE] | (WREQ << 16);     # state = wait state requested
		}
	}
	elsif($stat == INAC)
	{
		# CIO START, accepted if state == inactive
		$MD0[STATE] = 0;                                  # reset state
		$md0buf     = "";                                 # reset buffer
		$MD0[XLEN]  = 0;                                  # mpx transferred length in bytes				
		$r = $R[$r3] & 0x3;                               # requested operation
		if($r == 0)
		{
			# read a physical sector, CIO START INPUT
			$MD0[SEC]  = ($R[$r3] >> 2) & 0x1F;           # physical sector to read: 0..31
			$MD0[STATE] = $dn << 6;                       # set disc/unit number
			$MD0[STATE] = $MD0[STATE] | (IEXE << 16);     # state = input + execute
		  	activate_event(\@MD0,$MD0[RWTIME]);           # schedule event for moving to sector in cylinder 		
		}
		elsif($r == 1)
		{
			#  write a physical sector, CIO START OUTPUT
			$MD0[SEC]  = ($R[$r3] >> 2) & 0x1F;           # physical sector to write: 0..31
			$MD0[STATE] = $dn << 6;                       # set disc/unit number			
			$MD0[STATE] = $MD0[STATE] | (OEXC << 16);     # state = output + exchange
		  	activate_event(\@MD0,$MD0[RWTIME]);           # schedule event to start mpx transfer 		
		}
		elsif($r == 2)
		{
			# move heads of a unit to another cylinder (SEEK)
			
			# On the P856 the seek is slightly different from the P855/P860 seek
			# On the P855/P860 the delta number of cylinders to move with a direction is
			# specified. On the P856 just the cylinder to move to is specified in $R[$r3].
			#
			$MD0[STATE] = ($dn & 1) << 6;                 # set unit number
			$MD0[STATE] = $MD0[STATE] | (WREQ << 16);     # cu state = wait request
			$c=$R[$r3] >> 3;                              # cylinder to move to
			$p = $c * SPT * NH * LSEC * 2;                # new position in chars
			if( ($dn == 0 || $dn == 2) && $MD0[RDY])
			{
				# unit 0 ($MD0[CYL] has the current unit cylinder)
				if(sysseek(MD0, $p, 0) ne undef)      # position heads to requested cylinder
				{
					$d=abs($c-$MD0[CYL]); 						# number of cylinders to move
					$MD0[CYL] = $c;
					$MD0[STATE] = $MD0[STATE] | DONCYL;   						
					# schedule event for moving $d cylinders
					activate_event(\@MD0,(($d * $MD0[TRTIME]) + $MD0[RWTIME])); 
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);    # schedule event
				}
			}
			elsif( ($dn == 1 || $dn == 3) && $MD2[RDY])
			{
				# unit 1 ($MD2[CYL] has the current unit cylinder)
				if(sysseek(MD2, $p, 0) ne undef)      		# position heads to requested cylinder
				{
					$d=abs($c-$MD2[CYL]); 						# number of cylinders to move
					$MD2[CYL] = $c;
					$MD0[STATE] = $MD0[STATE] | DONCYL;   						
					# schedule event for moving $d cylinders
					activate_event(\@MD0,(($d * $MD0[TRTIME]) + $MD0[RWTIME])); 
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);    # schedule event
				}
			}
			else
			{
				# not attached, so not operable
				$MD0[STATE] = $MD0[STATE] | DSEEKERR| DONCYL | DNOTOP;  # set not operable error bit in state					
			  	activate_event(\@MD0,$MD0[RWTIME]);       # schedule event, 				
			}		
		}
		else
		{
			# seek unit to cylinder 0
			$MD0[STATE] = ($dn & 1) << 6;                # set unit number
			$MD0[STATE] = $MD0[STATE] | (WREQ << 16);    # cu state = wait request			
				
			if( ($dn == 0 || $dn == 2) && $MD0[RDY])
			{
				# unit 0 ($MD0[CYL] has the current unit cylinder)				
				if(sysseek(MD0, 0, 0) ne undef)           # position heads to cylinder 0
				{
					$d = $MD0[CYL];
					$MD0[CYL] = 0;
					$MD0[STATE] = $MD0[STATE] | DONCYL; 					
					# schedule event for moving to cylinder 0
					activate_event(\@MD0,(($d * $MD0[TRTIME]) + $MD0[RWTIME])); 
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
				}
			}
			elsif( ($dn == 1 || $dn == 3) && $MD2[RDY])
			{
				# unit 1 ($MD2[CYL] has the current unit cylinder)				
				if(sysseek(MD2, 0, 0) ne undef)              # position heads to cylinder 0
				{
					$d = $MD2[CYL];
					$MD2[CYL] = 0;
					$MD0[STATE] = $MD0[STATE] | DONCYL; 					
					# schedule event for moving to cylinder 0
					activate_event(\@MD0,(($MD0[CYL] * $MD0[TRTIME]) + $MD0[RWTIME])); 
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
				}
			}
			else
			{
				# not attached, so not operable
				$MD0[STATE] = $MD0[STATE] | DSEEKERR| DONCYL | DNOTOP;  # set not operable error bit in state					
			  	activate_event(\@MD0,$MD0[RWTIME]);          # schedule event, 				
			}
		}
		$MD0[DN] = $dn; # save active drive/unit number
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # CIO not accepted 
	}
}

sub svc_md0 {
	my $stat = $MD0[STATE] >> 16;
	my $dn   = $MD0[DN];                                 # number of active device
	my $pos;                                             # sector position in cylinder
    	my $cyl;                                          # current cylider
	if(($dn == 0) || $dn == 2)
	{
		$cyl=$MD0[CYL];                                	  # set current cylinder (unit 0)
	}
	else
	{
		$cyl=$MD2[CYL];                                	  # set current cylinder (unit 1)
	}
	
	if( $stat & WREQ)
	{
		# stop io requested (CIO STOP or end of SEEK)
    	$MD0[STATE] = $MD0[STATE] & 0xFFFF;               # reset cu state bits		
		$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
		$intreg = $intreg | IMD0;                         # set the ready flipflop			
	}	
	elsif( $stat == IEXE )                               # input + execute
	{
		$MD0[STATE]  = $MD0[STATE] & 0xFFFF;              # reset cu state bits		
		if(($dn == 0 && $MD0[RDY]) || ($dn == 2 && $MD0[RDY]) || 
		   ($dn == 1 && $MD2[RDY]) || ($dn == 3 && $MD2[RDY]))
		{
			# read requested physical sector into the device buffer
			$pos = (($cyl * SPT * NH) + $MD0[SEC]) * LSEC * 2;  # sector position in cylinder
			if($dn==0)
			{
				sysseek(MD0,$pos,0);                          # position and
				$stat = sysread(MD0, $md0buf, LSEC*2);        # read sector
			}
			elsif($dn==2)
			{
				sysseek(MD1,$pos,0);                          # position
				$stat = sysread(MD1, $md0buf, LSEC*2);        # read sector
			}
			elsif($dn==1)
			{
				sysseek(MD2,$pos,0);                          # position and
				$stat = sysread(MD2, $md0buf, LSEC*2);        # read sector			
			}
			elsif($dn==3)
			{
				sysseek(MD3,$pos,0);                          # position
				$stat = sysread(MD3, $md0buf, LSEC*2);        # read sector
			}			
			
			# check sector identifier as an integrity check. The real 
			# disk has the cylinder number as identifier in the 1st word of the sector.
			# cylinder: on the X1210/1215 cyl number fits in 8 bit.
			# on the X1216 in 10 bit
			my $s = unpack("C",substr($md0buf,0,1)) >> 3;     # physical sector number on disk
			my $c = unpack("C",substr($md0buf,1,1));          # cylinder (X1210/1215 cyl number fits in 8 bit)
   
			if(($stat == LSEC*2) && ($c == $cyl))    # check length an cylinder as on real X1210/X1215 
			{
				# internal buffer ok and ready for mpx transfer
				$MD0[BUFFER] = 0;                             # reset word buffer
				$MD0[STATE]  = ($MD0[STATE] | (IEXC << 16));  # state = input + exchange
				# wait for INR
				activate_event(\@MD0, $MD0[TIME]);            # schedule event for mpx transfer
			}
			else
			{
				#error
				$MD0[STATE] = $MD0[STATE] | (WAIT << 16);     # cu state = wait				
				if($stat != LSEC*2)
				{
					# incorrect length
					$MD0[STATE] = $MD0[STATE] | DLENGERR;     # set error bit in state
				}
				else
				{
					# sector identifier error
					$MD0[STATE] = $MD0[STATE] | DDATAERR;     # set error bit in state					
				}
				$intreg = $intreg | IMD0;                     # set the ready flipflop					
			}
		}
		else
		{
			# active disk not operable
			$MD0[STATE] = $MD0[STATE] | DNOTOP;               # set not operable error bit in state			
			$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
		   $intreg = $intreg | IMD0;                         # set the ready flipflop	
		}		
	}
	elsif( $stat == IEXC )                                    # input + exchange
	{
		# move internal device buffer via mpx into memory buffer
		if($MD0[BUFFER] == 0)
		{
			$brkreg = $brkreg | BMD0;                         # set an mpx break request
			activate_event(\@MD0,$MD0[TIME]);                 # schedule next mpx event			
		}
		else
		{
			# throughput error (INR/break did not process)
    	   $MD0[STATE] = $MD0[STATE] & 0xFFFF;               # reset cu state bits				
			$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
			$MD0[STATE] = $MD0[STATE] | DTHROERR;             # set throughput error bit in state
		   $intreg = $intreg | IMD0;                         # set the ready flipflop			
		}
	}
	elsif( $stat == OEXC )                                    # output + exchange
	{
		# move memory buffer into internal device buffer via mpx
		$brkreg = $brkreg | BMD0;                             # set an mpx break request
		activate_event(\@MD0,$MD0[TIME]);                     # schedule next mpx event
	}	
	elsif( $stat == OEXE )
	{
		# write device buffer into the requested sector
		$MD0[STATE] = $MD0[STATE] & 0xFFFF;                   # reset cu state bits			
		if(($dn == 0 && $MD0[RDY]) || ($dn == 2 && $MD0[RDY]) ||
	      ($dn == 1 && $MD2[RDY]) || ($dn == 3 && $MD2[RDY]))
		{
			# check sector identifier to write as a disk integrety check
			# format sector identifier
			# xxxxx sector (5 bit) 0..31 (!!this is the physical sector number!!)
			#      00        (2 bit)
			#        xxxxxxxxx cylinder (9 bit) 0..202
			my $l = length($md0buf);                          # length of the requested buffer to write
			my $s = unpack("C", substr($md0buf,0,1)) >> 3;    # physical sector number IN the sector buffer to write
			my $c = unpack("C", substr($md0buf,1,1));         # cylinder (X1210/1215 cyl number fits in 8 bit)
            		# $MD0[SEC] is the physical sector requested to write
			#if($l == 2*LSEC && $c == $cyl && $s == $MD0[SEC])
			if($l == 2*LSEC && $c == $cyl)			
			{
				# sector format ok, write the sector
				$pos = (($cyl * SPT * NH) + $MD0[SEC]) * LSEC * 2;  # sector position in cylinder
				if($dn==0)
				{
					sysseek(MD0,$pos,0);                      # position and				
					$stat = syswrite(MD0, $md0buf, LSEC*2);   # write sector
				}
				elsif($dn==2)
				{
					sysseek(MD1,$pos,0);                      # position and
					$stat = syswrite(MD1, $md0buf, LSEC*2);   # write sector				
				}
				elsif($dn==1)
				{
					sysseek(MD2,$pos,0);                      # position and				
					$stat = syswrite(MD2, $md0buf, LSEC*2);   # write sector
				}
				elsif($dn==3)
				{
					sysseek(MD3,$pos,0);                      # position and
					$stat = syswrite(MD3, $md0buf, LSEC*2);   # write sector				
				}				
				
				# and check the result
				if($stat == LSEC*2)            
				{
					# internal buffer transferred to disk
					$MD0[STATE]  = ($MD0[STATE] | (WREQ << 16));  # state = wait request
					activate_event(\@MD0, $MD0[RWTIME]);          # schedule event write event
				}
				else
				{
					# incorrect length				
					$MD0[STATE] = $MD0[STATE] | (WAIT << 16);     # cu state = wait
					$MD0[STATE] = $MD0[STATE] | DLENGERR;         # set error bit in state
					$intreg = $intreg | IMD0;                     # set the ready flipflop				
				}				
			}
			else
			{
				# error in the format of the sector buffer to write; not written to disk			
				$MD0[STATE] = $MD0[STATE] | (WAIT << 16);     # cu state = wait				
				if($l != 2*LSEC)
				{
					# incorrect length
					$MD0[STATE] = $MD0[STATE] | DLENGERR;       # set error bit in state
				}
				#if($s != $MD0[SEC] || $c != $cyl)               # check sector identifier
				if($c != $cyl)                                  # check sector identifier				
				{
					# sector identifier in buffer is corrupted => program error
					# !!!!!!on the real P800 disk, the sector number was not part of the
					# sector identier!!!!!!!!!! (so that test must be removed later)
					$MD0[STATE] = $MD0[STATE] | DPROGERR;         # set error bit in state
				}
				$intreg = $intreg | IMD0;                     # set the ready flipflop				
			}
		}
		else
		{
			# active disk not operable
			$MD0[STATE] = $MD0[STATE] | DNOTOP;               # set not operable error bit in state				
			$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
		    $intreg = $intreg | IMD0;                         # set the ready flipflop	
		}
	}
	else	
	{
		prlog("Error: MD0 service request with CU in wrong state: $stat\n");		
	}	
}

sub inr_md0 {
	# byte or word transfer (internal buffer => memory)
	my $t = shift;    # transfer mode: 0=byte, 1=word 
	my $c;
	my $pos = $MD0[XLEN];
	if( $pos < LSEC*2)
	{
		if($t)
		{
			# word transfer
			$c = substr($md0buf,$pos,1);
			$MD0[BUFFER] = unpack("C", $c) << 8;
			$c = substr($md0buf,($pos+1),1);
			$c = unpack("C", $c);
			$MD0[BUFFER] = $MD0[BUFFER] | $c;
			$MD0[XLEN]   = $MD0[XLEN] + 2;
		}
		else
		{
			# byte transfer
			$c = substr($md0buf,$pos,1);
			$MD0[BUFFER] = unpack("C", $c);
			$MD0[XLEN]   = $MD0[XLEN] + 1;			
		}
	}
	else
	{
		$MD0[BUFFER] = 0;
	}
}

sub otr_md0 {
	# byte or word transfer (memory => internal buffer)		
	my $t = shift;    # transfer mode: 0=byte, 1=word
	my $c;
	if($t)
	{
		# word transfer
        $c = pack("C", (($MD0[BUFFER] >> 8) & 0xFF));		
		$md0buf=$md0buf . $c;
   		$c = pack("C", ($MD0[BUFFER] & 0xFF));  
		$md0buf=$md0buf . $c;
		$MD0[XLEN] = $MD0[XLEN] + 2;	
	}
	else
	{

		# byte transfer
		$c = pack("C", ($MD0[BUFFER] & 0xFF));
		$md0buf=$md0buf . $c;
		$MD0[XLEN] = $MD0[XLEN] + 1;		
	}
}

sub sst_md0 {
	if( ($MD0[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$MD0[STATE] = $MD0[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $MD0[STATE];                        # r3 = status
		$intreg = $intreg & (~ IMD0);                 # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}
}

sub tst_md0 {
	if( ($MD0[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # CU busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}
	$CR=0;
}

sub state_md0 {
    $t1 = length($md0buf);
	prlog("CU" . prhex(20,$MD0[STATE]) . " CR" . prhex(4,$CR) . " BUFFER-LENGTH" . prhex(8,$t1) . " TRL" . prhex(16,$MD0[XLEN]) . "\n");
    prlog("(Last) selected drive(dn):$MD0[DN] PSector:$MD0[SEC] Cylinder: Unit-0 $MD0[CYL] Unit-1 $MD2[CYL]\n");
    if($MD0[FILE] ne "")
	{
		prlog("Device MD0 (a=0x02 d=0 u=0 dn=0) is attached to file: $MD0[FILE]\n");
	}
	else
	{
		prlog("Device MD0 (a=0x02 d=0 u=0 dn=0) is not attached to a file\n");
	}
    if($MD1[FILE] ne "")
	{
		prlog("Device MD1 (a=0x22 d=1 u=0 dn=2) is attached to file: $MD1[FILE]\n");
	}
	else
	{
		prlog( "Device MD1 (a=0x22 d=1 u=0 dn=2) is not attached to a file\n");
	}
    if($MD2[FILE] ne "")
	{
		prlog("Device MD2 (a=0x12 d=0 u=1 dn=1) is attached to file: $MD2[FILE]\n");
	}
	else
	{
		prlog("Device MD2 (a=0x12 d=0 u=1 dn=1) is not attached to a file\n");
	}
    if($MD3[FILE] ne "")
	{
		prlog("Device MD3 (a=0x32 d=1 u=1 dn=3) is attached to file: $MD3[FILE]\n");
	}
	else
	{
		prlog("Device MD3 (a=0x32 d=1 u=1 dn=3) is not attached to a file\n");
	}
	if($MD0[RDY])
	{
		prlog("X1215 disk unit-0 is ready\n");
	}
	else
	{
		prlog("X1215 disk unit-0 is NOT ready\n");
	}
	if($MD2[RDY])
	{
		prlog("X1215 disk unit-1 is ready\n");
	}
	else
	{
		prlog("X1215 disk unit-1 is NOT ready\n");
	}	
}

sub reset_md0 {
	$intreg = $intreg & (~ IMD0);                     # reset the ready flipflop
	$brkreg = $brkreg & (~ BMD0);                     # reset break flipflop	
	$MD0[STATE]  = INAC << 16;                        # state = inactive + 0
	$MD0[BUFFER] = 0;                                 # reset buffer
	$MD0[XLEN]   = 0;                                 # reset transferred length
	$MD0[DN]     = 0;                                 # reset to primary device
	$MD0[SEC]    = 0;                                 # reset physical sector
	$MD0[CYL]    = 0;								  # unit 0 at load point
	$MD2[CYL]    = 0;                                 # unit 1 at load point
	$md0buf      = "";
	# if attached, reattach (set file at load point)
	if($MD0[FILE] ne "")
	{
		$file = $MD0[FILE];
		close (MD0) || die "can't close $file: $!";
		open (MD0, "+<:raw", $file) || die "Cannot open  disk file $file: $!";
		prlog("File $file reattached to device MD0\n");
	}
	if($MD1[FILE] ne "")
	{
		$file = $MD1[FILE];
		close (MD1) || die "can't close $file: $!";
		open (MD1, "+<:raw", $file) || die "Cannot open  disk file $file: $!";
		prlog("File $file reattached to device MD1\n");
	}
	if($MD2[FILE] ne "")
	{
		$file = $MD2[FILE];
		close (MD2) || die "can't close $file: $!";
		open (MD2, "+<:raw", $file) || die "Cannot open  disk file $file: $!";
		prlog("File $file reattached to device MD2\n");
	}
	if($MD3[FILE] ne "")
	{
		$file = $MD3[FILE];
		close (MD3) || die "can't close $file: $!";
		open (MD3, "+<:raw", $file) || die "Cannot open  disk file $file: $!";
		prlog("File $file reattached to device MD3\n");
	}
	if($MD0[FILE] ne "" && $MD1[FILE] ne "")
	{
		$MD0[RDY]=1;
		prlog("X1215 disk unit-0 is ready\n");
	}
	else
	{
		$MD0[RDY]=0;
		prlog("X1215 disk unit-0 is NOT ready\n");
	}
	if($MD2[FILE] ne "" && $MD3[FILE] ne "")
	{
		$MD2[RDY]=1;
		prlog("X1215 disk unit-1 is ready\n");
	}
	else
	{
		$MD2[RDY]=0;
		prlog("X1215 disk unit-1 is NOT ready\n");
	}	
	
}

######################### CRD device routines
# The control unit is wired to a multiplex channel; so only an interrupt when the
# cu switches to wait state. Data is transferred to memory from the internal 
# cardreader buffer by the multiplexer/control unit by means of breaks
#
sub cio_crd {
	my $f = shift;                  # function bit
	my $stat = $CRD[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == IEXE) || ($stat == IEXC))
		{
			$CRD[STATE] = $CRD[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START INPUT, accepted if state == inactive
	    $CRD[STATE] = (IEXE << 16);                       # state = reset + input + execute
        $crdbuf = "";                                     # reset buffer
		activate_event(\@CRD,$CRD[RTIME]);                # schedule a event, move next card
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # not accepted                    
	}
}		
		
sub svc_crd {
	my $stat = ($CRD[STATE] >> 16);  
	
	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
    	$CRD[STATE] = $CRD[STATE] & 0xFFFF;                   # reset cu state bits			
		$CRD[STATE] = $CRD[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | ICR;                              # set the ready flipflop			
	}	
	elsif( $stat == IEXE )                                    # input + execute
	{
		if($CRD[FILE] eq "")
		{
			# not attached, so not operable
    	    $CRD[STATE] = $CRD[STATE] & 0xFFFF;               # reset cu state bits				
	        $CRD[STATE] = $CRD[STATE] | (WAIT << 16);         # cu state = wait
			$CRD[STATE] = $CRD[STATE] | 1;                    # set not operable error bit in state
		    $intreg = $intreg | ICR;                          # set the ready flipflop	
		}
		else
		{
			# read next card into the device buffer
	        my $fc = $CRD[FC];	
			$CRD[RCLEN] = 0;                                 # read length in bytes			
			$stat = read($fc, $crdbuf, 160);
			if($stat == 160)            
			{
				$CRD[BUFFER] = 0;                            # reset word buffer
    	        $CRD[STATE] = $CRD[STATE] & 0xFFFF;          # reset cu state bits					
				$CRD[STATE]  = ($CRD[STATE] | (IEXC << 16));  # state = input + exchange
				# wait for INR
				# schedule next CRD event
				activate_event(\@CRD, $CRD[TIME]);
			}
			else
			{
				# incorrect length (or no card)
    	        $CRD[STATE] = $CRD[STATE] & 0xFFFF;           # reset cu state bits						
				$CRD[STATE] = $CRD[STATE] | (WAIT << 16);     # cu state = wait
				$CRD[STATE] = $CRD[STATE] | 0x20;             # set eof bit in state
		        $intreg = $intreg | ICR;                      # set the ready flipflop				
			}
		}
	}
	elsif( $stat == IEXC )                                    # input + exchange
	{
		if($CRD[BUFFER] == 0)
		{
			$brkreg = $brkreg | BCR;                          # set an mpx break request
			activate_event(\@CRD,$CRD[TIME]);                 # schedule next event			
		}
		else
		{
			# throughput error (INR/break did not process)
    	    $CRD[STATE] = $CRD[STATE] & 0xFFFF;               # reset cu state bits				
			$CRD[STATE] = $CRD[STATE] | (WAIT << 16);         # cu state = wait
			$CRD[STATE] = $CRD[STATE] | 2;                    # set throughput error bit in state
		    $intreg = $intreg | ICR;                          # set the ready flipflop			
		}
	}
	else
	{
		prlog("Error: CRD service request with CU in wrong state: $stat\n");
	}
}

sub inr_crd {
	# word transfer is assumed!
    my $pos = $CRD[RCLEN]; 
	my $c;
	if($pos < 159)
	{
		$c = substr($crdbuf,$pos,1);
		$CRD[BUFFER] = unpack("C", $c) << 8;
		$c = substr($crdbuf,($pos+1),1);
		$c = unpack("C", $c);
		$CRD[BUFFER] = $CRD[BUFFER] | $c;
		$CRD[RCLEN] = $CRD[RCLEN] + 2;
	}
	else
	{
		$CRD[BUFFER] = 0;
	}
}

sub sst_crd {
	if( ($CRD[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$CRD[STATE] = $CRD[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $CRD[STATE];                        # r3 = status
		$intreg = $intreg & (~ ICR);                  # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_crd {
	if( ($CRD[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}	
	$CR = 0;
}

sub state_crd {
    $t1 = length($crdbuf);
	prlog("CU" . prhex(20,$CRD[STATE]) . " CR" . prhex(4,$CR) . " BUFFER-LENGTH" . prhex(8,$t1) . " TRL" . prhex(8,$CRD[RCLEN]) . "\n");
    if($CRD[FILE] ne "")
	{
		prlog("Device CRD is attached to file: $CRD[FILE]\n");
	}
	else
	{
		prlog("Device CRD is not attached to a file\n");
	}				
}

sub reset_crd {
	$intreg = $intreg & (~ ICR);                      # reset the ready flipflop
	$brkreg = $brkreg & (~ BCR);                      # reset break flipflop	
	$CRD[STATE]  = INAC << 16;                        # state = inactive + 0
	$CRD[BUFFER] = 0;                                 # reset buffer
	$CRD[RCLEN]  = 0;                                 # reset transferred length
	$crdbuf      = "";
	# if attached, reattach (set file at load point)
	if($CRD[FILE] ne "")
	{
		$file = $CRD[FILE];
		close ($CRD[FC]) || die "can't close $file: $!";
		open (CRD, $file) || die "Cannot open card reader inputfile $file: $!";
		$CRD[FC] = \*CRD;
		binmode $CRD[FC], ":raw";	
		prlog("File $file reattached to device CRD\n");
	}

}

######################### LPT device routines
# The control unit is wired to a multiplex channel; so only an interrupt when the
# cu switches to wait state. Data is transferred to the internal lineprinter buffer
# by the multiplexer/control unit by means of breaks
# The 1st 2 bytes in the buffer are interpreted as carriage control characters,
# supported as:
# - 0D0A => skip 1 line, after printing the line
# - 0A0A => skip 2 lines, after printing the line
# - 0D0C => print line and skip to top of next new page
# - 0A0C => print line, skip 1 line and skip to top of next new page
#
sub cio_lpt {
	my $f = shift;                  # function bit
	my $stat = $LPT[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == OEXE) || ($stat == OEXC))
		{
			$LPT[STATE] = $LPT[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START OUTPUT, accepted if state == inactive
	    $LPT[STATE] = (OEXC << 16);                       # state = reset + output + exchange
        $lptbuf = "";                                     # reset buffer
		activate_event(\@LPT,$LPT[TIME]);                 # schedule a event
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # not accepted                    
	}
}

sub svc_lpt {
	my $stat = ($LPT[STATE] >> 16); 

	if( $stat & WREQ)
    {
		$LPT[STATE] = $LPT[STATE] & 0xFFFF;                   # reset cu state bits
		$LPT[STATE] = $LPT[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | ILP;                              # set the ready flipflop	
	}
	elsif( $stat == OEXC )                                    # output + exchange
	{
		$brkreg = $brkreg | BLP;                              # set an mpx break request
		activate_event(\@LPT,$LPT[TIME]);                     # schedule next event
	}
	elsif( $stat == OEXE )                                    # output + execute	
	{
		if($LPT[FILE] eq "")
		{
			# not attached, so not operable
		    $LPT[STATE] = $LPT[STATE] & 0xFFFF;               # reset cu state bits			
	        $LPT[STATE] = $LPT[STATE] | (WAIT << 16);         # cu state = wait
			$LPT[STATE] = $LPT[STATE] | 1;                    # set not operable error bit in state
			$intreg = $intreg | ILP;                          # set the ready flipflop
		}
		else
		{
	        my $fc = $LPT[FC];
			my $c1 = ord(substr($lptbuf,0,1));
			my $c2 = ord(substr($lptbuf,1,1));
			my $l  = length($lptbuf);
			# print $l . prhex(8,$c1) . prhex(8,$c2) . "\n";
			
			# skip before printing the buffer
			if($LPT[SKIP] == SKIP_BP)
			{
				if($c1 == 0xA)
				{
					# new line
					pr_lpt("\r\n");
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xA)
				{
					# new line
					pr_lpt("\r\n");
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xC && $LPT[LCUR] != 0)
				{
					# new page if not on new page position
					while( $LPT[LCUR] !=  $LPT[LMAX] + $LPT[NPAG])
					{
						pr_lpt("\r\n");
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
				if($LPT[LCUR] == $LPT[LMAX] )			# end of page ?
				{
					# new page
					while( $LPT[LCUR] != ( $LPT[LMAX] + $LPT[NPAG]))
					{
						pr_lpt("\r\n");
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
			}			
			
			if($LPT[SKIP] == SKIP_NO)
			{
				# no printer control => just print the buffer
				pr_lpt($lptbuf);
            }
			else
			{
				# printer control: either skip before or after print (set by SKIP_BP or SKIP_AP)
				if($l > 2)
				{
					# strip control characters and print the line
					$lptbuf = substr($lptbuf,2,$l-2);
					pr_lpt($lptbuf);
				}
			}
			
			# skip page after printing the buffer
			if($LPT[SKIP] == SKIP_AP)
			{
				if($c1 == 0xA)
				{
					# new line
					pr_lpt("\r\n");
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xA)
				{
					# new line
					pr_lpt("\r\n");
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xC && $LPT[LCUR] != 0)
				{
					# new page if not on new page position
					while( $LPT[LCUR] !=  $LPT[LMAX] + $LPT[NPAG])
					{
						pr_lpt("\r\n");
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
				if($LPT[LCUR] == $LPT[LMAX] )			# end of page ?
				{
					# new page
					while( $LPT[LCUR] != ( $LPT[LMAX] + $LPT[NPAG]))
					{
						pr_lpt("\r\n");
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
			}
			
			$LPT[STATE] = $LPT[STATE] | (WREQ << 16);             # state = wait state requested
			activate_event(\@LPT,$LPT[PTIME]);                    # schedule next event (print time)
		}
	}
	else
	{
		prlog("LP-SVC: Lineprinter CU in unexpected state\n");
	}
}

sub otr_lpt {

	# cu in exchange state (byte transfer is assumed!)
	$lptbuf = $lptbuf . chr($LPT[BUFFER]);
}

sub sst_lpt {
	if( ($LPT[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$LPT[STATE] = $LPT[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $LPT[STATE];                        # r3 = status
		$intreg = $intreg & (~ ILP);                  # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_lpt {
	if( ($LPT[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}
	$CR = 0;
}

sub state_lpt {
    $t1 = length($lptbuf);
	prlog("CU" . prhex(20,$LPT[STATE]) . " CR" . prhex(4,$CR) . " BUFFER-LENGTH" . prhex(8,$t1) . "\n");
	prlog("BUFFER:" . $lptbuf . "\n");
    if($LPT[FILE] ne "")
	{
		prlog("Device LPT is attached to file: $LPT[FILE]\n");
	}
	else
	{
		prlog("Device LPT is not attached to a file\n");
	}
}

sub reset_lpt {
	$intreg = $intreg & (~ ILP);                      # reset the ready flipflop
	$brkreg = $brkreg & (~ BLP);                      # reset break flipflop
	$LPT[STATE]  = INAC << 16;                        # state = inactive + 0
    $lptbuf	= "";                                     # reset buffer
	$LPT[LCUR]  = 0;                                  # set on line 0 current page 	
	# if attached, reattach
	if($LPT[FILE] ne "")
	{
		$file = $LPT[FILE];
		close ($LPT[FC]) || die "can't close $file: $!";
		open (LPT, ">$file") || die "Cannot open line printer output file $file: $!";	
		$LPT[FC] = \*LPT;
		prlog("File $file reattached to device LPT\n");
	}
}

sub pr_lpt {
    # print a string on the lpt
    # in case lpt has a window, the string is also printed to that window
    my $str = shift;
    my $fc = $LPT[FC];
    print $fc $str;
    if($lptwin ne "")
    {
        print LPW $str;
    }
}

############################################################################

sub prx {
my $v = shift;
my $s = sprintf ("%x", $v);
print $s;
}

sub prhex {
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf(" %01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf(" %02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf(" %03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf(" %04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf(" %05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf(" %06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf(" %07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf(" %08X", $v);
	}
	# print "$s ";
	return $s;
}

sub prhex0 {
    ######no spaces around the string#########
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf("%01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf("%02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf("%03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf("%04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf("%05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf("%06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf("%07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf("%08X", $v);
	}
	return $s;
}

sub rtrim {
	my $s = shift;
	my $l = length($s);
	my $i = $l;
	if ($l > 0)
	{
		while($i > 0)
		{
			if((substr($s,$i-1,1)) ne " ")
			{ 
				last;
			}
			else
			{
				$i=$i-1;
			}
		}
	}
	else
	{
		return "";
	}
	if($i == 0)
	{
		return "";
	}
	else
	{
		return substr($s,0,$i);
	}
}

sub proct
{
	# convert 16bit number into an  octal string
	my $n=shift;
	$n=$n & 0xFFFF;
	my $d1=$n >> 15 + "0";
	my $d2= (($n >> 12) & 0x7) + "0";
	my $d3= (($n >>  9) & 0x7) + "0";
	my $d4= (($n >>  6) & 0x7) + "0";	
	my $d5= (($n >>  3) & 0x7) + "0";	
	my $d6= ($n & 0x7) + "0";
	return " " . $d1 . $d2 . $d3 . $d4 . $d5 . $d6;
}

sub prdec
{
	# print number
	# $n is number of digits (string)
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	my $format="\%0" . $n . "d";
	$s = sprintf($format, $v);
	# print "$s ";
	return "  $s  ";
}

sub prdec0
{
    ######no spaces around the string#########
	# print number
	# $n is number of digits (string)
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	my $format="\%0".$n."d";
	$s = sprintf($format, $v);
	return $s;
}

# convert hex string in expression to binary
sub cvhex
{
	my $s = shift;
	my $i = 0;
	my $r=0;
	my $d=0;
	my $c;
	if(($c=substr($s,$i,1)) =~ /[0-9A-F]/ )	# at least 1 hex digit required
	{
		while (($c = substr($s,$i,1)) =~ /[0-9A-F]/)
		{
			$d = $cvdig{$c};
			$r = ($r << 4) + $d;
			$i = $i + 1;
		}
		
		if ($r > 0xFFFF)
		{
			return -1;
		}
		return $r;
	}
	else
	{
		return -1;
	}
}

# convert decimal string in expression to binary
sub cvdec
{
	my $s = shift;
	my $i = 0;
	my $r=0;
	my $d=0;
	my $c;
	if(($c=substr($s,$i,1)) =~ /[0-9]/ )	# at least 1 hex digit required
	{
		while (($c = substr($s,$i,1)) =~ /[0-9]/)
		{
			$d = $cvdig{$c};
			$r = ($r * 10) + $d;
			$i = $i + 1;
		}
		
		if ($r > 0xFFFF)
		{
			return -1;
		}
		return $r;
	}
	else
	{
		return -1;
	}
}

# translate string to upper case
sub UP {
	my $s = shift;
    $_ = $s;
	tr/a-z/A-Z/;
	return $s = $_;
}

# read octad from papertape, attached to filecode IN (used by SCP load commands)
sub nextchar {
	my $c;
	read(IN,$c,1);
	$c = unpack("C", $c); 	#get new current char
	return ($c & 0xFF);
}

# read word (2 octads) from papertape
sub nextword {
	my $w = nextchar();
	$w = $w << 8;
	$w = $w + nextchar();;
	return $w;
}

# read octad from papertape, attached to filecode PTR
sub read_ptr {
	my $c;
	if(read($PTR[FC],$c,1) )
	{
		$c = unpack("C", $c) & 0xFF; 	 # get next char
		print(prhex(8, $c) . "\r");     # show progress
		return ($c | 0x8000);            # nonzero if not eot
	}
	else
	{
		return 0;
	}
}

# punch a byte on papertape connected to filecode PTP
sub punch_ptp {
    my($a) = @_;
    my $byte;
	my $fc = $PTP[FC];
    $byte = pack("C", ($a & 0xFF) );
    print $fc $byte;
}

sub punch {
    my($a) = @_;
    my $byte;
    $byte = pack("C", ($a & 0xFF) );
    print BIN $byte;
}


# The instruction decoder
############################# decoder data ###########################

sub initdecoder {
	# Fill tables
	
# direct translated instruction codes
$IDir{0xC020} = 'MUK';
$IDir{0xC820} = 'DVK';
$IDir{0xD020} = 'DAK';
$IDir{0xD820} = 'DSK';
$IDir{0x8F20} = 'ABL';
$IDir{0x8E20} = 'ABL(6)';
$IDir{0x8D20} = 'ABL(5)';
$IDir{0x8C20} = 'ABL(4)';
$IDir{0x8B20} = 'ABL(3)';
$IDir{0x8A20} = 'ABL(2)';
$IDir{0x8920} = 'ABL(1)';
$IDir{0x8820} = 'ABL(0)';
$IDir{0xF021} = 'EXK';
$IDir{0x207F} = 'HLT';
$IDir{0x20BF} = 'INH';
$IDir{0x2840} = 'ENB';
$IDir{0x2804} = 'LKM';
$IDir{0x2801} = 'SMD';
$IDir{0xC900} = 'FFL';
$IDir{0xC901} = 'FFX';
$IDir{0x6800} = 'MVS'; # P854/P858/P859
$IDir{0x6840} = 'MVA'; # P854/P858/P859
$IDir{0x6880} = 'CS';  # P854/P858/P859
$IDir{0x68A0} = 'CA';  # P854/P858/P859

$IDirl{0xC020} = 2;
$IDirl{0xC820} = 2;
$IDirl{0xD020} = 3;
$IDirl{0xD820} = 3;
$IDirl{0x8F20} = 2;
$IDirl{0x8E20} = 2;
$IDirl{0x8D20} = 2;
$IDirl{0x8C20} = 2;
$IDirl{0x8B20} = 2;
$IDirl{0x8A20} = 2;
$IDirl{0x8920} = 2;
$IDirl{0x8820} = 2;
$IDirl{0xF001} = 2;
$IDirl{0x207F} = 1;
$IDirl{0x20BF} = 1;
$IDirl{0x2840} = 1;
$IDirl{0x2804} = 2;
$IDirl{0x2801} = 1;
$IDirl{0xC900} = 1;
$IDirl{0xC901} = 1;
$IDirl{0x6800} = 1;  # P854/P858/P859
$IDirl{0x6840} = 1;  # P854/P858/P859
$IDirl{0x6880} = 1;  # P854/P858/P859
$IDirl{0x68A0} = 1;  # P854/P858/P859

# format 1 long constant instruction codes (r2=0)

$IL1{0x8020} = 'LDKL';
$IL1{0x9020} = 'ADKL';
$IL1{0x9820} = 'SUKL';
$IL1{0xE820} = 'CWK';
$IL1{0xA020} = 'ANKL';
$IL1{0xA820} = 'ORKL';
$IL1{0xB020} = 'XRKL';
$IL1{0xE020} = 'LCK';
$IL1{0xE821} = 'CCK';
$IL1{0xF021} = 'CF';   # r1 # 0 (else EXK: r1=0, r2=0)

# format 1 instructions with r1=0

$ILS{0x9821} = 'C2R';
$ILS{0x9841} = 'C2';
$ILS{0x9861} = 'C2*';
$ILS{0x9021} = 'IMR';
$ILS{0x9041} = 'IM';
$ILS{0x9061} = 'IM*';
$ILS{0xA021} = 'CMR';
$ILS{0xA041} = 'CM';
$ILS{0xA061} = 'CM*';
$ILS{0xF020} = 'RTN';
$ILS{0xF001} = 'EXR';
$ILS{0xF021} = 'EXR*';
$ILS{0xF041} = 'EX';
$ILS{0xF061} = 'EX*';
$ILS{0xB820} = 'TLR';
$ILS{0xB840} = 'TL';
$ILS{0xB860} = 'TL*';
$ILS{0xB821} = 'TSR';
$ILS{0xB841} = 'TS';
$ILS{0xB861} = 'TS*';
$ILS{0xC041} = 'TSB';
$ILS{0xC061} = 'TSB*';
$ILS{0xC021} = 'TSBR';
$ILS{0xC841} = 'TRB';
$ILS{0xC861} = 'TRB*';
$ILS{0xC821} = 'TRBR';
$ILS{0xD041} = 'TB';
$ILS{0xD061} = 'TB*';
$ILS{0xD021} = 'TBR';
$ILS{0xD840} = 'DS';
$ILS{0xD860} = 'DS*';
$ILS{0xD800} = 'DSR';
$ILS{0xD820} = 'DSR*';
$ILS{0xD040} = 'DA';
$ILS{0xD060} = 'DA*';
$ILS{0xD000} = 'DAR';
$ILS{0xD020} = 'DAR*';
$ILS{0xC840} = 'DV';
$ILS{0xC860} = 'DV*';
$ILS{0xC800} = 'DVR';
$ILS{0xC820} = 'DVR*';
$ILS{0xC040} = 'MU';
$ILS{0xC060} = 'MU*';
$ILS{0xC000} = 'MUR';
$ILS{0xC020} = 'MUR*';

$ILS{0xF821} = 'C1RS';	# r1 = 0
$ILS{0xF841} = 'C1S';	# r1 = 0
$ILS{0xF861} = 'C1S*';	# r1 = 0

# format 1 instruction codes, using r1 and r2

$I1{0x8040} = 'LD';
$I1{0x8060} = 'LD*';
$I1{0x8000} = 'LDR';
$I1{0x8020} = 'LDR*';
$I1{0x8041} = 'ST';
$I1{0x8061} = 'ST*';
$I1{0x8021} = 'STR';

$I1{0x9000} = 'ADR';
$I1{0x9020} = 'ADR*';
$I1{0x9021} = 'ADRS';	# r1 # 0
$I1{0x9040} = 'AD';
$I1{0x9041} = 'ADS';	# r1 # 0
$I1{0x9060} = 'AD*';
$I1{0x9061} = 'ADS*';	# r1 # 0

$I1{0x9800} = 'SUR';
$I1{0x9820} = 'SUR*';
$I1{0x9821} = 'SURS';	# r1 # 0
$I1{0x9840} = 'SU';
$I1{0x9841} = 'SUS';	# r1 # 0
$I1{0x9860} = 'SU*';
$I1{0x9861} = 'SUS*';	# r1 # 0

$I1{0xE800} = 'CWR';
$I1{0xE820} = 'CWR*';
$I1{0xE840} = 'CW';
$I1{0xE860} = 'CW*';

$I1{0xF800} = 'C1R';
$I1{0xF820} = 'C1R*';

$I1{0xF840} = 'C1';

$I1{0xF860} = 'C1*';

$I1{0x9801} = 'NGR';

$I1{0xA000} = 'ANR';	# r1 # 0
$I1{0xA020} = 'ANR*';
$I1{0xA021} = 'ANRS';	# r1 # 0
$I1{0xA040} = 'AN';
$I1{0xA041} = 'ANS';	# r1 # 0
$I1{0xA060} = 'AN*';
$I1{0xA061} = 'ANS*';	# r1 # 0

$I1{0xA001} = 'TM';	# r1 # 0

$I1{0xA800} = 'ORR';
$I1{0xA820} = 'ORR*';
$I1{0xA821} = 'ORRS';	# r1 # 0
$I1{0xA840} = 'OR';
$I1{0xA841} = 'ORS';	# r1 # 0
$I1{0xA860} = 'OR*';
$I1{0xA861} = 'ORS*';	# r1 # 0

$I1{0xB000} = 'XRR';	# r1 # 0
$I1{0xB020} = 'XRR*';   # r1 # 0
$I1{0xB021} = 'XRRS';	# r1 # 0
$I1{0xB040} = 'XR';     # r1 # 0
$I1{0xB041} = 'XRS';	# r1 # 0
$I1{0xB060} = 'XR*';    # r1 # 0
$I1{0xB061} = 'XRS*';	# r1 # 0

$I1{0xB001} = 'TNM';	# r1 # 0

$I1{0xE000} = 'ECR';	# r1 # 0
$I1{0xE020} = 'LCR';    # r1 # 0
$I1{0xE021} = 'SCR';	# r1 # 0
$I1{0xE040} = 'LC';     # r1 # 0
$I1{0xE041} = 'SC'; 	# r1 # 0
$I1{0xE060} = 'LC*';    # r1 # 0
$I1{0xE061} = 'SC*';	# r1 # 0

$I1{0xE821} = 'CCR';    # r1 # 0
$I1{0xE841} = 'CC'; 	# r1 # 0
$I1{0xE861} = 'CC*';	# r1 # 0

$I1{0xF001} = 'CFR';    # r1 # 0
$I1{0xF021} = 'CFR*';   # r1 # 0
$I1{0xF041} = 'CFI'; 	# r1 # 0
$I1{0xF061} = 'CFI*';	# r1 # 0

$I1{0xD040} = 'EL';
$I1{0xD060} = 'EL*';
$I1{0xD020} = 'ELR';
$I1{0xD041} = 'ES';
$I1{0xD061} = 'ES*';
$I1{0xD021} = 'ESR';

# format 1 FPP instruction codes using r2 (r1 is nonzero)
$IFPP{0xC120}  = 'FLDR';
$IFPP{0xC140}  = 'FLD';
$IFPP{0xC160}  = 'FLD*';
$IFPP{0xC121}  = 'FSTR';
$IFPP{0xC141}  = 'FST';
$IFPP{0xC161}  = 'FST*';
$IFPP{0xCC20}  = 'FADR';
$IFPP{0xCC21}  = 'FADRS';
$IFPP{0xCC40}  = 'FAD';
$IFPP{0xCC41}  = 'FADS';
$IFPP{0xCC60}  = 'FAD*';
$IFPP{0xCC61}  = 'FADS*';
$IFPP{0xCD20}  = 'FSUR';
$IFPP{0xCD21}  = 'FSURS';
$IFPP{0xCD40}  = 'FSU';
$IFPP{0xCD41}  = 'FSUS';
$IFPP{0xCD60}  = 'FSU*';
$IFPP{0xCD61}  = 'FSUS*';
$IFPP{0xCE20}  = 'FMUR';
$IFPP{0xCE21}  = 'FMURS';
$IFPP{0xCE40}  = 'FMU';
$IFPP{0xCE41}  = 'FMUS';
$IFPP{0xCE60}  = 'FMU*';
$IFPP{0xCE61}  = 'FMUS*';
$IFPP{0xCF20}  = 'FDVR';
$IFPP{0xCF21}  = 'FDVRS';
$IFPP{0xCF40}  = 'FDV';
$IFPP{0xCF41}  = 'FDVS';
$IFPP{0xCF60}  = 'FDV*';
$IFPP{0xCF61}  = 'FDVS*';

# register strings: A0..A15
$R2[0]  = 'A0';
$R2[2]  = 'A1';
$R2[4]  = 'A2';
$R2[6]  = 'A3';
$R2[8]  = 'A4';
$R2[10] = 'A5';
$R2[12] = 'A6';
$R2[14] = 'A7';
$R2[1]  = 'A8';
$R2[3]  = 'A9';
$R2[5]  = 'A10';
$R2[7]  = 'A11';
$R2[9]  = 'A12';
$R2[11] = 'A13';
$R2[13] = 'A14';
$R2[15] = 'A15';

# register strings: A0..A7
$R3[0] = 'A0';
$R3[1] = 'A1';
$R3[2] = 'A2';
$R3[3] = 'A3';
$R3[4] = 'A4';
$R3[5] = 'A5';
$R3[6] = 'A6';
$R3[7] = 'A7';

# condition codes: (0)..(7)
$CND[0] = '(0)';
$CND[1] = '(1)';
$CND[2] = '(2)';
$CND[3] = '(3)';
$CND[4] = '(4)';
$CND[5] = '(5)';
$CND[6] = '(6)';
$CND[7] = '(7)';
}


sub decode {
    # decodes the instruction at physical address $pc
    
    my $PC = $pcsaved;
    my $M  = $M[$PC] & 0xFFFF;
    my $s1 = " --" . prhex(16,$PC) . " (" . prhex0(16,($PC-$latrace)) . ")";
    my $s2 = "   " . prhex(16,$M); # instruction hex
    my $str= chars($M);             # translated to characters
    my $s;
    my $l;
	my $nl = "\n";
    my $result1 = $s1 . $s2 . $str;
    my $result2 = "";
    $PC=$PC+2;                      # operand address
    
    if($M == 0)
    {
        $s = "DATA 0";
        $result1 = $result1 . $s;
    }
    elsif($IDir{$M} ne undef)
    {
        # direct printable instruction
        $s=$IDir{$M};               # instruction mnemonic
        $result1 = $result1 . $s;
        $l=$IDirl{$M};              # instruction length (words)
        if($l>1)
        {
			# operand; 1st word
	        $M = $M[$PC];           # operand, 1st word
			$str=chars($M);         # word translated to characters
			$result1 = $result1 . ' ' . prhex(16,$M);
            $result2 = "   " . prhex(16,$PC) . '          ' . prhex(16,$M) . $str;
            $PC=$PC+2;              # operand address; 2nd word
        }
        if($l>2)
        {
			# operand; 2nd word
            $M = $M[$PC];           # operand, 2nd word
            $str=chars($M);         # word translated to characters				
            $result1 = $result1 . ',' . prhex(16,$M);
            $result2 = $result2 . $nl . "   " .  prhex(16,$PC) . '          ' . prhex(16,$M) . $str;			
            $PC=$PC+2;
        }
        if($result2 ne "")
        {
            $result1 =  $result1 . $nl . $result2;
        }
    }
    else
    {
        $s = decodeformat0($M);
        if($s)
        {
            $result1 = $result1 . $s;
        }
        else
        {
            $s = decodeformat1($M);
            if($s)
            {
                $result1 = $result1 . $s;
            }
            else
            {
                $s = "DATA  " . prhex(16,$M);
                $result1 = $result1 . $s;
            }
        }
    }
    return $result1;
}

sub chars
{
	my $cc = shift;
	my $c = $cc >> 8;
	my $str = "  ";
	if($ch{$c} ne undef)
	{
		$str = $str . $ch{$c};
	}
	else
	{
		$str = $str . '.';
	}
	$c = $cc & 0xFF;
	if($ch{$c} ne undef)
	{
		$str = $str . $ch{$c};
	}
	else
	{
		$str = $str . '.';
	}
	$str = $str . '| ';
	return $str;
}

sub decodeformat0
{
	my $ir  = shift;
	if($ir & 0x8000) # format 0 ?
	{
		return "";    # no => format 1
	}
	my $PC = $pcsaved;
	my $opc = ($ir & 0xF800) >> 11;
	my $r3  = ($ir & 0x0700) >> 8;
	my $k   = $ir & 0xFF;
	my $r2;
	my $D;
	my $s;
	my $dev;
	my $n;
	my $sc;
	my $loc;
	
	if($opc == 0)
	{
		# LDK r3,k
		$s= 'LDK ';
		$s= $s . $R3[$r3] . ',' . prhex0(8,$k);
		#print $s;
		return $s;
	}
	elsif($opc == 1)
	{
		# AB cnd,k		
		$s = 'AB';
		$s= $s . $CND[$r3] . ' ' . prhex0(8,$k);
		#print $s;			
		return $s;	
	}
	elsif($opc == 2)
	{
		# ADK r3,k
		$s= 'ADK ';
		$s= $s . $R3[$r3] . ',' . prhex0(8,$k);
		#print $s;
		return $s;		
	}	
	elsif($opc == 3)
	{
		# SUK r3,k
		$s= 'SUK ';
		$s= $s . $R3[$r3] . ',' . prhex0(8,$k);
		#print $s;
		return $s;		
	}
	elsif($opc == 4)
	{
		# ANK r3,k
		# RIT k
		$ir=$ir & 0xFFC0;
		if($ir == 0x20C0)
		{
			$s = 'RIT ';
			$s= $s . prhex0(8,$k);
		}
		else
		{
			$s = 'ANK ';
			$s= $s . $R3[$r3] . ',' . prhex0(8,$k);
		}
		#print $s;
		return $s;			
	}	
	elsif($opc == 5)
	{
		# ORK r3,k
		$s= 'ORK ';
		$s= $s . $R3[$r3] . ',' . prhex0(8,$k);
		#print $s;
		return $s;			
	}		
	elsif($opc == 6)
	{
		# XRK r3,k
		$s= 'XRK ';
		$s= $s . $R3[$r3] . ',' . prhex0(8,$k);
		#print $s;
		return $s;			
	}
	elsif($opc == 7)
	{
		# shift instructions
		$sc = ($ir & 0xE0) >> 5;  # shift code
		$n  = $ir & 0x1F;         # nbr of shifts      
		$r2 = ($n & 0x1E) >> 1;
		# print "sc=$sc n=$n r3=$r3\n";
		if($r3 == 0)
		{
			# double shifts
			if($sc == 0)
			{
				$s = 'DLA ';
			}
			elsif($sc == 1)
			{
				$s = 'DRA ';
			}
			elsif($sc == 2)
			{
				$s = 'DLL ';
			}
			elsif($sc == 3)
			{
				$s = 'DRL ';
			}
			elsif($sc == 6)
			{
				$s = 'DLC ';
			}			
			elsif($sc == 7)
			{
				$s = 'DRC ';
			}
			else
			{
				if($sc == 4)
				{
					$s = 'DLN ';
				}
				elsif($sc == 5)
				{
					$s = 'DRN ';
				}
				else
				{
					prlog("Instruction Decode Error: Shift Double\n");
					return "";
				}
				$s=$s . $R2[$r2];
				#print $s;
				return $s;				
			}
			$s=$s . prhex(8,$n);
			#print $s;
			return $s;
		}
		else
		{
			if($sc == 0)
			{
				$s = 'SLA ';
			}
			elsif($sc == 1)
			{
				$s = 'SRA ';
			}
			elsif($sc == 2)
			{
				$s = 'SLL ';
			}
			elsif($sc == 3)
			{
				$s = 'SRL ';
			}
			elsif($sc == 6)
			{
				$s = 'SLC ';
			}			
			elsif($sc == 7)
			{
				$s = 'SRC ';
			}
			else
			{
				if($sc == 4)
				{
					$s = 'SLN ';
				}
				elsif($sc == 5)
				{
					$s = 'SRN ';
				}
				else
				{
					prlog("Instruction Decode Error: Shift Single\n");
					return "";
				}
				$s=$s . $R3[$r3] . ',' . $R2[$r2];
				#print $s;
				return $s;				
				
			}
			$s=$s . $R3[$r3] . ',' . prhex0(8,$n);
			#print $s;
			return $s;
		}
	}
	elsif($opc == 8)
	{
		# CIO r3,f,dev		
		# OTR r3,f,dev
		if($ir & 0x80)
		{
			# CIO
			$s = 'CIO ';
		}
		else
		{
			# OTR
			$s = 'OTR ';
		}
		$s = $s . $R3[$r3];
		if($ir & 0x40)
		{
			# f=1
			$s=$s . ',1';
		}
		else
		{	
			# f=0
			$s=$s . ',0';
		}
		$dev=$ir & 0x3F;
		$s=$s . ',' . prhex0(8,$dev);
		#print $s;
		return $s;		
	}		
	elsif($opc == 9)
	{
		# INR r3,f,dev
		# SST r3,dev
		# TST r3,dev
		if($ir & 0x80)
		{
			# SST r3,dev
			# TST r3,dev
			if($ir & 0x40)
			{
				$s = 'SST ';
			}
			else
			{
				$s = 'TST ';
			}
			$s = $s . $R3[$r3];	
			$dev=$ir & 0x3F;
			$s=$s . ',' . prhex0(8,$dev);
			#print $s;
			return $s;				
		}
		else
		{
			# INR r3,f,dev
			$s = 'INR ';
			$s = $s . $R3[$r3];
			if($ir & 0x40)
			{
				# f=1
				$s=$s . ',1';
			}
			else
			{	
				# f=0
				$s=$s . ',0';
			}
			$dev=$ir & 0x3F;
			$s=$s . ',' . prhex0(8,$dev);
			#print $s;
			return $s;					
		}
	}	
	elsif($opc == 0xA)
	{
		# RF cond,D	
		$s = 'RF';
		$D = $k + 2;
		$loc = $PC + $D;		
		$s= $s . $CND[$r3] . ' ' . '*+' . prhex0(8,$D);
		$s = $s . ' [logical' . prhex(16,$loc) . ']';
		#print $s;
		return $s;			
	}		
	elsif($opc == 0xB)
	{
		# RB cond,D
		$s = 'RB';
		$D = $k - 2;
		$loc = $PC - $D;
		$s = $s . $CND[$r3] . ' ' . '*-' . prhex0(8,$D);
		$s = $s . ' [logical:' . prhex(16,$loc) . ']';
		#print $s;
		return $s;			
	}
	elsif($opc == 0xD)
	{
		# string MVS
		# string MVA
		# string CS
		# string CA
        # printed directly
	}	
	elsif($opc == 0xE)
	{
		# WER r3,addr
		# move table forward MVF
		# move table from system to user area MVSU
		$r2=($ir & 0x1E) >> 1;
		$ir=$ir & 0xF080;
        if($r3 == 0)
        {
            if($ir == 0x7000)
            {
                # MVF
                $s = 'MVF ';
                $s = $s . $R2[$r2];
                #print $s;
                return $s;				
            }
            elsif($ir == 0x7080)
            {
                # MVSU
                $s = 'MVSU ';
                $s = $s . $R2[$r2];	
                #print $s;
                return $s;				
            }
            else
            {
                return "";
            }
        }
		else
		{
			$s = 'WER ';
			$s = $s . $R3[$r3] . ',' . prhex0(8,$k);
			#print $s;
			return $s;
		}
	}	
	elsif($opc == 0xF)
	{
		# RER r3,addr
		# move table backward MVB
		# move table from user to system area MVUS
		$r2=($ir & 0x1E) >> 1;
		$ir=$ir & 0xF080;
        if($r3 == 0)
        {
            if($ir == 0x7000)
            {
                # MVB
                $s = 'MVB ';
                $s = $s . $R2[$r2];
                #print $s;
                return $s;				
            }
            elsif($ir == 0x7080)
            {
                # MVUS
                $s = 'MVUS ';
                $s = $s . $R2[$r2];	
                #print $s;
                return $s;
			}
            else
            {
                return "";
            }
		}
		else
		{
			$s = 'RER ';
		    $s = $s . $R3[$r3] . ',' . prhex0(8,$k);
			#print $s;
			return $s;
		}			
	}	
	return "";
}

sub decodeformat1
{
	my $ir = shift;
    my $M;
    my $PC = $pcsaved + 2;
	my $opc = ($ir & 0x7800) >> 11;
	my $r1  = ($ir & 0x0780) >> 7;
	my $r2  = ($ir & 0x001E) >> 1;
	my $md  = ($ir & 0x60)   >> 5;
	my $iml = $ir & 0xF87F;   # long constant instr mask (mask r1)
	my $ims = $ir & 0xFFE1;   # mask instr with r1=0 (mask r2)
	my $im1 = $ir & 0xF861;   # r1-r2 instr mask
    my $nl = "\n";
    my $str;
	my $s="";
    my $s1="";

	if($s = f1exp($ir)) # exceptions
	{
		return $s; # format 1 exeptions: ABL and ML/MS
	}
	elsif($IL1{$iml} ne undef) # check 1st for a long constant instruction (r2 == 0)
	{
		# long constant instructions (r2==0)
		$s=$IL1{$iml};     # code
        # operand
		$M = $M[$PC];
		$str=chars($M);
        $s=$s . ' ' . $R2[$r1] . ',' . prhex0(16,$M);
		$s1= "   " . prhex(16,$PC) . '          ' . prhex(16,$M) . $str;
		$PC=$PC+2;
		return $s . $nl . $s1;
	}
	elsif($ILS{$ims} ne undef)
	{
		# instr with r1 == 0
		$s=$ILS{$ims};
		if($md == 2 || $md == 3)
		{
            # operand
			$M = $M[$PC];
			$str=chars($M);
            $s=$s . prhex(16,$M);
			$s1= "   " . prhex(16,$PC) . '          ' . prhex(16,$M) . $str;			
			$PC=$PC+2;
			if($r2 != 0)
			{
				$s=$s . ',' . $R2[$r2];
			}
		}
		else
		{
			$s=$s . ' ' . $R2[$r2];
			$s1="";
		}
		if($s1 ne "")
		{
            $s = $s . $nl . $s1;
		}
		return $s;
	}
	elsif($I1{$im1} ne undef)
	{
		# normal format 1 instr with both r1 and r2
		$s=$I1{$im1};
		if($md == 2 || $md == 3)
		{
            # operand
			$M = $M[$PC];
			$str=chars($M);
            $s=$s . ' ' . $R2[$r1] . ',' . prhex0(16,$M);
			$s1= "   " . prhex(16,$PC) . '          ' . prhex(16,$M) .  $str;				
			$PC=$PC+2;
			if($r2 != 0)
			{
				$s=$s . ',' . $R2[$r2];
			}
		}
		else
		{
			$s=$s . ' ' . $R2[$r1] . ',' . $R2[$r2];
			$s1="";
		}
		if($s1 ne "")
		{
            $s = $s . $nl . $s1;
		}
		return $s;
	}	
	return "";
}

# format 1 exceptions
sub f1exp
{
	my $ir = shift;
    my $M;
    my $PC = $pcsaved + 2;
    my $str;
	my $opc = ($ir & 0x7800) >> 11;
	my $n   = ($ir & 0x0780) >> 7;
    my $r1  = ($ir & 0x0780) >> 7;
	my $r2  = ($ir & 0x001E) >> 1;
	my $md  = ($ir & 0x60)   >> 5;
	my $cnd = $n >> 1;
    my $nl = "\n";
	my $s;	
	my $s1="";
	my $i;
	
	if(($opc == 1) && (($ir & 1) == 0))
	{
		# absolute conditional branch
		if($md == 0)
		{
			$s = 'ABR';
			$s = $s . $CND[$cnd] . ' ' . $R2[$r2];
		}
		elsif($md == 1)
		{
			$s = 'ABR';
            $s = $s . $CND[$cnd] . '* ' . $R2[$r2];
		}
		elsif($md == 2)
		{
			$s = 'ABI';
			$s = $s . $CND[$cnd];
			$M = $M[$PC];
			$str=chars($M);			
			$s=$s . ' ' . prhex(16,$M);
			$s1= "   " . prhex(16,$PC) . '    ' . prhex(16,$M) . $str;				
			$PC=$PC+2;
			if($r2 != 0)
			{
				$s=$s . ',' . $R2[$r2];
			}          
		}
		elsif($md == 3)
		{
			$s = 'ABI';
			$s = $s . $CND[$cnd] . '*';
			$M = $M[$PC];
			$str=chars($M);			
			$s=$s . ' ' . prhex(16,$M);
			$s1= "   " . prhex(16,$PC) . '    ' . prhex(16,$M) . $str;				
			$PC=$PC+2;
			if($r2 != 0)
			{
				$s=$s . ',' . $R2[$r2];
			}         
		}
		if($s1 ne "")
		{
            $s = $s . $nl . $s1;
		}		
		return $s;
	}
	elsif($opc == 7)
	{
        if($r1 == 0)
        {
            return 0;
        }
		# ML/MS
		if($ir & 1)
		{
			# MS
			if($md == 1)
			{
				$s='MSR ';
                $s=$s . prdec0(2,$n) . ',' . $R2[$r2];
				return $s;
			}
			elsif($md == 2)
			{
				$s='MS ';
				$M = $M[$PC];
				$str=chars($M);				
				$s=$s . prdec0(2,$n) . ',' . prhex0(16,$M);
				$s1= "   " . prhex(16,$PC) . '          ' . prhex(16,$M) .  $str;
				$PC=$PC+2;
				if($r2 != 0)
				{
					$s=$s . ',' . $R2[$r2];
				}
				if($s1 ne "")
				{
                    $s = $s . $nl . $s1;
				}
				return $s;
			}
			elsif($md == 3)
			{
				$s='MS* ';
				$M = $M[$PC];
				$str=chars($M);				
				$s=$s . prdec0(2,$n) . ',' . prhex0(16,$M);
				$s1="   " . prhex(16,$PC) . '          ' . prhex(16,$M) .  $str;
				$PC=$PC+2;
				if($r2 != 0)
				{
					$s=$s . ',' . $R2[$r2];
				}
				if($s1 ne "")
				{
                    $s = $s . $nl . $s1;
				}
				return $s;
			}
			else
			{
				return "";
			}
		}
		else
		{
			# ML
			if($md == 1)
			{
				if($r2 == 0)
				{
					# MLK	
					$s='MLK ';
					$s=$s . prdec0(2,$n);
                    $i=0;
					while($i<$n)
					{
						$s = $s . $nl;
						$M = $M[$PC];
						$str=chars($M);						
						$s=$s . "   " . prhex(16,$PC) . '          ' . prhex(16,$M) .  $str;
						$PC=$PC+2;
						$i=$i+1;
					}
					return $s;
				}
				else
				{
					# MLR
					$s='MLR ';
                    $s=$s . prdec0(2,$n) . ',' . $R2[$r2];
					return $s;
				}
			}
			elsif($md == 2)
			{
				# ML
				$s='ML ';
				$M = $M[$PC];
				$str=chars($M);				
				$s=$s . prdec0(2,$n) . ',' . prhex(16,$M);
				$s1= "   " . prhex(16,$PC) . '          ' . prhex(16,$M) .  $str;
				$PC=$PC+2;
				if($r2 != 0)
				{
					$s=$s . ',' . $R2[$r2];
				}
				if($s1 ne "")
				{
		            $s = $s . $nl. $s1;
				}
				return $s;
			}
			elsif($md == 3)
			{
				# ML*
				$s='ML* ';
				$M = $M[$PC];
				$str=chars($M);				
				$s=$s . prdec(2,$n) . ',' . prhex(16,$M);
				$s1= "   " . prhex(16,$PC) . '          ' . prhex(16,$M) .  $str;
				$PC=$PC+2;
				if($r2 != 0)
				{
					$s=$s . ',' . $R2[$r2];
				}
                if($s1 ne "")
				{
		            $s = $s . $nl. $s1;
				}
				return $s;
			}
        }
    }
	else
	{
		return "";
	}
	return "";
}









